package org.processmining.plugins.abstractions.encoding;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.JFrame;
import javax.swing.JOptionPane;

import org.deckfour.xes.extension.XExtension;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.out.XesXmlGZIPSerializer;
import org.processmining.plugins.abstractions.datatypes.EncodedLog;
import org.processmining.plugins.abstractions.patterns.MinePatterns;
import org.processmining.plugins.abstractions.util.EquivalenceClass;

/**
 * 
 * @author R. P. Jagadeesh Chandra 'JC' Bose
 * @date 15 May 2011 Eindhoven University of Technology (TU/e)
 * 
 *       Reproducing or using whole or part of this code/application is to be done with proper
 *       acknowledgments by citing "R. P. Jagadeesh Chandra Bose and Wil M. P.
 *       van der Aalst. Abstractions in Process Mining: A Taxonomy of Patterns.
 *       Proceedings of the International Conference on Business Process
 *       Management (BPM). Volume 5701 of LNCS., Springer-Verlag (2009) pages
 *       159-175"
 * 
 */
public class TransformLog {
	XLog log;
	EncodedLog encodedLog;
	String transformLogDirectory;
	String transformLogFileName;

	Set<Set<String>> individualActivitiesAbstractionSet;
	Map<Set<String>, Set<String>> pAANSetMap;
	Set<String> sNIEASet;
	Map<String, Set<Set<String>>> aNPASetMap;
	Map<Set<String>, TreeSet<String>> patternAlphabetPatternSetMap;
	
	boolean debug = false;
	
	public TransformLog(XLog log, EncodedLog encodedLog, String transformLogAbsoluteFileName){
		this.log = log;
		this.encodedLog = encodedLog;
		this.transformLogFileName = transformLogAbsoluteFileName;
	}
	
	class DescendingLengthComparator implements Comparator<String>{
		public int compare(String str1, String str2) {
			int lengthDiff = str2.length()-str1.length();
			return (lengthDiff != 0? lengthDiff : str2.compareTo(str1));
		}
	}
	
	public TransformLog(){

	}
	
	private void updatePatternAlphabetPatternSetMapWithCyclicPermutations(){
		int encodingLength=encodedLog.getEncodingLength();
		Set<String> cyclicPermutationSet = new TreeSet<String>(new DescendingLengthComparator());
		String repeat,permutedRepeat;
		int activityStartIndex;
		Set<String> nonStartingActivitySet = new HashSet<String>();
		TreeSet<String> patternAlphabetPatternSet;
		for(Set<String> patternAlphabet : patternAlphabetPatternSetMap.keySet()){
			patternAlphabetPatternSet = patternAlphabetPatternSetMap.get(patternAlphabet);
			if(patternAlphabetPatternSet.size() == patternAlphabet.size())
				continue;
			else{
				repeat = patternAlphabetPatternSet.iterator().next();
				
				cyclicPermutationSet.clear();
				cyclicPermutationSet.addAll(patternAlphabetPatternSet);
				
				nonStartingActivitySet.clear();
				nonStartingActivitySet.addAll(patternAlphabet);
				
				for(String tr : patternAlphabetPatternSet)
					nonStartingActivitySet.remove(tr.substring(0, encodingLength));
				
				for(String nonStartingActivity : nonStartingActivitySet){
					activityStartIndex = repeat.indexOf(nonStartingActivity)/encodingLength;
					permutedRepeat = repeat.substring(activityStartIndex*encodingLength);
					permutedRepeat += repeat.substring(0,activityStartIndex*encodingLength);
					cyclicPermutationSet.add(permutedRepeat);
				}
				
				patternAlphabetPatternSet.addAll(cyclicPermutationSet);
				patternAlphabetPatternSetMap.put(patternAlphabet, patternAlphabetPatternSet);
			}
		}
	}
	
	private void printPatternAlphabetAbstractionNameSetMap(){
		Set<String> decodedPatternAlphabet = new HashSet<String>();
		Map<String, String> charActivityMap = encodedLog.getCharActivityMap();
		for(Set<String> patternAlphabet : pAANSetMap.keySet()){
			decodedPatternAlphabet.clear();
			for(String encodedActivity : patternAlphabet)
				decodedPatternAlphabet.add(charActivityMap.get(encodedActivity));
		}
	}
	
	public void transform(Map<Set<String>, TreeSet<String>> patternAlphabetPatternSetMap, Map<Set<String>, Set<String>> patternAlphabetAbstractionNameSetMap, Set<String> sNIEASet){
		this.patternAlphabetPatternSetMap = patternAlphabetPatternSetMap;
		this.pAANSetMap = patternAlphabetAbstractionNameSetMap;
		this.sNIEASet = sNIEASet;
		
		printPatternAlphabetAbstractionNameSetMap();
		
		updatePatternAlphabetPatternSetMapWithCyclicPermutations();
		
		Set<String> activitiesInvolvedInAbstractionSet = new HashSet<String>();
		for(Set<String> patternAlphabet : patternAlphabetAbstractionNameSetMap.keySet()){
			activitiesInvolvedInAbstractionSet.addAll(patternAlphabet);
		}
		
		Set<String> pAANSet;
		Map<String, Set<String>> aANSetMap = new HashMap<String, Set<String>>();
		Set<String> aANSet;
		for(Set<String> patternAlphabet : patternAlphabetAbstractionNameSetMap.keySet()){
			pAANSet = patternAlphabetAbstractionNameSetMap.get(patternAlphabet);
			for(String activity : patternAlphabet){
				if(aANSetMap.containsKey(activity)){
					aANSet = aANSetMap.get(activity);
				}else{
					aANSet = new HashSet<String>();
				}
				aANSet.addAll(pAANSet);
				aANSetMap.put(activity, aANSet);
			}
		}

		processPatternAlphabets();

		String fileSeparator = System.getProperty("file.separator");
		
		XFactory factory = XFactoryRegistry.instance().currentDefault();
		XLog outputLog = factory.createLog();
		
		for (XExtension extension : log.getExtensions())
			outputLog.getExtensions().add(extension);

		XAttributeMap attributeMap = (XAttributeMap)log.getAttributes().clone();
		String parentLogLocation="", parentLogConceptName="";
		
		XAttribute parentFileAttribute = null;
		for(String attribute : attributeMap.keySet()){
			if(attribute.equals("parentFileName")){
				parentFileAttribute = attributeMap.get(attribute);
			}else if(attribute.equals("fileLocation")){
				parentLogLocation = attributeMap.get(attribute).toString();
			}else if(attribute.equals("concept:name")){
				parentLogConceptName = attributeMap.get(attribute).toString();
			}
		}
		
		outputLog.setAttributes((XAttributeMap)attributeMap.clone());
		
		XAttribute fileLocationAttribute = factory.createAttributeLiteral("fileLocation", new File(transformLogFileName).getParent()+fileSeparator,
				XConceptExtension.instance());
		
		XAttribute conceptNameAttribute = factory.createAttributeLiteral("concept:name", new File(transformLogFileName).getName(),
				XConceptExtension.instance());
		
		if(!parentLogLocation.equals("") && !parentLogConceptName.equals("")){
			XAttribute parentFileAbsPathAttribute = factory.createAttributeLiteral("parentFileName", parentLogLocation+parentLogConceptName,XConceptExtension.instance());
			parentFileAbsPathAttribute.getAttributes().put("parentFileName", parentFileAttribute);
			outputLog.getAttributes().put("parentFileName", parentFileAbsPathAttribute);
		}else{
			XAttribute parentFileAbsPathAttribute = factory.createAttributeLiteral("parentFileName", parentLogConceptName, XConceptExtension.instance());
			outputLog.getAttributes().put("parentFileName", parentFileAbsPathAttribute);
		}
			 
		outputLog.getAttributes().put("fileLocation", fileLocationAttribute);
		outputLog.getAttributes().put("concept:name", conceptNameAttribute);
		
		System.out.println("Selected Abs: "+aNPASetMap.keySet());
		Map<String, XLog> abstractionNameSubLogMap = new HashMap<String, XLog>();
		XLog abstractionSubLog;
		for(String abstractionName : aNPASetMap.keySet()){
			abstractionSubLog = factory.createLog();
			abstractionSubLog.setAttributes(log.getAttributes());
			for (XExtension extension : log.getExtensions())
				abstractionSubLog.getExtensions().add(extension);
			abstractionNameSubLogMap.put(abstractionName, abstractionSubLog);
		}
		
		XTrace originalTrace, outputTrace, pATrace;
		XEvent outputEvent;
		
		EquivalenceClass equivalenceClass = new EquivalenceClass();
		Set<String> allPatternSet = new HashSet<String>();
		Set<String> aSFCSet = new HashSet<String>();
		
		for(Set<String> patternAlphabet : patternAlphabetPatternSetMap.keySet()){
			allPatternSet.addAll(patternAlphabetPatternSetMap.get(patternAlphabet));
			aSFCSet.addAll(patternAlphabet);
		}

		int encodingLength = encodedLog.getEncodingLength();
		Map<String, TreeSet<String>> startSymbolPatternSetMap = equivalenceClass.getStartSymbolEquivalenceClassMap(encodingLength, allPatternSet, true);
		TreeSet<String> startSymbolPatternSet;

		int patternLength;
		
		Map<String, Set<String>> patternPatternAlphabetMap = new HashMap<String, Set<String>>();
		Set<String> patternAlphabet;
		for(String pattern : allPatternSet){
			patternAlphabet = new HashSet<String>();
			patternLength = pattern.length()/encodingLength;
			for(int i = 0; i < patternLength; i++)
				patternAlphabet.add(pattern.substring(i*encodingLength, (i+1)*encodingLength));
			patternPatternAlphabetMap.put(pattern, patternAlphabet);
		}
		
		
		List<String> encodedTraceList = encodedLog.getEncodedTraceList();
		int noEncodedTraces = encodedTraceList.size();
		
		int MAX_STRLENGTH = MinePatterns.MAX_STRLENGTH_TRANSFORMLOG;
		
		int encodedTraceLength, currentTraceIndex, currentSubTraceLength, lastPatternMatchedIndex;
		String encodedTrace, currentSymbol, currentSubTrace, pANme, cANe;
		
		Set<String> iEASet = new HashSet<String>();
		iEASet.addAll(encodedLog.getCharActivityMap().keySet());
		iEASet.removeAll(sNIEASet);
		iEASet.removeAll(aSFCSet);
		
		
		Set<String> nonPatternAlphabet = new HashSet<String>();

		Pattern regExPattern;
		Matcher regExMatcher;
		String nonPatternAlphabetPatternStr, nonPatternAlphabetPatternRegEx, matchingSubSequence, tempMatchingSubSequence;
		List<Integer> patternActivityIndexList = new ArrayList<Integer>();
		
		int noMatches, patternStartIndex, patternEndIndex, patternActivityIndex;
		int startJ;
		
		String modifiedTrace, pASubSequence;
		String[] currentSubTraceArray;
		List<String> encodedTraceActivityList = new ArrayList<String>();
		
		for(int i = 0; i < noEncodedTraces; i++){
			originalTrace = (XTrace)log.get(i).clone();
			
			outputTrace = factory.createTrace();
			outputTrace.setAttributes(originalTrace.getAttributes());
			
			for (XExtension extension : originalTrace.getExtensions())
				outputTrace.getExtensions().add(extension);
			
			encodedTrace = encodedTraceList.get(i);
			encodedTraceLength = encodedTrace.length()/encodingLength;
			
			encodedTraceActivityList.clear();
			for(int p = 0; p < encodedTraceLength; p++)
				encodedTraceActivityList.add(encodedTrace.substring(p*encodingLength, (p+1)*encodingLength));
			
			modifiedTrace = "";
			
			pANme = ""; 
			pASubSequence = "";
			pATrace = null;
			
			currentTraceIndex = 0;
			while (currentTraceIndex < encodedTraceLength) {
				if (encodedTraceLength - currentTraceIndex < MAX_STRLENGTH) {
					currentSubTrace = encodedTrace.substring(currentTraceIndex * encodingLength, encodedTraceLength
							* encodingLength);
				} else {
					currentSubTrace = encodedTrace.substring(currentTraceIndex * encodingLength,
							(currentTraceIndex + MAX_STRLENGTH) * encodingLength);
				}
				
				currentSubTraceLength = currentSubTrace.length()/encodingLength;
				
				currentSubTraceArray = new String[currentSubTraceLength];
				for(int p = 0; p < currentSubTraceLength; p++)
					currentSubTraceArray[p] = currentSubTrace.substring(p*encodingLength, (p+1)*encodingLength);
				
				lastPatternMatchedIndex = currentSubTraceLength;
				
				for(int j = 0; j < currentSubTraceLength; j++){
					startJ = j;
					currentSymbol = currentSubTrace.substring(j*encodingLength, (j+1)*encodingLength);
					
					
					if(iEASet.contains(currentSymbol)){
						lastPatternMatchedIndex = j;
						continue;
					}
					
					if(!startSymbolPatternSetMap.containsKey(currentSymbol)){
						if(!pANme.equals("")){
							 if(activitiesInvolvedInAbstractionSet.contains(currentSymbol)){
								if(aANSetMap.containsKey(currentSymbol)){
									aANSet = aANSetMap.get(currentSymbol);
									if(aANSet.contains(pANme)){
										pATrace.add((XEvent)originalTrace.get(currentTraceIndex+j).clone());
										pASubSequence += currentSymbol;
										continue;
									}
								}
							 }
							
							outputEvent = factory.createEvent();
							outputEvent.setAttributes(originalTrace.get(currentTraceIndex+j-1).getAttributes());
							XAttribute name = factory.createAttributeLiteral("concept:name", pANme,
									XConceptExtension.instance());
							outputEvent.getAttributes().put("concept:name", name);
							XAttribute transition = factory.createAttributeLiteral("lifecycle:transition", "complete", XConceptExtension.instance()); 
							outputEvent.getAttributes().put("lifecycle:transition", transition);
							
							outputTrace.add(outputEvent);
							
							if(!abstractionNameSubLogMap.containsKey(pANme)){
								
							}else{
								abstractionSubLog = abstractionNameSubLogMap.get(pANme);
								abstractionSubLog.add(pATrace);
								abstractionNameSubLogMap.put(pANme, abstractionSubLog);
							}
							
							modifiedTrace += pANme;
						}
						
						modifiedTrace += currentSymbol;
						
						outputTrace.add((XEvent)(originalTrace.get(currentTraceIndex+j).clone()));
						lastPatternMatchedIndex = j;
						pANme = "";
						pASubSequence = "";
						continue;
					}
					
					startSymbolPatternSet = startSymbolPatternSetMap.get(currentSymbol);
					
					int maxStartSymbolPatternLength = startSymbolPatternSet.iterator().next().length()/encodingLength;
					if(currentSubTraceLength-j < maxStartSymbolPatternLength && (encodedTraceLength-(currentTraceIndex+j) > maxStartSymbolPatternLength)){
						break;
					}

					for(String pattern : startSymbolPatternSet){
						if(currentSubTrace.indexOf(pattern, j*encodingLength) == j*encodingLength){
							regExPattern = Pattern.compile("(" + pattern + "){1,}");
							regExMatcher = regExPattern.matcher(currentSubTrace);
							patternLength = pattern.length()/encodingLength;
							
							if (regExMatcher.find(j * encodingLength) && regExMatcher.start() == j * encodingLength) {
								noMatches = (regExMatcher.end() - regExMatcher.start()) / pattern.length();
								
								matchingSubSequence = regExMatcher.group();
								
								patternStartIndex = j;
								patternEndIndex = patternStartIndex + noMatches * patternLength;
								
								j += noMatches * patternLength - 1;
								lastPatternMatchedIndex = j;
								
								
								patternAlphabet = patternPatternAlphabetMap.get(pattern);
								
								if(!patternAlphabetAbstractionNameSetMap.containsKey(patternAlphabet)){
	
								}
								
								pAANSet = patternAlphabetAbstractionNameSetMap.get(patternAlphabet);
								
								if(pAANSet.size() == 1){
									cANe = pAANSet.iterator().next();
								}else{
									if(!pANme.equals("") && pAANSet.contains(pANme)){
										cANe = pANme;
									}else{
										cANe = pAANSet.iterator().next();
									}
								}
								
								if(cANe.equals(pANme)){
									for(int p = patternStartIndex; p < patternEndIndex; p++){
										pATrace.add((XEvent)originalTrace.get(currentTraceIndex+p).clone());
									}
									pASubSequence += matchingSubSequence;
								}else{
									if(!pANme.equals("")){
										outputEvent = factory.createEvent();
										outputEvent.setAttributes(originalTrace.get(currentTraceIndex+patternStartIndex-1).getAttributes());
										
										XAttribute name = factory.createAttributeLiteral("concept:name", pANme,
												XConceptExtension.instance());
										outputEvent.getAttributes().put("concept:name", name);
										XAttribute transition = factory.createAttributeLiteral("lifecycle:transition", "complete", XConceptExtension.instance()); 
										outputEvent.getAttributes().put("lifecycle:transition", transition);
										
										outputTrace.add(outputEvent);
										
										if(!abstractionNameSubLogMap.containsKey(pANme)){
	
										}else{
											abstractionSubLog = abstractionNameSubLogMap.get(pANme);
											abstractionSubLog.add(pATrace);
											abstractionNameSubLogMap.put(pANme, abstractionSubLog);
										}
										
										modifiedTrace += pANme;
										
										pANme = "";
										pASubSequence = "";
										
									}

									pANme = cANe;
									pATrace = factory.createTrace();
									pATrace.setAttributes(originalTrace.getAttributes());
									for(int p = patternStartIndex; p < patternEndIndex; p++){
										pATrace.add((XEvent)originalTrace.get(currentTraceIndex+p).clone());
									}
									pASubSequence = matchingSubSequence;
								}
							}
							break;
						}else{
							patternLength = pattern.length()/encodingLength;
							
							nonPatternAlphabet.clear();
							nonPatternAlphabet.addAll(encodedLog.getCharActivityMap().keySet());
							if(patternPatternAlphabetMap.containsKey(pattern)){
								nonPatternAlphabet.removeAll(patternPatternAlphabetMap.get(pattern));
							}
							
							nonPatternAlphabetPatternStr = "(";
							
							Iterator<String> it = nonPatternAlphabet.iterator();
							while(it.hasNext()){
								nonPatternAlphabetPatternStr += "("+it.next()+")";
								if(it.hasNext())
									nonPatternAlphabetPatternStr += "|";
							}
														
							nonPatternAlphabetPatternStr += ")*";
							
							nonPatternAlphabetPatternRegEx = "(";
							for(int rr = 0; rr < patternLength-1; rr++){
								nonPatternAlphabetPatternRegEx += "("+pattern.substring(rr*encodingLength, (rr+1)*encodingLength)+")"+nonPatternAlphabetPatternStr;
							}
							nonPatternAlphabetPatternRegEx += pattern.substring((patternLength-1)*encodingLength, patternLength*encodingLength);
							nonPatternAlphabetPatternRegEx += "){1}";
							
							try{
								regExPattern = Pattern.compile(nonPatternAlphabetPatternRegEx);
								regExMatcher = regExPattern.matcher(currentSubTrace);
							
							if (regExMatcher.find(j * encodingLength) && regExMatcher.start() == j * encodingLength) {
								matchingSubSequence = regExMatcher.group();
							
								tempMatchingSubSequence = matchingSubSequence;
								for(String encodedActivity : nonPatternAlphabet)
									tempMatchingSubSequence = tempMatchingSubSequence.replaceAll(encodedActivity, "");
								
								noMatches = tempMatchingSubSequence.length()/pattern.length();
								
								patternActivityIndexList.clear();
								patternActivityIndex = -1;
								int patternMatchEndIndex = 0;
								String tempSubSequence;
								for(int m = 0; m < noMatches; m++){
									tempSubSequence = matchingSubSequence.substring(patternMatchEndIndex*encodingLength);
									for(int p = 0; p < patternLength; p++){
										patternActivityIndex = tempSubSequence.indexOf(pattern.substring(p*encodingLength, (p+1)*encodingLength),patternMatchEndIndex*encodingLength)/encodingLength;
										
										patternActivityIndexList.add(j+patternActivityIndex);
										patternMatchEndIndex = patternActivityIndex+1;
									}
									patternMatchEndIndex = (patternActivityIndex+1);
								}
								
								String tempSymbol;
								String tempActivityEncodedTrace;
								XEvent tempEvent;
								for(int m = 0; m < noMatches; m++){
									for(int k = 0; k < patternLength; k++){
										tempSymbol = currentSubTraceArray[patternActivityIndexList.get(m*patternLength+k)];
										tempActivityEncodedTrace = encodedTraceActivityList.get(currentTraceIndex+patternActivityIndexList.get(m*patternLength+k));
										tempEvent = originalTrace.get(currentTraceIndex+patternActivityIndexList.get(m*patternLength+k));
										
										for(int p = patternActivityIndexList.get(m*patternLength+k); p > j+k; p--){
											currentSubTraceArray[p] = currentSubTraceArray[p-1];
											encodedTraceActivityList.set(currentTraceIndex+p, encodedTraceActivityList.get(currentTraceIndex+p-1));
											originalTrace.set(currentTraceIndex+p, originalTrace.get(currentTraceIndex+p-1));
										}
										currentSubTraceArray[j+(m*patternLength)+k] = tempSymbol;
										encodedTraceActivityList.set(currentTraceIndex+j+(m*patternLength)+k, tempActivityEncodedTrace);
										originalTrace.set(currentTraceIndex+j+(m*patternLength)+k, tempEvent);
									}
								}
								
								StringBuilder tempStrBuilder = new StringBuilder();
								for(int p = 0; p < currentSubTraceLength; p++)
									tempStrBuilder.append(currentSubTraceArray[p]);
								currentSubTrace = tempStrBuilder.toString();
								
								patternAlphabet = patternPatternAlphabetMap.get(pattern);
								pAANSet = patternAlphabetAbstractionNameSetMap.get(patternAlphabet);
								
								if(individualActivitiesAbstractionSet.contains(patternAlphabet)){
									patternStartIndex = j;
									patternEndIndex = patternStartIndex + patternLength * noMatches;
									for(int p = patternStartIndex; p < patternEndIndex; p++)
										outputTrace.add((XEvent)originalTrace.get(currentTraceIndex+p).clone());
									
									modifiedTrace += tempMatchingSubSequence;
									j += patternLength * noMatches -1;
									lastPatternMatchedIndex = j;
									
									break;
								}
								
								if(pAANSet.size() == 1){
									cANe = pAANSet.iterator().next();
								}else{
									if(!pANme.equals("") && pAANSet.contains(pANme)){
										cANe = pANme;
									}else{
										cANe = pAANSet.iterator().next();
									}
								}

								if(pANme.equals(cANe)){
									patternStartIndex = j;
									patternEndIndex = patternStartIndex + patternLength * noMatches;
									for(int p = patternStartIndex; p < patternEndIndex; p++)
										pATrace.add((XEvent)originalTrace.get(currentTraceIndex+p).clone());
									pASubSequence += tempMatchingSubSequence;
								}else{
									if(!pANme.equals("")){
													
										outputEvent = factory.createEvent();
										outputEvent.setAttributes(originalTrace.get(currentTraceIndex+j-1).getAttributes());
										
										XAttribute name = factory.createAttributeLiteral("concept:name", pANme,
												XConceptExtension.instance());
										outputEvent.getAttributes().put("concept:name", name);
										XAttribute transition = factory.createAttributeLiteral("lifecycle:transition", "complete", XConceptExtension.instance()); 
										outputEvent.getAttributes().put("lifecycle:transition", transition);
										
										outputTrace.add(outputEvent);
										
										if(!abstractionNameSubLogMap.containsKey(pANme)){

										}else{
											abstractionSubLog = abstractionNameSubLogMap.get(pANme);
											abstractionSubLog.add(pATrace);
											abstractionNameSubLogMap.put(pANme, abstractionSubLog);
										}
										
										modifiedTrace += pANme;
										
										pANme = "";
										pASubSequence = "";
									}
									pANme = cANe;
									pASubSequence = tempMatchingSubSequence;
								}
								
								pANme = cANe;
								pATrace = factory.createTrace();
								pATrace.setAttributes(originalTrace.getAttributes());
								patternStartIndex = j;
								patternEndIndex = patternStartIndex + patternLength * noMatches;
								
								for(int p = patternStartIndex; p < patternEndIndex; p++){
									pATrace.add((XEvent)originalTrace.get(currentTraceIndex+p).clone());
								}
								pASubSequence = tempMatchingSubSequence;
								j += patternLength * noMatches-1;
								lastPatternMatchedIndex = j;
								break;
							}else{
								Set<String> currentPatternAlphabet = patternPatternAlphabetMap.get(pattern);
								
								String approximatePatternMatchRegEx="(";
								it = currentPatternAlphabet.iterator();
								while(it.hasNext()){
									approximatePatternMatchRegEx += "("+it.next()+")";
									if(it.hasNext())
										approximatePatternMatchRegEx += "|";
								}
								approximatePatternMatchRegEx += "){"+patternLength+"}";
								
								
								
								regExPattern = Pattern.compile(approximatePatternMatchRegEx);
								regExMatcher = regExPattern.matcher(currentSubTrace);
								if (regExMatcher.find(j * encodingLength) && regExMatcher.start() == j * encodingLength) {
									matchingSubSequence = regExMatcher.group();
									int matchingSubSequenceLength = matchingSubSequence.length()/encodingLength;
									int jcb;
									Set<String> matchingSubSequenceActivitySet = new HashSet<String>();
									for(jcb = 0; jcb < matchingSubSequenceLength; jcb++)
										if(matchingSubSequenceActivitySet.contains(matchingSubSequence.substring(jcb*encodingLength, (jcb+1)*encodingLength)))
											break;
										else
											matchingSubSequenceActivitySet.add(matchingSubSequence.substring(jcb*encodingLength, (jcb+1)*encodingLength));
									if(jcb < matchingSubSequenceLength)
										matchingSubSequence = matchingSubSequence.substring(0,jcb*encodingLength);
									matchingSubSequenceLength = jcb;
										
									if(matchingSubSequenceLength != patternLength)
										continue;
									
									patternStartIndex = j;
									patternEndIndex = patternStartIndex + matchingSubSequenceLength;
									
									j += matchingSubSequenceLength-1;
									lastPatternMatchedIndex = j;
									
									
									patternAlphabet = patternPatternAlphabetMap.get(pattern);
									
									if(!patternAlphabetAbstractionNameSetMap.containsKey(patternAlphabet)){
									
									}
									
									pAANSet = patternAlphabetAbstractionNameSetMap.get(patternAlphabet);
									
									if(pAANSet.size() == 1){
										cANe = pAANSet.iterator().next();
									}else{
										if(!pANme.equals("") && pAANSet.contains(pANme)){
											cANe = pANme;
										}else{
											cANe = pAANSet.iterator().next();
										}
									}
									
									if(cANe.equals(pANme)){
										for(int p = patternStartIndex; p < patternEndIndex; p++){
											pATrace.add((XEvent)originalTrace.get(currentTraceIndex+p).clone());
										}
										pASubSequence += matchingSubSequence;
									}else{
										if(!pANme.equals("")){
											outputEvent = factory.createEvent();
											outputEvent.setAttributes(originalTrace.get(currentTraceIndex+patternStartIndex-1).getAttributes());
											
											XAttribute name = factory.createAttributeLiteral("concept:name", pANme,
													XConceptExtension.instance());
											outputEvent.getAttributes().put("concept:name", name);
											XAttribute transition = factory.createAttributeLiteral("lifecycle:transition", "complete", XConceptExtension.instance()); 
											outputEvent.getAttributes().put("lifecycle:transition", transition);
											
											outputTrace.add(outputEvent);
											
											if(!abstractionNameSubLogMap.containsKey(pANme)){
										
											}else{
												abstractionSubLog = abstractionNameSubLogMap.get(pANme);
												abstractionSubLog.add(pATrace);
												abstractionNameSubLogMap.put(pANme, abstractionSubLog);
											}
											
											modifiedTrace += pANme;

											pANme = "";
											pASubSequence = "";
											
										}

										pANme = cANe;
										pATrace = factory.createTrace();
										pATrace.setAttributes(originalTrace.getAttributes());
										for(int p = patternStartIndex; p < patternEndIndex; p++){
											pATrace.add((XEvent)originalTrace.get(currentTraceIndex+p).clone());
										}
										pASubSequence = matchingSubSequence;
										break;
									}
								}
							}
							}catch(StackOverflowError e){
								JOptionPane.showMessageDialog(new JFrame(), "Stackoverflow: Increase the size of the stack (to say, 8MB) of your JVM using -Xss8M", "Error", JOptionPane.ERROR_MESSAGE);
								return;
							}
						}
					}
					
					if(startJ == j){
						if (!pANme.equals("")){
							 if(activitiesInvolvedInAbstractionSet.contains(currentSymbol)){
								if(aANSetMap.containsKey(currentSymbol)){
									aANSet = aANSetMap.get(currentSymbol);
									if(aANSet.contains(pANme)){
										pATrace.add((XEvent)originalTrace.get(currentTraceIndex+j).clone());
										pASubSequence += currentSymbol;
										continue;
									}
								}
							 }
							outputEvent = factory.createEvent();
							outputEvent.setAttributes(originalTrace.get(currentTraceIndex+j-1).getAttributes());
							XAttribute name = factory.createAttributeLiteral("concept:name", pANme,
									XConceptExtension.instance());
							outputEvent.getAttributes().put("concept:name", name);
							XAttribute transition = factory.createAttributeLiteral("lifecycle:transition", "complete", XConceptExtension.instance()); 
							outputEvent.getAttributes().put("lifecycle:transition", transition);
							
							outputTrace.add(outputEvent);
							
							if(!abstractionNameSubLogMap.containsKey(pANme)){
							
							}else{
								abstractionSubLog = abstractionNameSubLogMap.get(pANme);
								abstractionSubLog.add(pATrace);
								abstractionNameSubLogMap.put(pANme, abstractionSubLog);
							}
							
							modifiedTrace += pANme;
						}
						
						modifiedTrace += currentSymbol;
						
						outputTrace.add((XEvent)(originalTrace.get(currentTraceIndex+j).clone()));
						lastPatternMatchedIndex = j;
						pANme = "";
						pASubSequence = "";
						continue;
					}
				}
				
				StringBuilder tempStrBuilder = new StringBuilder();
				for(int p = 0; p < encodedTraceLength; p++)
					tempStrBuilder.append(encodedTraceActivityList.get(p));
				encodedTrace = tempStrBuilder.toString();
				
				if(encodedTraceLength-currentTraceIndex < MAX_STRLENGTH)
					currentTraceIndex = encodedTraceLength + 1;
				else
					currentTraceIndex += lastPatternMatchedIndex + 1;
			}
			
			if(!pANme.equals("")){
				outputEvent = factory.createEvent();
				outputEvent.setAttributes(originalTrace.get(encodedTraceLength-1).getAttributes());
				
				XAttribute name = factory.createAttributeLiteral("concept:name", pANme,
						XConceptExtension.instance());
				outputEvent.getAttributes().put("concept:name", name);
				XAttribute transition = factory.createAttributeLiteral("lifecycle:transition", "complete", XConceptExtension.instance()); 
				outputEvent.getAttributes().put("lifecycle:transition", transition);
				
				outputTrace.add(outputEvent);
				
				if(!abstractionNameSubLogMap.containsKey(pANme)){

				}else{
					abstractionSubLog = abstractionNameSubLogMap.get(pANme);
					abstractionSubLog.add(pATrace);
					abstractionNameSubLogMap.put(pANme, abstractionSubLog);
				}
				
				modifiedTrace += pANme;
			}
			outputLog.add(outputTrace);
		}
		
		try {
			FileOutputStream out = new FileOutputStream(transformLogFileName);
			XesXmlGZIPSerializer xesGzipSerializer = new XesXmlGZIPSerializer();
			xesGzipSerializer.serialize(outputLog, out);
			out.close();
		} catch (FileNotFoundException e) {
			JOptionPane.showMessageDialog(new JFrame(), e.getMessage());
		} catch (IOException e) {
			JOptionPane.showMessageDialog(new JFrame(), e.getMessage());
		}
		
		String sublogDirectory = transformLogFileName.substring(0, transformLogFileName.indexOf("."));
		if(new File(sublogDirectory).exists()){
			deleteDirectory(new File(sublogDirectory));
		}
		new File(sublogDirectory).mkdirs();
		String validFileName, validAbsPathFileName;
		FileOutputStream out;
		XesXmlGZIPSerializer xesGzipSerializer;
		for(String abstraction : abstractionNameSubLogMap.keySet()){
			validFileName = getValidFileName(abstraction);
			validAbsPathFileName = sublogDirectory + fileSeparator + validFileName + ".xes.gz";
			abstractionSubLog = abstractionNameSubLogMap.get(abstraction);

			try{
				if(!abstractionSubLog.isEmpty()){
					out = new FileOutputStream(validAbsPathFileName);
					xesGzipSerializer = new XesXmlGZIPSerializer();
					xesGzipSerializer.serialize(abstractionSubLog, out);
					out.close();
				}
			}catch (FileNotFoundException e) {
				JOptionPane.showMessageDialog(new JFrame(), e.getMessage());
			} catch (IOException e) {
				JOptionPane.showMessageDialog(new JFrame(), e.getMessage());
			}
		}
	}
	
	private void processPatternAlphabets(){
		aNPASetMap = new HashMap<String, Set<Set<String>>>();
		Set<Set<String>> abstractionNamePatternAlphabetSet;
		Set<String> patternAlphabetAbstractionNameSet;
		
		for(Set<String> patternAlphabet : pAANSetMap.keySet()){
			patternAlphabetAbstractionNameSet = pAANSetMap.get(patternAlphabet);
			for(String abstractionName : patternAlphabetAbstractionNameSet){
				if(aNPASetMap.containsKey(abstractionName)){
					abstractionNamePatternAlphabetSet = aNPASetMap.get(abstractionName);
				}else{
					abstractionNamePatternAlphabetSet = new HashSet<Set<String>>();
				}
				abstractionNamePatternAlphabetSet.add(patternAlphabet);
				aNPASetMap.put(abstractionName, abstractionNamePatternAlphabetSet);
			}
		}
	
		individualActivitiesAbstractionSet = new HashSet<Set<String>>();
		for(String abstractionName : aNPASetMap.keySet()){
			abstractionNamePatternAlphabetSet = aNPASetMap.get(abstractionName);
			if(abstractionNamePatternAlphabetSet.size() == 1){
				for(Set<String> patternAlphabet : abstractionNamePatternAlphabetSet)
					if(patternAlphabet.size() == 1)
						individualActivitiesAbstractionSet.add(patternAlphabet);
			}
		}

	}

	private boolean deleteDirectory(File path) {
		if (path.exists()) {
			File[] files = path.listFiles();
			for (int i = 0; i < files.length; i++) {
				if (files[i].isDirectory()) {
					deleteDirectory(files[i]);
				} else {
					files[i].delete();
				}
			}
		}
		return (path.delete());
	}
	
	private String getValidFileName(String fileName){
		String [] invalidCharInFileName = {"\\", "/", ":", "?","\"","<",">","|"};
		for(String invalidChar : invalidCharInFileName)
			fileName = fileName.replace(invalidChar, " ");
		return fileName;
	}
}
	
	

