package org.processmining.plugins.abstractions.util;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import org.jgrapht.alg.DijkstraShortestPath;
import org.jgrapht.graph.DefaultEdge;
import org.jgrapht.graph.SimpleDirectedGraph;

/**
 * 
 * @author R. P. Jagadeesh Chandra 'JC' Bose
 * @date 27 May 2010 (last modified)
 * @since 01 July 2009
 * @version 1.0 
 * @(c)R. P. Jagadeesh Chandra 'JC' Bose, Department of Mathematics
 *          and Computer Science, TU/e
 * 
 * This class computes the Hasse diagram for a given set of pattern alphabets
 * 		
 * 		 Reproducing or using whole or part of this code/application is to be done with proper
 *       acknowledgments by citing "R. P. Jagadeesh Chandra Bose and Wil M. P.
 *       van der Aalst. Abstractions in Process Mining: A Taxonomy of Patterns.
 *       Proceedings of the International Conference on Business Process
 *       Management (BPM). Volume 5701 of LNCS., Springer-Verlag (2009) pages
 *       159-175"
 */

public class AbstractionSetTheory {
	SimpleDirectedGraph<Set<String>, DefaultEdge> sg;
	Collection<Set<String>> patternAlphabetSet;
	Map<Integer, Set<Set<String>>> sizeVertexPatternAlphabetSetMap;
	Map<Set<String>, Set<Set<String>>> patternAlphabetMaximalElementSetMap;
	Map<Set<String>, Set<Set<String>>> maximalElementSubsumedPatternAlphabetSetMap;
	
	/*
	 * Store the sizes of the alphabet; It would be required when constructing
	 * the Hasse diagram; We try to put a link between two alphabets only if
	 * there is no path between them
	 */
	int[] sortedSizeArray;
	
	public AbstractionSetTheory(Collection<Set<String>> patternAlphabetSet){
		this.patternAlphabetSet = patternAlphabetSet;
		sg = new SimpleDirectedGraph<Set<String>, DefaultEdge>(DefaultEdge.class);
		for(Set<String> patternAlphabet : patternAlphabetSet)
			sg.addVertex(patternAlphabet);
		
		prepareSizeVertexPatternAlphabetSetMap();
		
		buildLayeredGraph();
//		printGraph();
	}
	
	private void prepareSizeVertexPatternAlphabetSetMap(){
		if(sizeVertexPatternAlphabetSetMap == null){
			sizeVertexPatternAlphabetSetMap = new HashMap<Integer, Set<Set<String>>>();
		}else{
			sizeVertexPatternAlphabetSetMap.clear();
		}
		
		int vertexPatternAlphabetSize;
		Set<Set<String>> sizeVertexPatternAlphabetSet;
		
		Set<Set<String>> vertexPatternAlphabetSet = sg.vertexSet();
		for(Set<String> vertexPatternAlphabet : vertexPatternAlphabetSet){
			vertexPatternAlphabetSize = vertexPatternAlphabet.size();
			if(sizeVertexPatternAlphabetSetMap.containsKey(vertexPatternAlphabetSize)){
				sizeVertexPatternAlphabetSet = sizeVertexPatternAlphabetSetMap.get(vertexPatternAlphabetSize);
			}else{
				sizeVertexPatternAlphabetSet = new HashSet<Set<String>>();
			}
			sizeVertexPatternAlphabetSet.add(vertexPatternAlphabet);
			sizeVertexPatternAlphabetSetMap.put(vertexPatternAlphabetSize, sizeVertexPatternAlphabetSet);
		}
		
		sortedSizeArray = new int[sizeVertexPatternAlphabetSetMap.size()];
		int index = 0;
		for(Integer size : sizeVertexPatternAlphabetSetMap.keySet())
			sortedSizeArray[index++] = size;
		Arrays.sort(sortedSizeArray);
	}

	protected void buildLayeredGraph(){
		int maxStepSize = sortedSizeArray.length-1;
		int lowerLayerPatternAlphabetSize, upperLayerPatternAlphabetSize;
		
		Set<Set<String>> lowerLayerPatternAlphabetSet, upperLayerPatternAlphabetSet;
		for(int step = 1; step <= maxStepSize; step++){
			for(int i = 0; i < sortedSizeArray.length-step; i++){
				lowerLayerPatternAlphabetSize = sortedSizeArray[i];
				upperLayerPatternAlphabetSize = sortedSizeArray[i+step];
				
				lowerLayerPatternAlphabetSet = sizeVertexPatternAlphabetSetMap.get(lowerLayerPatternAlphabetSize);
				upperLayerPatternAlphabetSet = sizeVertexPatternAlphabetSetMap.get(upperLayerPatternAlphabetSize);
				
				for(Set<String> patternAlphabetI : lowerLayerPatternAlphabetSet){
					for(Set<String> patternAlphabetJ : upperLayerPatternAlphabetSet){
						if (patternAlphabetJ.containsAll(patternAlphabetI)) {
							DijkstraShortestPath<Set<String>, DefaultEdge> d = new DijkstraShortestPath<Set<String>, DefaultEdge>(
									sg, patternAlphabetI, patternAlphabetJ);
							if (d.getPathEdgeList() == null)
								sg.addEdge(patternAlphabetI, patternAlphabetJ);
						}
					}
				}
			}
		}
	}
	
	public void extendMaximalElements(int minMaximalElementSizeThreshold, float commonElementsThreshold){
		List<Set<String>> maximalElementList = getMaximalElements();
		int noMaximalElements = maximalElementList.size();
		Set<String> extendedMaximalElement;
		Set<Set<String>> extendedMaximalElementSet = new HashSet<Set<String>>();
		float percentageCommonElements;
		Set<String> maximalElementI = new HashSet<String>(), maximalElementJ = new HashSet<String>();
		Set<String> commonActivitySet = new HashSet<String>();
		
		for(int i = 0; i < noMaximalElements-1; i++){
			if(maximalElementList.get(i).size() <= minMaximalElementSizeThreshold)
				continue;
			for(int j = i+1; j < noMaximalElements; j++){
				if(maximalElementList.get(j).size() <= minMaximalElementSizeThreshold)
					continue;
				maximalElementI.clear();
				maximalElementI.addAll(maximalElementList.get(i));
				
				maximalElementJ.clear();
				maximalElementJ.addAll(maximalElementList.get(j));
		
				commonActivitySet.clear();
				commonActivitySet.addAll(maximalElementI);
				commonActivitySet.retainAll(maximalElementJ);
				
				percentageCommonElements = commonActivitySet.size()*100.0f/Math.min(maximalElementI.size(), maximalElementJ.size());
				
				if(percentageCommonElements >= commonElementsThreshold){
					extendedMaximalElement = new TreeSet<String>();
					extendedMaximalElement.addAll(maximalElementI);
					extendedMaximalElement.addAll(maximalElementJ);
					extendedMaximalElementSet.add(extendedMaximalElement);
					//	Adding Extended Maximal Element: extendedMaximalElement
					sg.addVertex(extendedMaximalElement);
					sg.addEdge(maximalElementJ, extendedMaximalElement);
					sg.addEdge(maximalElementI, extendedMaximalElement);
				}
			}
		}
	}
	
	public List<Set<String>> getMaximalElements(){
		List<Set<String>> maximalElementList = new ArrayList<Set<String>>();
		
		for(Set<String> alphabet : sg.vertexSet()){
			if(sg.outDegreeOf(alphabet) == 0)
				maximalElementList.add(alphabet);
		}
		return maximalElementList;
	}	

	public Map<Set<String>, Set<Set<String>>> findAlphabetMaximalElementSetMap(){
		Map<Set<String>, Set<Set<String>>> alphabetMaximalElementSetMap = new HashMap<Set<String>, Set<Set<String>>>();
		List<Set<String>> maximalElementList = getMaximalElements();
		
		Set<Set<String>> alphabetMaximalElementSet;
		for(Set<String> alphabet : sg.vertexSet()){
			alphabetMaximalElementSet = new HashSet<Set<String>>();
			for(Set<String> maximalElement : maximalElementList){
				if(alphabet.equals(maximalElement))
					continue;
				
				DijkstraShortestPath<Set<String>, DefaultEdge> shortestPath = new DijkstraShortestPath<Set<String>, DefaultEdge>(
						sg, alphabet, maximalElement);
				if(shortestPath.getPathEdgeList() != null){
					alphabetMaximalElementSet.add(maximalElement);
				}
			}
			if(alphabetMaximalElementSet.size() > 0){
				alphabetMaximalElementSetMap.put(alphabet, alphabetMaximalElementSet);
			}
		}
		return alphabetMaximalElementSetMap;
	}
	
	
	public Set<Set<String>> getSubsumedPatternAlphabetSet(Set<Set<String>> maximalElementSet){
		Set<Set<String>> subsumedPatternAlphabetSet = new HashSet<Set<String>>();
		
		for(Set<String> alphabet : sg.vertexSet()){
			for(Set<String> maximalElement : maximalElementSet){
				if(alphabet.equals(maximalElement))
					continue;
				DijkstraShortestPath<Set<String>, DefaultEdge> shortestPath = new DijkstraShortestPath<Set<String>, DefaultEdge>(
						sg, alphabet, maximalElement);
				if(shortestPath.getPathEdgeList() != null){
					subsumedPatternAlphabetSet.add(alphabet);
				}
			}
		}
		return subsumedPatternAlphabetSet;
	}
	
	public Set<Set<String>> getUniquelySubsumedPatternAlphabetSet(Set<Set<String>> maximalElementSet){
		Set<Set<String>> uniquelySubsumedPatternAlphabetSet = new HashSet<Set<String>>();
		Map<Set<String>, Set<Set<String>>> alphabetAllMaximalElementSetMap = findAlphabetMaximalElementSetMap();
		//First get all subsumed alphabets in the given set of maximal elements

		for(Set<String> alphabet : alphabetAllMaximalElementSetMap.keySet()){
			if(maximalElementSet.containsAll(alphabetAllMaximalElementSetMap.get(alphabet))){
				uniquelySubsumedPatternAlphabetSet.add(alphabet);
			}
		}
		return uniquelySubsumedPatternAlphabetSet;
	}
	
	public void preparePatternAlphabetMaximalElementRelations(){
		List<Set<String>> maximalElementList = getMaximalElements();
		patternAlphabetMaximalElementSetMap = new HashMap<Set<String>, Set<Set<String>>>();
		maximalElementSubsumedPatternAlphabetSetMap = new HashMap<Set<String>, Set<Set<String>>>();
		
		Set<Set<String>> alphabetMaximalElementSet;
		Set<Set<String>> maximalElementSubsumedAlphabetSet;
		boolean maximalElementSubsumesAlphabet;
		
		for(Set<String> alphabet : patternAlphabetSet){
			for(Set<String> maximalElement : maximalElementList){
				maximalElementSubsumesAlphabet = false;
				if (!alphabet.equals(maximalElement)) {
					DijkstraShortestPath<Set<String>, DefaultEdge> d = new DijkstraShortestPath<Set<String>, DefaultEdge>(
							sg, alphabet, maximalElement);
					if (d.getPathEdgeList() != null) {
						if (patternAlphabetMaximalElementSetMap.containsKey(alphabet))
							alphabetMaximalElementSet = patternAlphabetMaximalElementSetMap.get(alphabet);
						else
							alphabetMaximalElementSet = new HashSet<Set<String>>();
						alphabetMaximalElementSet.add(maximalElement);
						patternAlphabetMaximalElementSetMap.put(alphabet, alphabetMaximalElementSet);
						
						maximalElementSubsumesAlphabet = true;
					}
				}else{
					alphabetMaximalElementSet = new HashSet<Set<String>>();
					alphabetMaximalElementSet.add(maximalElement);
					patternAlphabetMaximalElementSetMap.put(alphabet, alphabetMaximalElementSet);
					maximalElementSubsumesAlphabet = true;
				}
				
				if(maximalElementSubsumesAlphabet){
					if(maximalElementSubsumedPatternAlphabetSetMap.containsKey(maximalElement))
						maximalElementSubsumedAlphabetSet = maximalElementSubsumedPatternAlphabetSetMap.get(maximalElement);
					else
						maximalElementSubsumedAlphabetSet = new HashSet<Set<String>>();
					
					maximalElementSubsumedAlphabetSet.add(alphabet);
					maximalElementSubsumedPatternAlphabetSetMap.put(maximalElement, maximalElementSubsumedAlphabetSet);
				}
			}
		}
//		return maximalElementList;
	}
	
	
	public void printGraph(){
		System.out.println("AlphabetSet for Abstration");
		/*
		 * Print the alphabetSet First
		 */
		for(Integer size : sizeVertexPatternAlphabetSetMap.keySet()){
			System.out.println("Size "+size+": "+sizeVertexPatternAlphabetSetMap.get(size).size()+" : "+sizeVertexPatternAlphabetSetMap.get(size));
		}
		
		System.out.println("\nPrinting Graph\n");
		Set<Set<String>> sizeAlphabetSet;
		for(Integer size : sizeVertexPatternAlphabetSetMap.keySet()){
			System.out.println(size);
			System.out.println("----------------");
			sizeAlphabetSet = sizeVertexPatternAlphabetSetMap.get(size);
			for(Set<String> alphabet : sizeAlphabetSet){
				System.out.println(alphabet+" Outdegree: "+sg.outDegreeOf(alphabet));
				
				Set<DefaultEdge> outgoingEdgeSet = sg.outgoingEdgesOf(alphabet);
				for(DefaultEdge edge : outgoingEdgeSet)
					System.out.println("\t"+sg.getEdgeTarget(edge));
			}
		}
		
		System.out.println("Maximal Elements");
		List<Set<String>> maximalElementList = getMaximalElements();
		for(Set<String> maximalElement : maximalElementList)
			System.out.println(maximalElement);
	}

	public Map<Set<String>, Set<Set<String>>> getAlphabetMaximalElementSetMap() {
		return patternAlphabetMaximalElementSetMap;
	}

	public Map<Set<String>, Set<Set<String>>> getMaximalElementSubsumedAlphabetSetMap() {
		return maximalElementSubsumedPatternAlphabetSetMap;
	}
	
	public Set<Set<String>> getSubsumedAlphabetSet(Set<String> maximalElement){
		if(maximalElementSubsumedPatternAlphabetSetMap.containsKey(maximalElement))
			return maximalElementSubsumedPatternAlphabetSetMap.get(maximalElement);
		else
			return null;
	}
	
	public Set<Set<String>> getVertexSet(){
		return sg.vertexSet();
	}
}
