package org.processmining.plugins.abstractions.patterns;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import javax.swing.JOptionPane;

import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XLogImpl;
import org.processmining.plugins.abstractions.datatypes.ConservednessMetric;
import org.processmining.plugins.abstractions.datatypes.EncodedLog;
import org.processmining.plugins.abstractions.datatypes.MetricPatternAlphabetCount;
import org.processmining.plugins.abstractions.encoding.EncodeActivitySet;
import org.processmining.plugins.abstractions.encoding.EncodeTraces;
import org.processmining.plugins.abstractions.enumtypes.PatternLengthPreference;
import org.processmining.plugins.abstractions.enumtypes.PatternType;
import org.processmining.plugins.abstractions.exceptions.ActivityOverFlowException;
import org.processmining.plugins.abstractions.exceptions.EncodingNotFoundException;
import org.processmining.plugins.abstractions.util.EquivalenceClass;
import org.processmining.plugins.abstractions.util.FileIO;
import org.processmining.plugins.abstractions.util.LogUtils;
import org.processmining.plugins.abstractions.util.Logger;
import org.processmining.plugins.abstractions.util.UkkonenSuffixTree;
import org.processmining.plugins.abstractions.util.Watch;
import org.processmining.plugins.abstractions.visualization.PatternAbstractionFrame;

/**
 * 
 * @author R. P. Jagadeesh Chandra 'JC' Bose
 * @date 15 May 2011 Eindhoven University of Technology (TU/e)
 * 
 *       Reproducing or using whole or part of this code/application is to be done with proper
 *       acknowledgments by citing "R. P. Jagadeesh Chandra Bose and Wil M. P.
 *       van der Aalst. Abstractions in Process Mining: A Taxonomy of Patterns.
 *       Proceedings of the International Conference on Business Process
 *       Management (BPM). Volume 5701 of LNCS., Springer-Verlag (2009) pages
 *       159-175"
 * 
 */
public class MinePatterns {
	XLog log;
	
	EncodedLog encodedLog;
	EncodedLog encodedLogWithTraceDelimiters;
	
	PatternAbstractionFrame owner;
	public static int MAX_STRLENGTH = 350000;
	public static int MAX_STRLENGTH_TRANSFORMLOG = 750;
	
	public static int INSTANCECOUNT_PERCENTAGE_THRESHOLD = 0;
	public static int OVERLAPPINGCOUNT_CONSERVEDNESS_THRESHOLD = 0;
	
	String[] splitTraces;

	
	boolean debug = false;
	
	Map<TreeSet<String>, TreeSet<String>> entireLogTandemArrayPatternAlphabetPatternSetMap = new HashMap<TreeSet<String>, TreeSet<String>>();
	Map<TreeSet<String>, TreeSet<String>> entireLogMaximalRepeatPatternAlphabetPatternSetMap = new HashMap<TreeSet<String>, TreeSet<String>>();
	Map<TreeSet<String>, TreeSet<String>> entireLogCombinedPatternAlphabetPatternSetMap = new HashMap<TreeSet<String>, TreeSet<String>>();
	
	Map<TreeSet<String>, TreeSet<String>> entireLogTandemArrayBasePatternAlphabetPatternSetMap = new HashMap<TreeSet<String>, TreeSet<String>>();
	Map<TreeSet<String>, TreeSet<String>> entireLogMaximalRepeatBasePatternAlphabetPatternSetMap = new HashMap<TreeSet<String>, TreeSet<String>>();
	Map<TreeSet<String>, TreeSet<String>> entireLogCombinedBasePatternAlphabetPatternSetMap = new HashMap<TreeSet<String>, TreeSet<String>>();

	MetricPatternAlphabetCount overlappingPatternAlphabetCount;
	MetricPatternAlphabetCount nonOverlappingPatternAlphabetCount;
	MetricPatternAlphabetCount nonOverlappingShorterPatternAlphabetCount;
	MetricPatternAlphabetCount nonOverlappingLongerPatternAlphabetCount;
	MetricPatternAlphabetCount nonOverlappingGlobalShorterPatternAlphabetCount;
	MetricPatternAlphabetCount nonOverlappingGlobalLongerPatternAlphabetCount;

	Set<Set<String>> ignorablePatternAlphabetSet;
	
	Map<Set<String>, Float> patternAlphabetMeanCountMap;
	Map<Set<String>, Float> patternAlphabetStdDevCountMap;
	
	Map<String, Integer> activityCountMap;
	Map<String, Integer> nonDuplicateActivityCountMap;
	
	PatternLengthPreference patternLengthPreference;
	ConservednessMetric conservednessMetric;
	String outputDir;
	String delim = "@";
	
	
	public MinePatterns(PatternAbstractionFrame owner, XLog log){
		String fileSeparator = System.getProperty("file.separator");
		outputDir = System.getProperty("java.io.tmpdir")+fileSeparator+"Abstractions"+fileSeparator+"MinePatterns";
		Logger.startLog(outputDir, "MinePatterns.txt");
		
		this.owner = owner;
		this.log = log;
		encodeLog();
	}
	
	/**
	 * This method encodes the given log into character streams
	 */
	private void encodeLog() {
		/*
		 * traceIndex is used to keep track of the number of traces;
		 * totalNoEvents counts the number of events/activities in the entire
		 * event log. activitySet accumulates the set of distinct
		 * activities/events in the event log; it doesn't store the trace
		 * identifier for encoding; Encoding trace identifier is required only
		 * when any of the maximal repeat (alphabet) features is selected
		 */
		System.out.println("Encoding Log");
		Watch w = new Watch();
		w.start();
		
		int traceIndex = 0;
		int maximumTraceLength = 0;
		long totalNoEvents = 0;
		Set<String> activitySet = new HashSet<String>();
		Set<String> eventTypeSet = new HashSet<String>();
		String activity;
		
		for (XTrace trace : log) {
			totalNoEvents += trace.size();
			for (XEvent event : trace) {
				activitySet.add(LogUtils.getActivity(event));
				eventTypeSet.add(LogUtils.getEventType(event));
			}
			if(trace.size() > maximumTraceLength)
				maximumTraceLength = trace.size();
		}

		long averageTraceLength = totalNoEvents/log.size();
		try {
			EncodeActivitySet encodeActivitySet = new EncodeActivitySet(activitySet);
			int encodingLength = encodeActivitySet.getEncodingLength();

			Map<String, String> activityCharMap = encodeActivitySet.getActivityCharMap();
			Map<String, String> charActivityMap = encodeActivitySet.getCharActivityMap();
			int maximumActivityLength = encodeActivitySet.getMaximumActivityLength();
			/*
			 * Encode each trace to a charStream
			 */
			EncodeTraces encodeTraces = new EncodeTraces(activityCharMap, log);
			List<String> encodedTraceList = encodeTraces.getCharStreamList();

			List<String> uniqueEncodedTraceList = new ArrayList<String>();

			traceIndex = 0;
			Map<String, TreeSet<Integer>> encodedTraceIdenticalIndicesSetMap = new HashMap<String, TreeSet<Integer>>();
			TreeSet<Integer> encodedTraceIdenticalIndicesSet;
			
			activityCountMap = new HashMap<String, Integer>();
			nonDuplicateActivityCountMap = new HashMap<String, Integer>();
			int noIdenticalTraces, encodedTraceLength;
			int activityCount;
			
			if(encodedTraceList.size() >= 1){
				for (String encodedTrace : encodedTraceList) {
					if (encodedTraceIdenticalIndicesSetMap.containsKey(encodedTrace)) {
						encodedTraceIdenticalIndicesSet = encodedTraceIdenticalIndicesSetMap.get(encodedTrace);
					} else {
						encodedTraceIdenticalIndicesSet = new TreeSet<Integer>();
						uniqueEncodedTraceList.add(encodedTrace);
					}
					encodedTraceIdenticalIndicesSet.add(traceIndex);
					encodedTraceIdenticalIndicesSetMap.put(encodedTrace, encodedTraceIdenticalIndicesSet);
	
					traceIndex++;
				}
				
				for(String encodedTrace : uniqueEncodedTraceList){
					encodedTraceLength = encodedTrace.length()/encodingLength;
					noIdenticalTraces = encodedTraceIdenticalIndicesSetMap.get(encodedTrace).size();
					
					for(int i = 0; i < encodedTraceLength; i++){
						activity = encodedTrace.substring(i*encodingLength, (i+1)*encodingLength);
						activityCount = noIdenticalTraces;
						if(activityCountMap.containsKey(activity))
							activityCount += activityCountMap.get(activity);
						activityCountMap.put(activity, activityCount);
						
						activityCount = 1;
						if(nonDuplicateActivityCountMap.containsKey(activity))
							activityCount += nonDuplicateActivityCountMap.get(activity);
						nonDuplicateActivityCountMap.put(activity, activityCount);
					}
				}
			}else{
				for(String encodedTrace : encodedTraceList){
					encodedTraceLength = encodedTrace.length()/encodingLength;
					for(int i = 0; i < encodedTraceLength; i++){
						activity = encodedTrace.substring(i*encodingLength, (i+1)*encodingLength);
						activityCount = 1;
						if(activityCountMap.containsKey(activity))
							activityCount += activityCountMap.get(activity);
						activityCountMap.put(activity, activityCount);
					}
					nonDuplicateActivityCountMap.putAll(activityCountMap);
				}
			}
	
			encodedLog = new EncodedLog();
			encodedLog.setEncodingLength(encodingLength);
			encodedLog.setActivityCharMap(activityCharMap);
			encodedLog.setCharActivityMap(charActivityMap);
			encodedLog.setEncodedTraceList(encodedTraceList);
			encodedLog.setUniqueEncodedTraceList(uniqueEncodedTraceList);
			encodedLog.setAverageTraceLength(averageTraceLength);
			encodedLog.setTotalTraceLength(totalNoEvents);
			encodedLog.setMaximumActivityNameLength(maximumActivityLength);
			encodedLog.setMaximumTraceLength(maximumTraceLength);
			encodedLog.setEncodedTraceIdenticalTraceIndicesSetMap(encodedTraceIdenticalIndicesSetMap);
			encodedLog.print();
			encodedLog.printLog();
			FileIO io = new FileIO();
			io.writeToFile(outputDir, "ActivityCountMap.txt", activityCountMap, delim);
			io.writeToFile(outputDir, "NonDuplicateActivityCountMap.txt", nonDuplicateActivityCountMap, delim);
			System.out.println("Took "+w.msecs()+" msecs to encode log");
		} catch (ActivityOverFlowException e) {
			JOptionPane.showMessageDialog(owner, "Exception while encoding log", e.getMessage(), JOptionPane.ERROR_MESSAGE);
		} catch (EncodingNotFoundException e) {
			JOptionPane.showMessageDialog(owner, "Exception while encoding log", e.getMessage(), JOptionPane.ERROR_MESSAGE);
		}
	}

	private void encodeLogWithTraceDelimiters() {
		/*
		 * traceIndex is used to keep track of the number of traces;
		 * totalNoEvents counts the number of events/activities in the entire
		 * event log. activitySet accumulates the set of distinct
		 * activities/events in the event log; it doesn't store the trace
		 * identifier for encoding; Encoding trace identifier is required only
		 * when any of the maximal repeat (alphabet) features is selected
		 */
		System.out.println("Encoding Log With Trace Delimiters");
		Watch w = new Watch();
		w.start();
		
		int traceIndex = 0;
		int maximumTraceLength = 0;
		long totalNoEvents = 0;
		Set<String> activitySet = new HashSet<String>();
		Set<String> eventTypeSet = new HashSet<String>();

		for (XTrace trace : log) {
			totalNoEvents += trace.size();
			for (XEvent event : trace) {
				activitySet.add(LogUtils.getActivity(event));
				eventTypeSet.add(LogUtils.getEventType(event));
			}
			activitySet.add("Trace: "+traceIndex);
			if(trace.size() > maximumTraceLength)
				maximumTraceLength = trace.size();
			traceIndex++;
		}

		long averageTraceLength = totalNoEvents/log.size();
		try {
			EncodeActivitySet encodeActivitySet = new EncodeActivitySet(activitySet);
			int encodingLength = encodeActivitySet.getEncodingLength();

			Map<String, String> activityCharMap = encodeActivitySet.getActivityCharMap();
			Map<String, String> charActivityMap = encodeActivitySet.getCharActivityMap();
			int maximumActivityLength = encodeActivitySet.getMaximumActivityLength();
			/*
			 * Encode each trace to a charStream
			 */
			EncodeTraces encodeTraces = new EncodeTraces(activityCharMap, log);
			List<String> encodedTraceList = encodeTraces.getCharStreamList();

			List<String> uniqueEncodedTraceList = new ArrayList<String>();

			traceIndex = 0;
			Map<String, TreeSet<Integer>> encodedTraceIdenticalIndicesSetMap = new HashMap<String, TreeSet<Integer>>();
			TreeSet<Integer> encodedTraceIdenticalIndicesSet;
			
			if(encodedTraceList.size() >= 1){
				for (String encodedTrace : encodedTraceList) {
					if (encodedTraceIdenticalIndicesSetMap.containsKey(encodedTrace)) {
						encodedTraceIdenticalIndicesSet = encodedTraceIdenticalIndicesSetMap.get(encodedTrace);
					} else {
						encodedTraceIdenticalIndicesSet = new TreeSet<Integer>();
						uniqueEncodedTraceList.add(encodedTrace);
					}
					encodedTraceIdenticalIndicesSet.add(traceIndex);
					encodedTraceIdenticalIndicesSetMap.put(encodedTrace, encodedTraceIdenticalIndicesSet);
	
					traceIndex++;
				}
			}
			
			encodedLogWithTraceDelimiters = new EncodedLog();
			encodedLogWithTraceDelimiters.setEncodingLength(encodingLength);
			encodedLogWithTraceDelimiters.setActivityCharMap(activityCharMap);
			encodedLogWithTraceDelimiters.setCharActivityMap(charActivityMap);
			encodedLogWithTraceDelimiters.setEncodedTraceList(encodedTraceList);
			encodedLogWithTraceDelimiters.setUniqueEncodedTraceList(uniqueEncodedTraceList);
			encodedLogWithTraceDelimiters.setAverageTraceLength(averageTraceLength);
			encodedLogWithTraceDelimiters.setMaximumActivityNameLength(maximumActivityLength);
			encodedLogWithTraceDelimiters.setMaximumTraceLength(maximumTraceLength);
			encodedLogWithTraceDelimiters.setEncodedTraceIdenticalTraceIndicesSetMap(encodedTraceIdenticalIndicesSetMap);
			encodedLogWithTraceDelimiters.print();
			System.out.println("Took "+w.msecs()+" msecs to encode log with delimiters");
		} catch (ActivityOverFlowException e) {
			JOptionPane.showMessageDialog(owner, "Exception while encoding log", e.getMessage(), JOptionPane.ERROR_MESSAGE);
		} catch (EncodingNotFoundException e) {
			JOptionPane.showMessageDialog(owner, "Exception while encoding log", e.getMessage(), JOptionPane.ERROR_MESSAGE);
		}
	}
	
	public void findPatterns(Set<PatternType> patternTypeSet, PatternLengthPreference patternLengthPreference){
		this.patternLengthPreference = patternLengthPreference;
		for(PatternType patternType : patternTypeSet){
			switch(patternType){
			case TandemArrays:
				findTandemArrays();
				findBaseTandemArrays();
				break;
			case MaximalRepeats:
				findMaximalRepeats();
				findBaseMaximalRepeats();
				break;
			}
		}
		
		computeCombinedPatterns();
		computePatternMetrics();
	}
	
	private void findTandemArrays(){
		System.out.println("In find Tandem Arrays");
		Watch w = new Watch();
		w.start();
		int encodingLength = encodedLog.getEncodingLength();
		int encodedTraceLength;
		List<String> uniqueEncodedTraceList = encodedLog.getUniqueEncodedTraceList();
		UkkonenSuffixTree suffixTree; 
		
		entireLogTandemArrayPatternAlphabetPatternSetMap.clear();
		Map<TreeSet<String>, TreeSet<String>> currentTraceTandemArrayPatternAlphabetPatternSetMap;
		TreeSet<String> patternAlphabetPatternSet;
		
		for(String encodedTrace : uniqueEncodedTraceList){
			encodedTraceLength = encodedTrace.length()/encodingLength;
			if(encodedTraceLength <= 1)
				continue;
			suffixTree = new UkkonenSuffixTree(encodingLength, encodedTrace);
			suffixTree.LZDecomposition();
			
			currentTraceTandemArrayPatternAlphabetPatternSetMap = suffixTree.getPrimitiveTandemRepeats();
			for(TreeSet<String> tandemArrayPatternAlphabet : currentTraceTandemArrayPatternAlphabetPatternSetMap.keySet()){
				if(entireLogTandemArrayPatternAlphabetPatternSetMap.containsKey(tandemArrayPatternAlphabet)){
					patternAlphabetPatternSet = entireLogTandemArrayPatternAlphabetPatternSetMap.get(tandemArrayPatternAlphabet);
				}else{
					patternAlphabetPatternSet = new TreeSet<String>();
				}
				patternAlphabetPatternSet.addAll(currentTraceTandemArrayPatternAlphabetPatternSetMap.get(tandemArrayPatternAlphabet));
				entireLogTandemArrayPatternAlphabetPatternSetMap.put(tandemArrayPatternAlphabet, patternAlphabetPatternSet);
			}
		}
		int noPatterns = 0;
		for(Set<String> patternAlphabet : entireLogTandemArrayPatternAlphabetPatternSetMap.keySet())
			noPatterns += entireLogTandemArrayPatternAlphabetPatternSetMap.get(patternAlphabet).size();
		System.out.println("No. Tandem Arrays: "+noPatterns);
		System.out.println("No. Pattern Alphabets: "+entireLogTandemArrayPatternAlphabetPatternSetMap.size());
		
		System.out.println("Took "+w.msecs()+" msecs to find tandem arrays");
		/*
		 * Now for tandem arrays, let us try to add all cyclic permutations; 
		 */
		Set<String> cyclicPermutationSet = new TreeSet<String>(new DescendingStringLengthComparator());
		String tandemRepeat,permutedTandemRepeat;
		int activityStartIndex;
		Set<String> nonStartingActivitySet = new HashSet<String>();
		for(TreeSet<String> tandemArrayAlphabet : entireLogTandemArrayPatternAlphabetPatternSetMap.keySet()){
			patternAlphabetPatternSet = entireLogTandemArrayPatternAlphabetPatternSetMap.get(tandemArrayAlphabet);
			if(patternAlphabetPatternSet.size() == tandemArrayAlphabet.size())
				continue;
			else{
//				System.out.println("Adding Cyclic Permutations");
				tandemRepeat = patternAlphabetPatternSet.iterator().next();
				
				cyclicPermutationSet.clear();
				cyclicPermutationSet.addAll(patternAlphabetPatternSet);
				
				nonStartingActivitySet.clear();
				nonStartingActivitySet.addAll(tandemArrayAlphabet);
				
				for(String tr : patternAlphabetPatternSet)
					nonStartingActivitySet.remove(tr.substring(0, encodingLength));
				
				for(String nonStartingActivity : nonStartingActivitySet){
					activityStartIndex = tandemRepeat.indexOf(nonStartingActivity)/encodingLength;
					permutedTandemRepeat = tandemRepeat.substring(activityStartIndex*encodingLength);
					permutedTandemRepeat += tandemRepeat.substring(0,activityStartIndex*encodingLength);
					cyclicPermutationSet.add(permutedTandemRepeat);
				}
				
//				System.out.println("Adding Cyclic Permutations for TRA "+tandemArrayAlphabet);
//				System.out.println("Original TRs: "+patternAlphabetPatternSet);
				
				patternAlphabetPatternSet.addAll(cyclicPermutationSet);
//				System.out.println("Altered TRs: "+patternAlphabetPatternSet);
				entireLogTandemArrayPatternAlphabetPatternSetMap.put(tandemArrayAlphabet, patternAlphabetPatternSet);
			}
		}
		
		noPatterns = 0;
		for(Set<String> patternAlphabet : entireLogTandemArrayPatternAlphabetPatternSetMap.keySet())
			noPatterns += entireLogTandemArrayPatternAlphabetPatternSetMap.get(patternAlphabet).size();
		System.out.println("No. Tandem Arrays (with CPerm): "+noPatterns);
		System.out.println("No. Pattern Alphabets (with CPerm): "+entireLogTandemArrayPatternAlphabetPatternSetMap.size());
		
		System.out.println("Took "+w.msecs()+" msecs to find tandem arrays(2)");
		
		/*
		 * Identify ignorable pattern alphabets; These are pattern alphabets
		 * where the coefficient of correlation is > 1
		 */
		
		if(ignorablePatternAlphabetSet == null)
			ignorablePatternAlphabetSet = new HashSet<Set<String>>();
		if(patternAlphabetMeanCountMap == null)
			patternAlphabetMeanCountMap = new HashMap<Set<String>, Float>();
		if(patternAlphabetStdDevCountMap == null)
			patternAlphabetStdDevCountMap = new HashMap<Set<String>, Float>();
		
		float mean, stddev;
		for(Set<String> patternAlphabet : entireLogTandemArrayPatternAlphabetPatternSetMap.keySet()){
			
			mean = 0;
			for(String encodedActivity : patternAlphabet){
				mean += activityCountMap.get(encodedActivity);
			}
			mean /= patternAlphabet.size();
			stddev = 0;
			for(String encodedActivity : patternAlphabet){
				stddev += (activityCountMap.get(encodedActivity)-mean)*(activityCountMap.get(encodedActivity)-mean);
			}
			stddev = (float)Math.sqrt(stddev/patternAlphabet.size());
			
			patternAlphabetMeanCountMap.put(patternAlphabet, mean);
			patternAlphabetStdDevCountMap.put(patternAlphabet, stddev);
			
			if(stddev >= mean)
				ignorablePatternAlphabetSet.add(patternAlphabet);
		}
		
		System.out.println("No. Ignorable Pattern Alphabets: "+ignorablePatternAlphabetSet.size());
		
//		System.out.println("Printing Tandem Arrays");
//		printPatterns(entireLogTandemArrayPatternAlphabetPatternSetMap);
		FileIO io = new FileIO();
		io.writeToFile(outputDir, "EntireLogTandemArrayPatternAlphabetPatternSetMap.txt", entireLogTandemArrayPatternAlphabetPatternSetMap, delim);
	}
	
	private void findBaseTandemArrays(){
		Watch w = new Watch();
		w.start();
		entireLogTandemArrayBasePatternAlphabetPatternSetMap.clear();
		
		int encodingLength = encodedLog.getEncodingLength();
		int repeatLength;
		
		TreeSet<String> tandemArrayPatternAlphabetPatternSet;
		TreeSet<String> tandemArrayBasePatternAlphabetPatternSet;
		
		for(TreeSet<String> tandemArrayPatternAlphabet : entireLogTandemArrayPatternAlphabetPatternSetMap.keySet()){
			tandemArrayPatternAlphabetPatternSet = entireLogTandemArrayPatternAlphabetPatternSetMap.get(tandemArrayPatternAlphabet);
			tandemArrayBasePatternAlphabetPatternSet = new TreeSet<String>();
			
			for(String tandemArrayRepeat : tandemArrayPatternAlphabetPatternSet){
				repeatLength = tandemArrayRepeat.length()/encodingLength;
				if(repeatLength == tandemArrayPatternAlphabet.size()){
					tandemArrayBasePatternAlphabetPatternSet.add(tandemArrayRepeat);
				}
			}
			
			if(tandemArrayBasePatternAlphabetPatternSet.size() > 0){
				entireLogTandemArrayBasePatternAlphabetPatternSetMap.put(tandemArrayPatternAlphabet, tandemArrayBasePatternAlphabetPatternSet);
			}
		}
		int noBasePatterns = 0;
		for(Set<String> patternAlphabet : entireLogTandemArrayBasePatternAlphabetPatternSetMap.keySet())
			noBasePatterns += entireLogTandemArrayBasePatternAlphabetPatternSetMap.get(patternAlphabet).size();
		System.out.println("No. Base Tandem Arrays: "+noBasePatterns);
		System.out.println("No. Pattern Alphabets: "+entireLogTandemArrayBasePatternAlphabetPatternSetMap.size());
		System.out.println("Took "+w.msecs()+" msecs to find base tandem arrays");
//		System.out.println("Printing Base Tandem Arrays");
//		printPatterns(entireLogTandemArrayBasePatternAlphabetPatternSetMap);
		FileIO io = new FileIO();
		io.writeToFile(outputDir, "EntireLogTandemArrayBasePatternAlphabetPatternSetMap.txt", entireLogTandemArrayBasePatternAlphabetPatternSetMap, delim);
	}

	/**
	 * This method finds the maximal repeats in the event log In order to
	 * find maximal repeats, we need to combine the traces with a distinct
	 * delimiter between them So, we first generate a new encoding for the log
	 * so that the trace identifiers are also encoded to use them as delimiters
	 * 
	 * If the combined stream is longer than the MAX_STRLENGTH, subsets of
	 * traces are combined and repeats are discovered over all combinations of
	 * the splits
	 */
	private void findMaximalRepeats(){
		if(encodedLogWithTraceDelimiters == null)
			encodeLogWithTraceDelimiters();
		
		int encodingLengthWithTraceDelimiters = encodedLogWithTraceDelimiters.getEncodingLength();
		Map<String, String> charActivityMapWithTraceDelimiters = encodedLogWithTraceDelimiters.getCharActivityMap();
		
		getCombinedTrace();
		
		int noSplits = splitTraces.length;
		
		
		entireLogMaximalRepeatPatternAlphabetPatternSetMap.clear();
		Map<String, String> activityCharMap = encodedLog.getActivityCharMap();
		
		TreeSet<String> maximalRepeatAlphabet;
		TreeSet<String> maximalRepeatAlphabetPatternSet;
		Set<String> maximalRepeatAlphabetPatternSetWithTraceDelimiters;
		int repeatLength;
		StringBuilder repeatStringBuilder = new StringBuilder();
		Map<TreeSet<String>, TreeSet<String>> maximalRepeatPatternAlphabetPatternSetMap;
	
		EquivalenceClass equivalenceClass = new EquivalenceClass();
		UkkonenSuffixTree suffixTree;

//		System.out.println("No. Splits: "+noSplits);
		for(int i = 0; i < noSplits; i++){
			for(int k = i; k < noSplits; k++){
				if(noSplits == 1)
					suffixTree = new UkkonenSuffixTree(encodingLengthWithTraceDelimiters, splitTraces[i]);
				else if(noSplits > 1 && k+1 < noSplits)
					suffixTree = new UkkonenSuffixTree(encodingLengthWithTraceDelimiters, splitTraces[i].concat(splitTraces[k+1]));
				else
					continue;

				
				suffixTree.findLeftDiverseNodes();
//				System.out.println("Invoked Left Diverse Nodes");
				maximalRepeatPatternAlphabetPatternSetMap = equivalenceClass.getAlphabetEquivalenceClassMap(encodingLengthWithTraceDelimiters, suffixTree.getMaximalRepeats());
				
				for(Set<String> maximalRepeatAlphabetWithTraceDelimiters : maximalRepeatPatternAlphabetPatternSetMap.keySet()){
					if(maximalRepeatAlphabetWithTraceDelimiters == null)
						continue;
					
					maximalRepeatAlphabet = new TreeSet<String>();
					for(String encodedActivityWithTraceDelimiter : maximalRepeatAlphabetWithTraceDelimiters){
						if(activityCharMap.containsKey(charActivityMapWithTraceDelimiters.get(encodedActivityWithTraceDelimiter.trim())))
							maximalRepeatAlphabet.add(activityCharMap.get(charActivityMapWithTraceDelimiters.get(encodedActivityWithTraceDelimiter.trim())));
						else{
							System.out.println("This should have never happened");
							System.out.println(i+"@"+k+" "+charActivityMapWithTraceDelimiters.get(encodedActivityWithTraceDelimiter.trim())+" Not Present");
						}
					}
				
					maximalRepeatAlphabetPatternSetWithTraceDelimiters = maximalRepeatPatternAlphabetPatternSetMap.get(maximalRepeatAlphabetWithTraceDelimiters);
					maximalRepeatAlphabetPatternSet = new TreeSet<String>();
					for(String maximalRepeatWithTraceDelimiter : maximalRepeatAlphabetPatternSetWithTraceDelimiters){
						repeatLength = maximalRepeatWithTraceDelimiter.length()/encodingLengthWithTraceDelimiters;
						repeatStringBuilder.setLength(0);
						for(int r = 0; r < repeatLength; r++){
							repeatStringBuilder.append(activityCharMap.get(charActivityMapWithTraceDelimiters.get(maximalRepeatWithTraceDelimiter.substring(r*encodingLengthWithTraceDelimiters, (r+1)*encodingLengthWithTraceDelimiters))));
						}
						maximalRepeatAlphabetPatternSet.add(repeatStringBuilder.toString());
					}
					
					if(entireLogMaximalRepeatPatternAlphabetPatternSetMap.containsKey(maximalRepeatAlphabet)){
						maximalRepeatAlphabetPatternSet.addAll(entireLogMaximalRepeatPatternAlphabetPatternSetMap.get(maximalRepeatAlphabet));
					}
					entireLogMaximalRepeatPatternAlphabetPatternSetMap.put(maximalRepeatAlphabet, maximalRepeatAlphabetPatternSet);
				}
			}
		}
		//Get the total number of maximal repeats
		int noMaximalRepeats = 0;
		for(Set<String> patternAlphabet : entireLogMaximalRepeatPatternAlphabetPatternSetMap.keySet())
			noMaximalRepeats += entireLogMaximalRepeatPatternAlphabetPatternSetMap.get(patternAlphabet).size();
		
		/*
		 * Identify ignorable pattern alphabets; These are pattern alphabets
		 * where the coefficient of correlation is > 1
		 */
		
		if(ignorablePatternAlphabetSet == null)
			ignorablePatternAlphabetSet = new HashSet<Set<String>>();
		if(patternAlphabetMeanCountMap == null)
			patternAlphabetMeanCountMap = new HashMap<Set<String>, Float>();
		if(patternAlphabetStdDevCountMap == null)
			patternAlphabetStdDevCountMap = new HashMap<Set<String>, Float>();
		
		float mean, stddev;
		for(Set<String> patternAlphabet : entireLogMaximalRepeatPatternAlphabetPatternSetMap.keySet()){
			
			mean = 0;
			for(String encodedActivity : patternAlphabet){
				mean += activityCountMap.get(encodedActivity);
			}
			mean /= patternAlphabet.size();
			stddev = 0;
			for(String encodedActivity : patternAlphabet){
				stddev += (activityCountMap.get(encodedActivity)-mean)*(activityCountMap.get(encodedActivity)-mean);
			}
			stddev = (float)Math.sqrt(stddev/patternAlphabet.size());
			
			patternAlphabetMeanCountMap.put(patternAlphabet, mean);
			patternAlphabetStdDevCountMap.put(patternAlphabet, stddev);
			
			if(stddev >= mean)
				ignorablePatternAlphabetSet.add(patternAlphabet);
		}
		
		FileIO io = new FileIO();
		io.writeToFile(outputDir, "EntireLogMaximalRepeatPatternAlphabetPatternSetMap.txt", entireLogMaximalRepeatPatternAlphabetPatternSetMap, delim);
		io.writeToFile(outputDir, "IgnorablePatternAlphabetSet.txt", ignorablePatternAlphabetSet);
		io.writeToFile(outputDir, "ActitivtyCountMap.txt", activityCountMap, delim);
		io.writeToFile(outputDir, "PatternAlphabetMeanCountMap.txt", patternAlphabetMeanCountMap, delim);
		io.writeToFile(outputDir, "PatternAlphabetStdDevCountMap.txt", patternAlphabetStdDevCountMap, delim);
	}
	
	private void findBaseMaximalRepeats(){
		entireLogMaximalRepeatBasePatternAlphabetPatternSetMap.clear();
		
		int encodingLength = encodedLog.getEncodingLength();
		int repeatLength;
		
		TreeSet<String> maximalRepeatPatternAlphabetPatternSet;
		TreeSet<String> maximalRepeatBasePatternAlphabetPatternSet;
		
		for(TreeSet<String> maximalRepeatPatternAlphabet : entireLogMaximalRepeatPatternAlphabetPatternSetMap.keySet()){
			maximalRepeatPatternAlphabetPatternSet = entireLogMaximalRepeatPatternAlphabetPatternSetMap.get(maximalRepeatPatternAlphabet);
			maximalRepeatBasePatternAlphabetPatternSet = new TreeSet<String>();
			
			for(String maximalRepeat : maximalRepeatPatternAlphabetPatternSet){
				repeatLength = maximalRepeat.length()/encodingLength;
				if(repeatLength == maximalRepeatPatternAlphabet.size()){
					maximalRepeatBasePatternAlphabetPatternSet.add(maximalRepeat);
				}
			}
			
			if(maximalRepeatBasePatternAlphabetPatternSet.size() > 0){
				entireLogMaximalRepeatBasePatternAlphabetPatternSetMap.put(maximalRepeatPatternAlphabet, maximalRepeatBasePatternAlphabetPatternSet);
			}
		}
		
		int noBasePatterns = 0;
		for(Set<String> patternAlphabet : entireLogMaximalRepeatBasePatternAlphabetPatternSetMap.keySet()){
			noBasePatterns += entireLogMaximalRepeatBasePatternAlphabetPatternSetMap.get(patternAlphabet).size();
		}
		
		FileIO io = new FileIO();
		io.writeToFile(outputDir, "EntireLogMaximalRepeatBasePatternAlphabetPatternSetMap.txt", entireLogMaximalRepeatBasePatternAlphabetPatternSetMap, delim);
	}
		
	private void computeCombinedPatterns(){
		entireLogCombinedPatternAlphabetPatternSetMap.clear();
		entireLogCombinedBasePatternAlphabetPatternSetMap.clear();
		
		int encodingLength = encodedLog.getEncodingLength();
		int repeatLength;
		Set<TreeSet<String>> combinedAlphabetSet = new HashSet<TreeSet<String>>();
		combinedAlphabetSet.addAll(entireLogTandemArrayPatternAlphabetPatternSetMap.keySet());
		combinedAlphabetSet.addAll(entireLogMaximalRepeatPatternAlphabetPatternSetMap.keySet());
		
		TreeSet<String> patternAlphabetPatternSet;
		TreeSet<String> patternAlphabetBasePatternSet;
		
		for(TreeSet<String> patternAlphabet : combinedAlphabetSet){
			patternAlphabetPatternSet = new TreeSet<String>();
			if(entireLogTandemArrayPatternAlphabetPatternSetMap.containsKey(patternAlphabet)){
				patternAlphabetPatternSet.addAll(entireLogTandemArrayPatternAlphabetPatternSetMap.get(patternAlphabet));
			}
			if(entireLogMaximalRepeatPatternAlphabetPatternSetMap.containsKey(patternAlphabet)){
				patternAlphabetPatternSet.addAll(entireLogMaximalRepeatPatternAlphabetPatternSetMap.get(patternAlphabet));
			}
			
			entireLogCombinedPatternAlphabetPatternSetMap.put(patternAlphabet, patternAlphabetPatternSet);
			
			patternAlphabetBasePatternSet = new TreeSet<String>();
			for(String pattern : patternAlphabetPatternSet){
				repeatLength = pattern.length()/encodingLength;
				if(patternAlphabet.size() == repeatLength)
					patternAlphabetBasePatternSet.add(pattern);
			}
			
			if(patternAlphabetBasePatternSet.size() > 0)
				entireLogCombinedBasePatternAlphabetPatternSetMap.put(patternAlphabet, patternAlphabetBasePatternSet);
		}
		
		FileIO io = new FileIO();
		io.writeToFile(outputDir, "EntireLogCombinedPatternAlphabetPatternSetMap.txt", entireLogCombinedPatternAlphabetPatternSetMap, delim);
		io.writeToFile(outputDir, "EntireLogCombinedBasePatternAlphabetPatternSetMap.txt", entireLogCombinedBasePatternAlphabetPatternSetMap, delim);
	}
	
	private void getCombinedTrace(){
		int encodingLengthWithTraceDelimiters = encodedLogWithTraceDelimiters.getEncodingLength();
		List<String> uniqueEncodedTraceListWithTraceDelimiters = encodedLogWithTraceDelimiters.getUniqueEncodedTraceList();
		Map<String, String> activityCharMapWithTraceDelimiters = encodedLogWithTraceDelimiters.getActivityCharMap();
		
		StringBuilder combinedTraceStringBuilder = new StringBuilder();
		int traceIndex = 0;
		List<Integer> traceEndIndexInCombinedStreamList = new ArrayList<Integer>();
		
		int currentLength;
		
		for(String encodedTraceWithTraceDelimiter : uniqueEncodedTraceListWithTraceDelimiters){
			combinedTraceStringBuilder.append(encodedTraceWithTraceDelimiter);
			//add the distinct delimiter between traces
			combinedTraceStringBuilder.append(activityCharMapWithTraceDelimiters.get("Trace: "+traceIndex));
			currentLength = combinedTraceStringBuilder.length()/encodingLengthWithTraceDelimiters;
			traceEndIndexInCombinedStreamList.add(currentLength);
			traceIndex++;
		}
		
		splitTraces = new String[1];
		splitTraces[0] = combinedTraceStringBuilder.toString();
	}
	
	private void computePatternMetrics(){
		computeOverlappingMetrics();
		computeNOACMetrics();
		computeNOGACMetrics();
		computeConservednessMetric();
	}
	
	private void computeOverlappingMetrics(){
		System.out.println("In computeOverlappingMetrics()");
		Map<String, Integer> patternCountMap = new HashMap<String, Integer>();

		overlappingPatternAlphabetCount = new MetricPatternAlphabetCount();
		
		Map<TreeSet<String>, Integer> tandemArrayPatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayBasePatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayPatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayBasePatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		Map<TreeSet<String>, Integer> maximalRepeatPatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatBasePatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatPatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatBasePatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		Map<TreeSet<String>, Integer> combinedPatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedBasePatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedPatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedBasePatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		Map<TreeSet<String>, Integer> tandemArrayPatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayBasePatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayPatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayBasePatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		Map<TreeSet<String>, Integer> maximalRepeatPatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatBasePatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatPatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatBasePatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		Map<TreeSet<String>, Integer> combinedPatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedBasePatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedPatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedBasePatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();

		
		int encodingLength = encodedLog.getEncodingLength();
		List<String> uniqueEncodedTraceList = encodedLog.getUniqueEncodedTraceList();
		Map<String, TreeSet<Integer>> encodedTraceIdenticalTraceIndicesSetMap = encodedLog.getEncodedTraceIdenticalTraceIndicesSetMap();
		
		TreeSet<String> patternAlphabetPatternSet;

		int noIdenticalTraces, encodedTraceLength, noSplits, patternCount, patternAlphabetCount;

		
		UkkonenSuffixTree suffixTree;
		String[] splitTraces;
		int count;
		int noTraces = encodedLog.getNoTraces();
		int noUniqueTraces = encodedLog.getNoUniqueTraces();

		for(String encodedTrace : uniqueEncodedTraceList){
			noIdenticalTraces = encodedTraceIdenticalTraceIndicesSetMap.get(encodedTrace).size();
			encodedTraceLength = encodedTrace.length()/encodingLength;
			noSplits = new Double(Math.ceil(1.0*encodedTraceLength/MAX_STRLENGTH)).intValue();
			splitTraces = new String[noSplits];
			
			for(int i = 0; i < noSplits-1; i++){
				splitTraces[i] = encodedTrace.substring(i*MAX_STRLENGTH*encodingLength, (i+1)*MAX_STRLENGTH*encodingLength);
			}
			splitTraces[noSplits-1] = encodedTrace.substring((noSplits-1)*MAX_STRLENGTH*encodingLength);
			
			for(String splitTrace : splitTraces){
				suffixTree = new UkkonenSuffixTree(encodingLength, splitTrace);
				
				for(TreeSet<String> patternAlphabet : entireLogCombinedPatternAlphabetPatternSetMap.keySet()){
					if(ignorablePatternAlphabetSet != null && ignorablePatternAlphabetSet.contains(patternAlphabet))
						continue;
					patternAlphabetPatternSet = entireLogCombinedPatternAlphabetPatternSetMap.get(patternAlphabet);
					patternAlphabetCount = 0;
					
					for(String pattern : patternAlphabetPatternSet){
						patternCount = suffixTree.noMatches(pattern);
						patternCountMap.put(pattern, patternCount);
						patternAlphabetCount += patternCount;
					}
					
					if(patternAlphabetCount > 0){
						count = patternAlphabetCount;
						if(combinedPatternAlphabetNonDuplicateCountMap.containsKey(patternAlphabet))
							count += combinedPatternAlphabetNonDuplicateCountMap.get(patternAlphabet);
						combinedPatternAlphabetNonDuplicateCountMap.put(patternAlphabet, count);
						
						count = patternAlphabetCount*noIdenticalTraces;
						if(combinedPatternAlphabetCountMap.containsKey(patternAlphabet))
							count += combinedPatternAlphabetCountMap.get(patternAlphabet);
						combinedPatternAlphabetCountMap.put(patternAlphabet, count);
			
						count = noIdenticalTraces;
						if(combinedPatternAlphabetInstanceCountMap.containsKey(patternAlphabet))
							count += combinedPatternAlphabetInstanceCountMap.get(patternAlphabet);
						combinedPatternAlphabetInstanceCountMap.put(patternAlphabet, count);
						
						count = 1;
						if(combinedPatternAlphabetNonDuplicateInstanceCountMap.containsKey(patternAlphabet))
							count += combinedPatternAlphabetNonDuplicateInstanceCountMap.get(patternAlphabet);
						combinedPatternAlphabetNonDuplicateInstanceCountMap.put(patternAlphabet, count);
					}
					
					if(entireLogCombinedBasePatternAlphabetPatternSetMap.containsKey(patternAlphabet)){
						patternAlphabetPatternSet = entireLogCombinedBasePatternAlphabetPatternSetMap.get(patternAlphabet);
						patternAlphabetCount = 0;
						for(String pattern : patternAlphabetPatternSet){
							patternAlphabetCount += patternCountMap.get(pattern);
						}
						
						if(patternAlphabetCount > 0){
							count = patternAlphabetCount;
							if(combinedBasePatternAlphabetNonDuplicateCountMap.containsKey(patternAlphabet))
								count += combinedBasePatternAlphabetNonDuplicateCountMap.get(patternAlphabet);
							combinedBasePatternAlphabetNonDuplicateCountMap.put(patternAlphabet, count);
							
							count = patternAlphabetCount*noIdenticalTraces;
							if(combinedBasePatternAlphabetCountMap.containsKey(patternAlphabet))
								count += combinedBasePatternAlphabetCountMap.get(patternAlphabet);
							combinedBasePatternAlphabetCountMap.put(patternAlphabet, count);
					
							count = noIdenticalTraces;
							if(combinedBasePatternAlphabetInstanceCountMap.containsKey(patternAlphabet))
								count += combinedBasePatternAlphabetInstanceCountMap.get(patternAlphabet);
							combinedBasePatternAlphabetInstanceCountMap.put(patternAlphabet, count);
							
							count = 1;
							if(combinedBasePatternAlphabetNonDuplicateInstanceCountMap.containsKey(patternAlphabet))
								count += combinedBasePatternAlphabetNonDuplicateInstanceCountMap.get(patternAlphabet);
							combinedBasePatternAlphabetNonDuplicateInstanceCountMap.put(patternAlphabet, count);
						}
					}
					
					if(entireLogTandemArrayPatternAlphabetPatternSetMap.containsKey(patternAlphabet)){
						patternAlphabetPatternSet = entireLogTandemArrayPatternAlphabetPatternSetMap.get(patternAlphabet);
						patternAlphabetCount = 0;
						for(String pattern : patternAlphabetPatternSet){
							patternAlphabetCount += patternCountMap.get(pattern);
						}

						if(patternAlphabetCount > 0){
							count = patternAlphabetCount;
							if(tandemArrayPatternAlphabetNonDuplicateCountMap.containsKey(patternAlphabet))
								count += tandemArrayPatternAlphabetNonDuplicateCountMap.get(patternAlphabet);
							tandemArrayPatternAlphabetNonDuplicateCountMap.put(patternAlphabet, count);
							
							count = patternAlphabetCount*noIdenticalTraces;
							if(tandemArrayPatternAlphabetCountMap.containsKey(patternAlphabet))
								count += tandemArrayPatternAlphabetCountMap.get(patternAlphabet);
							tandemArrayPatternAlphabetCountMap.put(patternAlphabet, count);
						
							count = noIdenticalTraces;
							if(tandemArrayPatternAlphabetInstanceCountMap.containsKey(patternAlphabet))
								count += tandemArrayPatternAlphabetInstanceCountMap.get(patternAlphabet);
							tandemArrayPatternAlphabetInstanceCountMap.put(patternAlphabet, count);
							
							count = 1;
							if(tandemArrayPatternAlphabetNonDuplicateInstanceCountMap.containsKey(patternAlphabet))
								count += tandemArrayPatternAlphabetNonDuplicateInstanceCountMap.get(patternAlphabet);
							tandemArrayPatternAlphabetNonDuplicateInstanceCountMap.put(patternAlphabet, count);
						}
					}
					
					if(entireLogTandemArrayBasePatternAlphabetPatternSetMap.containsKey(patternAlphabet)){
						patternAlphabetPatternSet = entireLogTandemArrayBasePatternAlphabetPatternSetMap.get(patternAlphabet);
						patternAlphabetCount = 0;
						for(String pattern : patternAlphabetPatternSet){
							patternAlphabetCount += patternCountMap.get(pattern);
						}
						
						if(patternAlphabetCount > 0){
							count = patternAlphabetCount;
							if(tandemArrayBasePatternAlphabetNonDuplicateCountMap.containsKey(patternAlphabet))
								count += tandemArrayBasePatternAlphabetNonDuplicateCountMap.get(patternAlphabet);
							tandemArrayBasePatternAlphabetNonDuplicateCountMap.put(patternAlphabet, count);
							
							count = patternAlphabetCount*noIdenticalTraces;
							if(tandemArrayBasePatternAlphabetCountMap.containsKey(patternAlphabet))
								count += tandemArrayBasePatternAlphabetCountMap.get(patternAlphabet);
							tandemArrayBasePatternAlphabetCountMap.put(patternAlphabet, count);
							
							count = noIdenticalTraces;
							if(tandemArrayBasePatternAlphabetInstanceCountMap.containsKey(patternAlphabet))
								count += tandemArrayBasePatternAlphabetInstanceCountMap.get(patternAlphabet);
							tandemArrayBasePatternAlphabetInstanceCountMap.put(patternAlphabet, count);
							
							count = 1;
							if(tandemArrayBasePatternAlphabetNonDuplicateInstanceCountMap.containsKey(patternAlphabet))
								count += tandemArrayBasePatternAlphabetNonDuplicateInstanceCountMap.get(patternAlphabet);
							tandemArrayBasePatternAlphabetNonDuplicateInstanceCountMap.put(patternAlphabet, count);
						}
					}
					
					if(entireLogMaximalRepeatPatternAlphabetPatternSetMap.containsKey(patternAlphabet)){
						patternAlphabetPatternSet = entireLogMaximalRepeatPatternAlphabetPatternSetMap.get(patternAlphabet);
						patternAlphabetCount = 0;
						for(String pattern : patternAlphabetPatternSet){
							patternAlphabetCount += patternCountMap.get(pattern);
						}
						if(patternAlphabetCount > 0){
							count = patternAlphabetCount;
							if(maximalRepeatPatternAlphabetNonDuplicateCountMap.containsKey(patternAlphabet))
								count += maximalRepeatPatternAlphabetNonDuplicateCountMap.get(patternAlphabet);
							maximalRepeatPatternAlphabetNonDuplicateCountMap.put(patternAlphabet, count);
							
							count = patternAlphabetCount*noIdenticalTraces;
							if(maximalRepeatPatternAlphabetCountMap.containsKey(patternAlphabet))
								count += maximalRepeatPatternAlphabetCountMap.get(patternAlphabet);
							maximalRepeatPatternAlphabetCountMap.put(patternAlphabet, count);
							
							count = noIdenticalTraces;
							if(maximalRepeatPatternAlphabetInstanceCountMap.containsKey(patternAlphabet))
								count += maximalRepeatPatternAlphabetInstanceCountMap.get(patternAlphabet);
							maximalRepeatPatternAlphabetInstanceCountMap.put(patternAlphabet, count);
							
							count = 1;
							if(maximalRepeatPatternAlphabetNonDuplicateInstanceCountMap.containsKey(patternAlphabet))
								count += maximalRepeatPatternAlphabetNonDuplicateInstanceCountMap.get(patternAlphabet);
							maximalRepeatPatternAlphabetNonDuplicateInstanceCountMap.put(patternAlphabet, count);
						}
					}
					
					if(entireLogMaximalRepeatBasePatternAlphabetPatternSetMap.containsKey(patternAlphabet)){
						patternAlphabetPatternSet = entireLogMaximalRepeatBasePatternAlphabetPatternSetMap.get(patternAlphabet);
						patternAlphabetCount = 0;
						for(String pattern : patternAlphabetPatternSet){
							patternAlphabetCount += patternCountMap.get(pattern);
						}
						
						if(patternAlphabetCount  > 0){
							count = patternAlphabetCount;
							if(maximalRepeatBasePatternAlphabetNonDuplicateCountMap.containsKey(patternAlphabet))
								count += maximalRepeatBasePatternAlphabetNonDuplicateCountMap.get(patternAlphabet);
							maximalRepeatBasePatternAlphabetNonDuplicateCountMap.put(patternAlphabet, count);
							
							count = patternAlphabetCount*noIdenticalTraces;
							if(maximalRepeatBasePatternAlphabetCountMap.containsKey(patternAlphabet))
								count += maximalRepeatBasePatternAlphabetCountMap.get(patternAlphabet);
							maximalRepeatBasePatternAlphabetCountMap.put(patternAlphabet, count);
							
							count = noIdenticalTraces;
							if(maximalRepeatBasePatternAlphabetInstanceCountMap.containsKey(patternAlphabet))
								count += maximalRepeatBasePatternAlphabetInstanceCountMap.get(patternAlphabet);
							maximalRepeatBasePatternAlphabetInstanceCountMap.put(patternAlphabet, count);
							
							count = 1;
							if(maximalRepeatBasePatternAlphabetNonDuplicateInstanceCountMap.containsKey(patternAlphabet))
								count += maximalRepeatBasePatternAlphabetNonDuplicateInstanceCountMap.get(patternAlphabet);
							maximalRepeatBasePatternAlphabetNonDuplicateInstanceCountMap.put(patternAlphabet, count);
						}
					}
				}
			}
		}

		//The instance counts contain absolute count values, we need to change them to percentages
		int instanceCountPercentage;
		for(TreeSet<String> combinedBasePatternAlphabet : combinedBasePatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(combinedBasePatternAlphabetInstanceCountMap.get(combinedBasePatternAlphabet)*100.0f/noTraces);
			combinedBasePatternAlphabetInstanceCountMap.put(combinedBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> combinedBasePatternAlphabet : combinedBasePatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(combinedBasePatternAlphabetNonDuplicateInstanceCountMap.get(combinedBasePatternAlphabet)*100.0f/noUniqueTraces);
			combinedBasePatternAlphabetNonDuplicateInstanceCountMap.put(combinedBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> combinedPatternAlphabet : combinedPatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(combinedPatternAlphabetInstanceCountMap.get(combinedPatternAlphabet)*100.0f/noTraces);
			combinedPatternAlphabetInstanceCountMap.put(combinedPatternAlphabet, instanceCountPercentage);
			
			if(instanceCountPercentage <= INSTANCECOUNT_PERCENTAGE_THRESHOLD)
				ignorablePatternAlphabetSet.add(combinedPatternAlphabet);
		}
		for(TreeSet<String> combinedPatternAlphabet : combinedPatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(combinedPatternAlphabetNonDuplicateInstanceCountMap.get(combinedPatternAlphabet)*100.0f/noUniqueTraces);
			combinedPatternAlphabetNonDuplicateInstanceCountMap.put(combinedPatternAlphabet, instanceCountPercentage);
		}
		
		for(TreeSet<String> tandemArrayBasePatternAlphabet : tandemArrayBasePatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(tandemArrayBasePatternAlphabetInstanceCountMap.get(tandemArrayBasePatternAlphabet)*100.0f/noTraces);
			tandemArrayBasePatternAlphabetInstanceCountMap.put(tandemArrayBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> tandemArrayBasePatternAlphabet : tandemArrayBasePatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(tandemArrayBasePatternAlphabetNonDuplicateInstanceCountMap.get(tandemArrayBasePatternAlphabet)*100.0f/noUniqueTraces);
			tandemArrayBasePatternAlphabetNonDuplicateInstanceCountMap.put(tandemArrayBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> tandemArrayPatternAlphabet : tandemArrayPatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(tandemArrayPatternAlphabetInstanceCountMap.get(tandemArrayPatternAlphabet)*100.0f/noTraces);
			tandemArrayPatternAlphabetInstanceCountMap.put(tandemArrayPatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> tandemArrayPatternAlphabet : tandemArrayPatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(tandemArrayPatternAlphabetNonDuplicateInstanceCountMap.get(tandemArrayPatternAlphabet)*100.0f/noUniqueTraces);
			tandemArrayPatternAlphabetNonDuplicateInstanceCountMap.put(tandemArrayPatternAlphabet, instanceCountPercentage);
		}
		
		for(TreeSet<String> maximalRepeatBasePatternAlphabet : maximalRepeatBasePatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(maximalRepeatBasePatternAlphabetInstanceCountMap.get(maximalRepeatBasePatternAlphabet)*100.0f/noTraces);
			maximalRepeatBasePatternAlphabetInstanceCountMap.put(maximalRepeatBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> maximalRepeatBasePatternAlphabet : maximalRepeatBasePatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(maximalRepeatBasePatternAlphabetNonDuplicateInstanceCountMap.get(maximalRepeatBasePatternAlphabet)*100.0f/noUniqueTraces);
			maximalRepeatBasePatternAlphabetNonDuplicateInstanceCountMap.put(maximalRepeatBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> maximalRepeatPatternAlphabet : maximalRepeatPatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(maximalRepeatPatternAlphabetInstanceCountMap.get(maximalRepeatPatternAlphabet)*100.0f/noTraces);
			maximalRepeatPatternAlphabetInstanceCountMap.put(maximalRepeatPatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> maximalRepeatPatternAlphabet : maximalRepeatPatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(maximalRepeatPatternAlphabetNonDuplicateInstanceCountMap.get(maximalRepeatPatternAlphabet)*100.0f/noUniqueTraces);
			maximalRepeatPatternAlphabetNonDuplicateInstanceCountMap.put(maximalRepeatPatternAlphabet, instanceCountPercentage);
		}
		
		overlappingPatternAlphabetCount.setContributingTandemArrayPatternAlphabetPatternSetMap(entireLogTandemArrayPatternAlphabetPatternSetMap);
		overlappingPatternAlphabetCount.setContributingTandemArrayBasePatternAlphabetPatternSetMap(entireLogTandemArrayBasePatternAlphabetPatternSetMap);
		overlappingPatternAlphabetCount.setContributingMaximalRepeatPatternAlphabetPatternSetMap(entireLogMaximalRepeatPatternAlphabetPatternSetMap);
		overlappingPatternAlphabetCount.setContributingMaximalRepeatBasePatternAlphabetPatternSetMap(entireLogMaximalRepeatBasePatternAlphabetPatternSetMap);
		overlappingPatternAlphabetCount.setContributingCombinedPatternAlphabetPatternSetMap(entireLogCombinedPatternAlphabetPatternSetMap);
		overlappingPatternAlphabetCount.setContributingCombinedBasePatternAlphabetPatternSetMap(entireLogCombinedBasePatternAlphabetPatternSetMap);
		
		overlappingPatternAlphabetCount.setTandemArrayPatternAlphabetCountMap(tandemArrayPatternAlphabetCountMap);
		overlappingPatternAlphabetCount.setTandemArrayPatternAlphabetNonDuplicateCountMap(tandemArrayPatternAlphabetNonDuplicateCountMap);
		overlappingPatternAlphabetCount.setTandemArrayBasePatternAlphabetCountMap(tandemArrayBasePatternAlphabetCountMap);
		overlappingPatternAlphabetCount.setTandemArrayBasePatternAlphabetNonDuplicateCountMap(tandemArrayBasePatternAlphabetNonDuplicateCountMap);
			
		overlappingPatternAlphabetCount.setTandemArrayPatternAlphabetInstanceCountMap(tandemArrayPatternAlphabetInstanceCountMap);
		overlappingPatternAlphabetCount.setTandemArrayPatternAlphabetNonDuplicateInstanceCountMap(tandemArrayPatternAlphabetNonDuplicateInstanceCountMap);
		overlappingPatternAlphabetCount.setTandemArrayBasePatternAlphabetInstanceCountMap(tandemArrayBasePatternAlphabetInstanceCountMap);
		overlappingPatternAlphabetCount.setTandemArrayBasePatternAlphabetNonDuplicateInstanceCountMap(tandemArrayBasePatternAlphabetNonDuplicateInstanceCountMap);
		
		overlappingPatternAlphabetCount.setMaximalRepeatPatternAlphabetCountMap(maximalRepeatPatternAlphabetCountMap);
		overlappingPatternAlphabetCount.setMaximalRepeatPatternAlphabetNonDuplicateCountMap(maximalRepeatPatternAlphabetNonDuplicateCountMap);
		overlappingPatternAlphabetCount.setMaximalRepeatBasePatternAlphabetCountMap(maximalRepeatBasePatternAlphabetCountMap);
		overlappingPatternAlphabetCount.setMaximalRepeatBasePatternAlphabetNonDuplicateCountMap(maximalRepeatBasePatternAlphabetNonDuplicateCountMap);

		overlappingPatternAlphabetCount.setMaximalRepeatPatternAlphabetInstanceCountMap(maximalRepeatPatternAlphabetInstanceCountMap);
		overlappingPatternAlphabetCount.setMaximalRepeatPatternAlphabetNonDuplicateInstanceCountMap(maximalRepeatPatternAlphabetNonDuplicateInstanceCountMap);
		overlappingPatternAlphabetCount.setMaximalRepeatBasePatternAlphabetInstanceCountMap(maximalRepeatBasePatternAlphabetInstanceCountMap);
		overlappingPatternAlphabetCount.setMaximalRepeatBasePatternAlphabetNonDuplicateInstanceCountMap(maximalRepeatBasePatternAlphabetNonDuplicateInstanceCountMap);

		overlappingPatternAlphabetCount.setCombinedPatternAlphabetCountMap(combinedPatternAlphabetCountMap);
		overlappingPatternAlphabetCount.setCombinedPatternAlphabetNonDuplicateCountMap(combinedPatternAlphabetNonDuplicateCountMap);
		overlappingPatternAlphabetCount.setCombinedBasePatternAlphabetCountMap(combinedBasePatternAlphabetCountMap);
		overlappingPatternAlphabetCount.setCombinedBasePatternAlphabetNonDuplicateCountMap(combinedBasePatternAlphabetNonDuplicateCountMap);
	
		overlappingPatternAlphabetCount.setCombinedPatternAlphabetInstanceCountMap(combinedPatternAlphabetInstanceCountMap);
		overlappingPatternAlphabetCount.setCombinedPatternAlphabetNonDuplicateInstanceCountMap(combinedPatternAlphabetNonDuplicateInstanceCountMap);
		overlappingPatternAlphabetCount.setCombinedBasePatternAlphabetInstanceCountMap(combinedBasePatternAlphabetInstanceCountMap);
		overlappingPatternAlphabetCount.setCombinedBasePatternAlphabetNonDuplicateInstanceCountMap(combinedBasePatternAlphabetNonDuplicateInstanceCountMap);

		Map<Set<String>, Float> patternAlphabetOverlappingConservednessValueMap = new HashMap<Set<String>, Float>();
		float conservedness;
		float mean, stddev;
		for(Set<String> combinedPatternAlphabet : combinedPatternAlphabetCountMap.keySet()){
			conservedness = 
			mean = 1; stddev = 0;
			if(patternAlphabetMeanCountMap.containsKey(combinedPatternAlphabet))
				mean = patternAlphabetMeanCountMap.get(combinedPatternAlphabet);
			if(patternAlphabetStdDevCountMap.containsKey(combinedPatternAlphabet))
				stddev = patternAlphabetStdDevCountMap.get(combinedPatternAlphabet);
			conservedness = (combinedPatternAlphabetCountMap.get(combinedPatternAlphabet)/mean) * (1-stddev/mean)*100;
			patternAlphabetOverlappingConservednessValueMap.put(combinedPatternAlphabet, conservedness);
			if(conservedness < OVERLAPPINGCOUNT_CONSERVEDNESS_THRESHOLD)
				ignorablePatternAlphabetSet.add(combinedPatternAlphabet);
		}
	}
	
	private void computeNOACMetrics(){
		System.out.println("In computeNOACMetrics()");
		nonOverlappingShorterPatternAlphabetCount = new MetricPatternAlphabetCount();
		nonOverlappingLongerPatternAlphabetCount = new MetricPatternAlphabetCount();
		
		Set<String> contributingTandemArrayShorterPatternSet = new HashSet<String>();
		Set<String> contributingMaximalRepeatShorterPatternSet = new HashSet<String>();
		Set<String> contributingCombinedShorterPatternSet = new HashSet<String>();
		
		Set<String> contributingTandemArrayShorterBasePatternSet = new HashSet<String>();
		Set<String> contributingMaximalRepeatShorterBasePatternSet = new HashSet<String>();
		Set<String> contributingCombinedShorterBasePatternSet = new HashSet<String>();
		
		Map<TreeSet<String>, Integer> tandemArrayShorterPatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayShorterBasePatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayShorterPatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayShorterBasePatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		Map<TreeSet<String>, Integer> maximalRepeatShorterPatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatShorterBasePatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatShorterPatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatShorterBasePatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		Map<TreeSet<String>, Integer> combinedShorterPatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedShorterBasePatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedShorterPatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedShorterBasePatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		Map<TreeSet<String>, Integer> tandemArrayShorterPatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayShorterBasePatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayShorterPatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayShorterBasePatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		Map<TreeSet<String>, Integer> maximalRepeatShorterPatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatShorterBasePatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatShorterPatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatShorterBasePatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		Map<TreeSet<String>, Integer> combinedShorterPatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedShorterBasePatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedShorterPatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedShorterBasePatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();

		Set<String> contributingTandemArrayLongerPatternSet = new HashSet<String>();
		Set<String> contributingMaximalRepeatLongerPatternSet = new HashSet<String>();
		Set<String> contributingCombinedLongerPatternSet = new HashSet<String>();
		
		Set<String> contributingTandemArrayLongerBasePatternSet = new HashSet<String>();
		Set<String> contributingMaximalRepeatLongerBasePatternSet = new HashSet<String>();
		Set<String> contributingCombinedLongerBasePatternSet = new HashSet<String>();
		
		
		Map<TreeSet<String>, Integer> tandemArrayLongerPatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayLongerBasePatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayLongerPatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayLongerBasePatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		Map<TreeSet<String>, Integer> maximalRepeatLongerPatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatLongerBasePatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatLongerPatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatLongerBasePatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		Map<TreeSet<String>, Integer> combinedLongerPatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedLongerBasePatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedLongerPatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedLongerBasePatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		Map<TreeSet<String>, Integer> tandemArrayLongerPatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayLongerBasePatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayLongerPatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayLongerBasePatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		Map<TreeSet<String>, Integer> maximalRepeatLongerPatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatLongerBasePatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatLongerPatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatLongerBasePatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		Map<TreeSet<String>, Integer> combinedLongerPatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedLongerBasePatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedLongerPatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedLongerBasePatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		int encodingLength = encodedLog.getEncodingLength();
		List<String> uniqueEncodedTraceList = encodedLog.getUniqueEncodedTraceList();
		Map<String, TreeSet<Integer>> encodedTraceIdenticalTraceIndicesSetMap = encodedLog.getEncodedTraceIdenticalTraceIndicesSetMap();
		Set<String> combinedPatternAlphabetPatternSet, combinedBasePatternAlphabetPatternSet;
		
		UkkonenSuffixTree suffixTree;
		int[] indexPos;
		int noIndices, noIdenticalTraces;
		boolean isPatternAlphabetExists, isBasePatternAlphabetExists;
		
		Map<Integer, SortedSet<String>> indexPatternSetMap = new HashMap<Integer, SortedSet<String>>();
		Map<Integer, SortedSet<String>> indexBasePatternSetMap = new HashMap<Integer, SortedSet<String>>();
		
		SortedSet<String> indexPatternSet;
		SortedSet<String> indexBasePatternSet;
		
		SortedSet<Integer> patternIndicesSet = new TreeSet<Integer>();

		SortedSet<Integer> nonOverlappingShorterPatternIndicesSet = new TreeSet<Integer>();
		Map<Integer, String> nonOverlappingPatternIndexContributingShorterPatternMap = new HashMap<Integer, String>();
		
		SortedSet<Integer> nonOverlappingLongerPatternIndicesSet = new TreeSet<Integer>();
		Map<Integer, String> nonOverlappingPatternIndexContributingLongerPatternMap = new HashMap<Integer, String>();

		
		SortedSet<Integer> basePatternIndicesSet = new TreeSet<Integer>();
		SortedSet<Integer> nonOverlappingShorterBasePatternIndicesSet = new TreeSet<Integer>();
		Map<Integer, String> nonOverlappingBasePatternIndexContributingShorterPatternMap = new HashMap<Integer, String>();
		
		SortedSet<Integer> nonOverlappingLongerBasePatternIndicesSet = new TreeSet<Integer>();
		Map<Integer, String> nonOverlappingBasePatternIndexContributingLongerPatternMap = new HashMap<Integer, String>();
		
		
		Set<String> tandemArrayPatternAlphabetPatternSet, maximalRepeatPatternAlphabetPatternSet;
		Set<String> tandemArrayBasePatternAlphabetPatternSet, maximalRepeatBasePatternAlphabetPatternSet;
		
		for(String encodedTrace : uniqueEncodedTraceList){
			noIdenticalTraces = encodedTraceIdenticalTraceIndicesSetMap.get(encodedTrace).size();

			/*
			 * Build the suffix tree for this trace
			 */
			suffixTree = new UkkonenSuffixTree(encodingLength, encodedTrace);

			for(TreeSet<String> combinedPatternAlphabet : entireLogCombinedPatternAlphabetPatternSetMap.keySet()){
				if(ignorablePatternAlphabetSet != null && ignorablePatternAlphabetSet.contains(combinedPatternAlphabet))
					continue;
				
				
				isPatternAlphabetExists = false;

				indexPatternSetMap.clear();
				combinedPatternAlphabetPatternSet = entireLogCombinedPatternAlphabetPatternSetMap.get(combinedPatternAlphabet);
				
				for(String pattern : combinedPatternAlphabetPatternSet){
					if(suffixTree.noMatches(pattern)<=0)
						continue;
				
					indexPos = suffixTree.getMatches(pattern);
					if(indexPos == null)
						continue;
					isPatternAlphabetExists = true;
					noIndices = indexPos.length;
					
					for(int i = 0; i < noIndices; i++){
						if(indexPatternSetMap.containsKey(indexPos[i])){
							indexPatternSet = indexPatternSetMap.get(indexPos[i]);
						}else{
							indexPatternSet = new TreeSet<String>(new AscendingStringLengthComparator());
						}
						indexPatternSet.add(pattern);
						indexPatternSetMap.put(indexPos[i], indexPatternSet);
					}
				}
				if(isPatternAlphabetExists){
					/*
					 * Update the counts
					 */
					
					patternIndicesSet.clear();
					patternIndicesSet.addAll(indexPatternSetMap.keySet());
					
					nonOverlappingShorterPatternIndicesSet.clear();
					nonOverlappingPatternIndexContributingShorterPatternMap.clear();
					
					//Doing it for shorter patterns
					int nextShorterPatternIndex = -1;
					String contributingShorterPattern;
					int contributingShorterPatternLength;
					for(Integer patternIndex : patternIndicesSet){
						if(patternIndex < nextShorterPatternIndex)
							continue;
						
						nonOverlappingShorterPatternIndicesSet.add(patternIndex);
						
						contributingShorterPattern = indexPatternSetMap.get(patternIndex).first();
						contributingCombinedShorterPatternSet.add(contributingShorterPattern);
						nonOverlappingPatternIndexContributingShorterPatternMap.put(patternIndex, contributingShorterPattern);
						
						contributingShorterPatternLength = contributingShorterPattern.length()/encodingLength;
						nextShorterPatternIndex = patternIndex + contributingShorterPatternLength;
						
					}
					
					int count = nonOverlappingPatternIndexContributingShorterPatternMap.size()*noIdenticalTraces;
					if(combinedShorterPatternAlphabetCountMap.containsKey(combinedPatternAlphabet)){
						count += combinedShorterPatternAlphabetCountMap.get(combinedPatternAlphabet);
					}
					combinedShorterPatternAlphabetCountMap.put(combinedPatternAlphabet, count);
					
					count = nonOverlappingPatternIndexContributingShorterPatternMap.size();
					if(combinedShorterPatternAlphabetNonDuplicateCountMap.containsKey(combinedPatternAlphabet)){
						count += combinedShorterPatternAlphabetNonDuplicateCountMap.get(combinedPatternAlphabet);
					}
					combinedShorterPatternAlphabetNonDuplicateCountMap.put(combinedPatternAlphabet, count);
					
					int instanceCount = noIdenticalTraces;
					if(combinedShorterPatternAlphabetInstanceCountMap.containsKey(combinedPatternAlphabet))
						instanceCount += combinedShorterPatternAlphabetInstanceCountMap.get(combinedPatternAlphabet);
					combinedShorterPatternAlphabetInstanceCountMap.put(combinedPatternAlphabet, instanceCount);
					
					instanceCount = 1;
					if(combinedShorterPatternAlphabetNonDuplicateInstanceCountMap.containsKey(combinedPatternAlphabet))
						instanceCount += combinedShorterPatternAlphabetNonDuplicateInstanceCountMap.get(combinedPatternAlphabet);
					combinedShorterPatternAlphabetNonDuplicateInstanceCountMap.put(combinedPatternAlphabet, instanceCount);

					//Check if this pattern alphabet is also a tandem array and/or maximal repeat
					Set<Integer> tandemArrayShorterPatternIndicesSet = new HashSet<Integer>();
					Set<Integer> maximalRepeatShorterPatternIndicesSet = new HashSet<Integer>();
					if(entireLogTandemArrayPatternAlphabetPatternSetMap.containsKey(combinedPatternAlphabet)){
						tandemArrayPatternAlphabetPatternSet = entireLogTandemArrayPatternAlphabetPatternSetMap.get(combinedPatternAlphabet);
						for(Integer nonOverlappingShorterPatternIndex : nonOverlappingPatternIndexContributingShorterPatternMap.keySet()){
							if(tandemArrayPatternAlphabetPatternSet.contains(nonOverlappingPatternIndexContributingShorterPatternMap.get(nonOverlappingShorterPatternIndex))){
								contributingTandemArrayShorterPatternSet.add(nonOverlappingPatternIndexContributingShorterPatternMap.get(nonOverlappingShorterPatternIndex));
								tandemArrayShorterPatternIndicesSet.add(nonOverlappingShorterPatternIndex);
							}
						}
						
						if(tandemArrayShorterPatternIndicesSet.size() > 0){
							count = tandemArrayShorterPatternIndicesSet.size()*noIdenticalTraces;
							if(tandemArrayShorterPatternAlphabetCountMap.containsKey(combinedPatternAlphabet)){
								count += tandemArrayShorterPatternAlphabetCountMap.get(combinedPatternAlphabet);
							}
							tandemArrayShorterPatternAlphabetCountMap.put(combinedPatternAlphabet, count);
							
							count = tandemArrayShorterPatternIndicesSet.size();
							if(tandemArrayShorterPatternAlphabetNonDuplicateCountMap.containsKey(combinedPatternAlphabet)){
								count += tandemArrayShorterPatternAlphabetNonDuplicateCountMap.get(combinedPatternAlphabet);
							}
							tandemArrayShorterPatternAlphabetNonDuplicateCountMap.put(combinedPatternAlphabet, count);
							
							instanceCount = noIdenticalTraces;
							if(tandemArrayShorterPatternAlphabetInstanceCountMap.containsKey(combinedPatternAlphabet))
								instanceCount += tandemArrayShorterPatternAlphabetInstanceCountMap.get(combinedPatternAlphabet);
							tandemArrayShorterPatternAlphabetInstanceCountMap.put(combinedPatternAlphabet, instanceCount);
							
							instanceCount = 1;
							if(tandemArrayShorterPatternAlphabetNonDuplicateInstanceCountMap.containsKey(combinedPatternAlphabet))
								instanceCount += tandemArrayShorterPatternAlphabetNonDuplicateInstanceCountMap.get(combinedPatternAlphabet);
							tandemArrayShorterPatternAlphabetNonDuplicateInstanceCountMap.put(combinedPatternAlphabet, instanceCount);
						}
					}
					
					if(entireLogMaximalRepeatPatternAlphabetPatternSetMap.containsKey(combinedPatternAlphabet)){
						maximalRepeatPatternAlphabetPatternSet = entireLogMaximalRepeatPatternAlphabetPatternSetMap.get(combinedPatternAlphabet);
						
						for(Integer nonOverlappingShorterPatternIndex : nonOverlappingPatternIndexContributingShorterPatternMap.keySet()){
							if(maximalRepeatPatternAlphabetPatternSet.contains(nonOverlappingPatternIndexContributingShorterPatternMap.get(nonOverlappingShorterPatternIndex))){
								contributingMaximalRepeatShorterPatternSet.add(nonOverlappingPatternIndexContributingShorterPatternMap.get(nonOverlappingShorterPatternIndex));
								maximalRepeatShorterPatternIndicesSet.add(nonOverlappingShorterPatternIndex);
							}
						}
						if(maximalRepeatShorterPatternIndicesSet.size() > 0){
							count = maximalRepeatShorterPatternIndicesSet.size()*noIdenticalTraces;
							if(maximalRepeatShorterPatternAlphabetCountMap.containsKey(combinedPatternAlphabet)){
								count += maximalRepeatShorterPatternAlphabetCountMap.get(combinedPatternAlphabet);
							}
							maximalRepeatShorterPatternAlphabetCountMap.put(combinedPatternAlphabet, count);
							
							count = maximalRepeatShorterPatternIndicesSet.size();
							if(maximalRepeatShorterPatternAlphabetNonDuplicateCountMap.containsKey(combinedPatternAlphabet)){
								count += maximalRepeatShorterPatternAlphabetNonDuplicateCountMap.get(combinedPatternAlphabet);
							}
							maximalRepeatShorterPatternAlphabetNonDuplicateCountMap.put(combinedPatternAlphabet, count);
							
							instanceCount = noIdenticalTraces;
							if(maximalRepeatShorterPatternAlphabetInstanceCountMap.containsKey(combinedPatternAlphabet))
								instanceCount += maximalRepeatShorterPatternAlphabetInstanceCountMap.get(combinedPatternAlphabet);
							maximalRepeatShorterPatternAlphabetInstanceCountMap.put(combinedPatternAlphabet, instanceCount);
							
							instanceCount = 1;
							if(maximalRepeatShorterPatternAlphabetNonDuplicateInstanceCountMap.containsKey(combinedPatternAlphabet))
								instanceCount += maximalRepeatShorterPatternAlphabetNonDuplicateInstanceCountMap.get(combinedPatternAlphabet);
							maximalRepeatShorterPatternAlphabetNonDuplicateInstanceCountMap.put(combinedPatternAlphabet, instanceCount);
						}
					}
					
					//	Doing it for longer patterns
					nonOverlappingLongerPatternIndicesSet.clear();
					nonOverlappingPatternIndexContributingLongerPatternMap.clear();
					
					int nextLongerPatternIndex = -1;

					String contributingLongerPattern;
					int contributingLongerPatternLength;
					
					for(Integer patternIndex : patternIndicesSet){
						if(patternIndex < nextLongerPatternIndex)
							continue;
						
						nonOverlappingLongerPatternIndicesSet.add(patternIndex);
						
						contributingLongerPattern = indexPatternSetMap.get(patternIndex).last();
						contributingCombinedLongerPatternSet.add(contributingLongerPattern);
						nonOverlappingPatternIndexContributingLongerPatternMap.put(patternIndex, contributingLongerPattern);
						
						contributingLongerPatternLength = contributingLongerPattern.length()/encodingLength;
						nextLongerPatternIndex = patternIndex + contributingLongerPatternLength;
						
					}
					
					count = nonOverlappingPatternIndexContributingLongerPatternMap.size()*noIdenticalTraces;
					if(combinedLongerPatternAlphabetCountMap.containsKey(combinedPatternAlphabet)){
						count += combinedLongerPatternAlphabetCountMap.get(combinedPatternAlphabet);
					}
					combinedLongerPatternAlphabetCountMap.put(combinedPatternAlphabet, count);
					
					count = nonOverlappingPatternIndexContributingLongerPatternMap.size();
					if(combinedLongerPatternAlphabetNonDuplicateCountMap.containsKey(combinedPatternAlphabet)){
						count += combinedLongerPatternAlphabetNonDuplicateCountMap.get(combinedPatternAlphabet);
					}
					combinedLongerPatternAlphabetNonDuplicateCountMap.put(combinedPatternAlphabet, count);
					
					instanceCount = noIdenticalTraces;
					if(combinedLongerPatternAlphabetInstanceCountMap.containsKey(combinedPatternAlphabet))
						instanceCount += combinedLongerPatternAlphabetInstanceCountMap.get(combinedPatternAlphabet);
					combinedLongerPatternAlphabetInstanceCountMap.put(combinedPatternAlphabet, instanceCount);
					
					instanceCount = 1;
					if(combinedLongerPatternAlphabetNonDuplicateInstanceCountMap.containsKey(combinedPatternAlphabet))
						instanceCount += combinedLongerPatternAlphabetNonDuplicateInstanceCountMap.get(combinedPatternAlphabet);
					combinedLongerPatternAlphabetNonDuplicateInstanceCountMap.put(combinedPatternAlphabet, instanceCount);

					//Check if this pattern alphabet is also a tandem array and/or maximal repeat
					Set<Integer> tandemArrayLongerPatternIndicesSet = new HashSet<Integer>();
					Set<Integer> maximalRepeatLongerPatternIndicesSet = new HashSet<Integer>();
					if(entireLogTandemArrayPatternAlphabetPatternSetMap.containsKey(combinedPatternAlphabet)){
						tandemArrayPatternAlphabetPatternSet = entireLogTandemArrayPatternAlphabetPatternSetMap.get(combinedPatternAlphabet);
						for(Integer nonOverlappingLongerPatternIndex : nonOverlappingPatternIndexContributingLongerPatternMap.keySet()){
							if(tandemArrayPatternAlphabetPatternSet.contains(nonOverlappingPatternIndexContributingLongerPatternMap.get(nonOverlappingLongerPatternIndex))){
								contributingTandemArrayLongerPatternSet.add(nonOverlappingPatternIndexContributingLongerPatternMap.get(nonOverlappingLongerPatternIndex));
								tandemArrayLongerPatternIndicesSet.add(nonOverlappingLongerPatternIndex);
							}
						}
						
						if(tandemArrayLongerPatternIndicesSet.size() > 0){
//							System.out.println("TAPIS: "+tandemArrayLongerPatternIndicesSet);
							count = tandemArrayLongerPatternIndicesSet.size()*noIdenticalTraces;
							if(tandemArrayLongerPatternAlphabetCountMap.containsKey(combinedPatternAlphabet)){
								count += tandemArrayLongerPatternAlphabetCountMap.get(combinedPatternAlphabet);
							}
							tandemArrayLongerPatternAlphabetCountMap.put(combinedPatternAlphabet, count);
							
							count = tandemArrayLongerPatternIndicesSet.size();
							if(tandemArrayLongerPatternAlphabetNonDuplicateCountMap.containsKey(combinedPatternAlphabet)){
								count += tandemArrayLongerPatternAlphabetNonDuplicateCountMap.get(combinedPatternAlphabet);
							}
							tandemArrayLongerPatternAlphabetNonDuplicateCountMap.put(combinedPatternAlphabet, count);
							
							instanceCount = noIdenticalTraces;
							if(tandemArrayLongerPatternAlphabetInstanceCountMap.containsKey(combinedPatternAlphabet))
								instanceCount += tandemArrayLongerPatternAlphabetInstanceCountMap.get(combinedPatternAlphabet);
							tandemArrayLongerPatternAlphabetInstanceCountMap.put(combinedPatternAlphabet, instanceCount);
							
							instanceCount = 1;
							if(tandemArrayLongerPatternAlphabetNonDuplicateInstanceCountMap.containsKey(combinedPatternAlphabet))
								instanceCount += tandemArrayLongerPatternAlphabetNonDuplicateInstanceCountMap.get(combinedPatternAlphabet);
							tandemArrayLongerPatternAlphabetNonDuplicateInstanceCountMap.put(combinedPatternAlphabet, instanceCount);
						}
					}
					
					if(entireLogMaximalRepeatPatternAlphabetPatternSetMap.containsKey(combinedPatternAlphabet)){
						maximalRepeatPatternAlphabetPatternSet = entireLogMaximalRepeatPatternAlphabetPatternSetMap.get(combinedPatternAlphabet);
						
						for(Integer nonOverlappingLongerPatternIndex : nonOverlappingPatternIndexContributingLongerPatternMap.keySet()){
							if(maximalRepeatPatternAlphabetPatternSet.contains(nonOverlappingPatternIndexContributingLongerPatternMap.get(nonOverlappingLongerPatternIndex))){
								contributingMaximalRepeatLongerPatternSet.add(nonOverlappingPatternIndexContributingLongerPatternMap.get(nonOverlappingLongerPatternIndex));
								maximalRepeatLongerPatternIndicesSet.add(nonOverlappingLongerPatternIndex);
							}
						}
						if(maximalRepeatLongerPatternIndicesSet.size() > 0){
							count = maximalRepeatLongerPatternIndicesSet.size()*noIdenticalTraces;
							if(maximalRepeatLongerPatternAlphabetCountMap.containsKey(combinedPatternAlphabet)){
								count += maximalRepeatLongerPatternAlphabetCountMap.get(combinedPatternAlphabet);
							}
							maximalRepeatLongerPatternAlphabetCountMap.put(combinedPatternAlphabet, count);
							
							count = maximalRepeatLongerPatternIndicesSet.size();
							if(maximalRepeatLongerPatternAlphabetNonDuplicateCountMap.containsKey(combinedPatternAlphabet)){
								count += maximalRepeatLongerPatternAlphabetNonDuplicateCountMap.get(combinedPatternAlphabet);
							}
							maximalRepeatLongerPatternAlphabetNonDuplicateCountMap.put(combinedPatternAlphabet, count);
							
							instanceCount = noIdenticalTraces;
							if(maximalRepeatLongerPatternAlphabetInstanceCountMap.containsKey(combinedPatternAlphabet))
								instanceCount += maximalRepeatLongerPatternAlphabetInstanceCountMap.get(combinedPatternAlphabet);
							maximalRepeatLongerPatternAlphabetInstanceCountMap.put(combinedPatternAlphabet, instanceCount);
							
							instanceCount = 1;
							if(maximalRepeatLongerPatternAlphabetNonDuplicateInstanceCountMap.containsKey(combinedPatternAlphabet))
								instanceCount += maximalRepeatLongerPatternAlphabetNonDuplicateInstanceCountMap.get(combinedPatternAlphabet);
							maximalRepeatLongerPatternAlphabetNonDuplicateInstanceCountMap.put(combinedPatternAlphabet, instanceCount);
						}
					}
				}
			}
			
			/*
			 * Do it for base patterns now
			 */
			for(TreeSet<String> combinedBasePatternAlphabet : entireLogCombinedBasePatternAlphabetPatternSetMap.keySet()){
				if(ignorablePatternAlphabetSet != null && ignorablePatternAlphabetSet.contains(combinedBasePatternAlphabet))
					continue;
				
				
				isBasePatternAlphabetExists = false;

				indexBasePatternSetMap.clear();
				combinedBasePatternAlphabetPatternSet = entireLogCombinedBasePatternAlphabetPatternSetMap.get(combinedBasePatternAlphabet);
				
				for(String basePattern : combinedBasePatternAlphabetPatternSet){
					if(suffixTree.noMatches(basePattern)<=0)
						continue;
				
					indexPos = suffixTree.getMatches(basePattern);
					if(indexPos == null)
						continue;
					isBasePatternAlphabetExists = true;
					noIndices = indexPos.length;
					
					for(int i = 0; i < noIndices; i++){
						if(indexBasePatternSetMap.containsKey(indexPos[i])){
							indexBasePatternSet = indexBasePatternSetMap.get(indexPos[i]);
						}else{
							indexBasePatternSet = new TreeSet<String>(new AscendingStringLengthComparator());
						}
						indexBasePatternSet.add(basePattern);
						indexBasePatternSetMap.put(indexPos[i], indexBasePatternSet);
					}
				}
				if(isBasePatternAlphabetExists){
					/*
					 * Update the counts
					 */
					
					basePatternIndicesSet.clear();
					basePatternIndicesSet.addAll(indexBasePatternSetMap.keySet());
					
					//Doing it for shorter patterns
					
					nonOverlappingShorterBasePatternIndicesSet.clear();
					nonOverlappingBasePatternIndexContributingShorterPatternMap.clear();
					
					int nextShorterPatternIndex = -1;
					String contributingShorterBasePattern;
					int contributingShorterBasePatternLength;
					for(Integer patternIndex : basePatternIndicesSet){
						if(patternIndex < nextShorterPatternIndex)
							continue;
						
						nonOverlappingShorterBasePatternIndicesSet.add(patternIndex);
						
						contributingShorterBasePattern = indexBasePatternSetMap.get(patternIndex).first();
						contributingCombinedShorterBasePatternSet.add(contributingShorterBasePattern);
						nonOverlappingBasePatternIndexContributingShorterPatternMap.put(patternIndex, contributingShorterBasePattern);
						
						contributingShorterBasePatternLength = contributingShorterBasePattern.length()/encodingLength;
						nextShorterPatternIndex = patternIndex + contributingShorterBasePatternLength;
						
					}
					
					int count = nonOverlappingBasePatternIndexContributingShorterPatternMap.size()*noIdenticalTraces;
					if(combinedShorterBasePatternAlphabetCountMap.containsKey(combinedBasePatternAlphabet)){
						count += combinedShorterBasePatternAlphabetCountMap.get(combinedBasePatternAlphabet);
					}
					combinedShorterBasePatternAlphabetCountMap.put(combinedBasePatternAlphabet, count);
					
					count = nonOverlappingBasePatternIndexContributingShorterPatternMap.size();
					if(combinedShorterBasePatternAlphabetNonDuplicateCountMap.containsKey(combinedBasePatternAlphabet)){
						count += combinedShorterBasePatternAlphabetNonDuplicateCountMap.get(combinedBasePatternAlphabet);
					}
					combinedShorterBasePatternAlphabetNonDuplicateCountMap.put(combinedBasePatternAlphabet, count);
					
					int instanceCount = noIdenticalTraces;
					if(combinedShorterBasePatternAlphabetInstanceCountMap.containsKey(combinedBasePatternAlphabet))
						instanceCount += combinedShorterBasePatternAlphabetInstanceCountMap.get(combinedBasePatternAlphabet);
					combinedShorterBasePatternAlphabetInstanceCountMap.put(combinedBasePatternAlphabet, instanceCount);
					
					instanceCount = 1;
					if(combinedShorterBasePatternAlphabetNonDuplicateInstanceCountMap.containsKey(combinedBasePatternAlphabet))
						instanceCount += combinedShorterBasePatternAlphabetNonDuplicateInstanceCountMap.get(combinedBasePatternAlphabet);
					combinedShorterBasePatternAlphabetNonDuplicateInstanceCountMap.put(combinedBasePatternAlphabet, instanceCount);

					//Check if this basePattern alphabet is also a tandem array and/or maximal repeat
					Set<Integer> tandemArrayShorterBasePatternIndicesSet = new HashSet<Integer>();
					Set<Integer> maximalRepeatShorterBasePatternIndicesSet = new HashSet<Integer>();
					if(entireLogTandemArrayBasePatternAlphabetPatternSetMap.containsKey(combinedBasePatternAlphabet)){
						tandemArrayBasePatternAlphabetPatternSet = entireLogTandemArrayBasePatternAlphabetPatternSetMap.get(combinedBasePatternAlphabet);
						for(Integer nonOverlappingShorterPatternIndex : nonOverlappingBasePatternIndexContributingShorterPatternMap.keySet()){
							if(tandemArrayBasePatternAlphabetPatternSet.contains(nonOverlappingBasePatternIndexContributingShorterPatternMap.get(nonOverlappingShorterPatternIndex))){
								contributingTandemArrayShorterBasePatternSet.add(nonOverlappingBasePatternIndexContributingShorterPatternMap.get(nonOverlappingShorterPatternIndex));
								tandemArrayShorterBasePatternIndicesSet.add(nonOverlappingShorterPatternIndex);
							}
						}
						
						if(tandemArrayShorterBasePatternIndicesSet.size() > 0){
							count = tandemArrayShorterBasePatternIndicesSet.size()*noIdenticalTraces;
							if(tandemArrayShorterBasePatternAlphabetCountMap.containsKey(combinedBasePatternAlphabet)){
								count += tandemArrayShorterBasePatternAlphabetCountMap.get(combinedBasePatternAlphabet);
							}
							tandemArrayShorterBasePatternAlphabetCountMap.put(combinedBasePatternAlphabet, count);
							
							count = tandemArrayShorterBasePatternIndicesSet.size();
							if(tandemArrayShorterBasePatternAlphabetNonDuplicateCountMap.containsKey(combinedBasePatternAlphabet)){
								count += tandemArrayShorterBasePatternAlphabetNonDuplicateCountMap.get(combinedBasePatternAlphabet);
							}
							tandemArrayShorterBasePatternAlphabetNonDuplicateCountMap.put(combinedBasePatternAlphabet, count);
							
							instanceCount = noIdenticalTraces;
							if(tandemArrayShorterBasePatternAlphabetInstanceCountMap.containsKey(combinedBasePatternAlphabet))
								instanceCount += tandemArrayShorterBasePatternAlphabetInstanceCountMap.get(combinedBasePatternAlphabet);
							tandemArrayShorterBasePatternAlphabetInstanceCountMap.put(combinedBasePatternAlphabet, instanceCount);
							
							instanceCount = 1;
							if(tandemArrayShorterBasePatternAlphabetNonDuplicateInstanceCountMap.containsKey(combinedBasePatternAlphabet))
								instanceCount += tandemArrayShorterBasePatternAlphabetNonDuplicateInstanceCountMap.get(combinedBasePatternAlphabet);
							tandemArrayShorterBasePatternAlphabetNonDuplicateInstanceCountMap.put(combinedBasePatternAlphabet, instanceCount);
						}
					}
					
					if(entireLogMaximalRepeatBasePatternAlphabetPatternSetMap.containsKey(combinedBasePatternAlphabet)){
						maximalRepeatBasePatternAlphabetPatternSet = entireLogMaximalRepeatBasePatternAlphabetPatternSetMap.get(combinedBasePatternAlphabet);
						
						for(Integer nonOverlappingShorterPatternIndex : nonOverlappingBasePatternIndexContributingShorterPatternMap.keySet()){
							if(maximalRepeatBasePatternAlphabetPatternSet.contains(nonOverlappingBasePatternIndexContributingShorterPatternMap.get(nonOverlappingShorterPatternIndex))){
								contributingMaximalRepeatShorterBasePatternSet.add(nonOverlappingBasePatternIndexContributingShorterPatternMap.get(nonOverlappingShorterPatternIndex));
								maximalRepeatShorterBasePatternIndicesSet.add(nonOverlappingShorterPatternIndex);
							}
						}
						if(maximalRepeatShorterBasePatternIndicesSet.size() > 0){
							count = maximalRepeatShorterBasePatternIndicesSet.size()*noIdenticalTraces;
							if(maximalRepeatShorterBasePatternAlphabetCountMap.containsKey(combinedBasePatternAlphabet)){
								count += maximalRepeatShorterBasePatternAlphabetCountMap.get(combinedBasePatternAlphabet);
							}
							maximalRepeatShorterBasePatternAlphabetCountMap.put(combinedBasePatternAlphabet, count);
							
							count = maximalRepeatShorterBasePatternIndicesSet.size();
							if(maximalRepeatShorterBasePatternAlphabetNonDuplicateCountMap.containsKey(combinedBasePatternAlphabet)){
								count += maximalRepeatShorterBasePatternAlphabetNonDuplicateCountMap.get(combinedBasePatternAlphabet);
							}
							maximalRepeatShorterBasePatternAlphabetNonDuplicateCountMap.put(combinedBasePatternAlphabet, count);
							
							instanceCount = noIdenticalTraces;
							if(maximalRepeatShorterBasePatternAlphabetInstanceCountMap.containsKey(combinedBasePatternAlphabet))
								instanceCount += maximalRepeatShorterBasePatternAlphabetInstanceCountMap.get(combinedBasePatternAlphabet);
							maximalRepeatShorterBasePatternAlphabetInstanceCountMap.put(combinedBasePatternAlphabet, instanceCount);
							
							instanceCount = 1;
							if(maximalRepeatShorterBasePatternAlphabetNonDuplicateInstanceCountMap.containsKey(combinedBasePatternAlphabet))
								instanceCount += maximalRepeatShorterBasePatternAlphabetNonDuplicateInstanceCountMap.get(combinedBasePatternAlphabet);
							maximalRepeatShorterBasePatternAlphabetNonDuplicateInstanceCountMap.put(combinedBasePatternAlphabet, instanceCount);
						}
					}
					
					// Doing it for longer patterns
					nonOverlappingLongerBasePatternIndicesSet.clear();
					nonOverlappingBasePatternIndexContributingLongerPatternMap.clear();
					
					int nextLongerPatternIndex = -1;
					String contributingLongerBasePattern;
					int contributingLongerBasePatternLength;
					for(Integer patternIndex : basePatternIndicesSet){
						if(patternIndex < nextLongerPatternIndex)
							continue;
						
						nonOverlappingLongerBasePatternIndicesSet.add(patternIndex);
						
						contributingLongerBasePattern = indexBasePatternSetMap.get(patternIndex).first();
						contributingCombinedLongerBasePatternSet.add(contributingLongerBasePattern);
						nonOverlappingBasePatternIndexContributingLongerPatternMap.put(patternIndex, contributingLongerBasePattern);
						
						contributingLongerBasePatternLength = contributingLongerBasePattern.length()/encodingLength;
						nextLongerPatternIndex = patternIndex + contributingLongerBasePatternLength;
						
					}
					
					count = nonOverlappingBasePatternIndexContributingLongerPatternMap.size()*noIdenticalTraces;
					if(combinedLongerBasePatternAlphabetCountMap.containsKey(combinedBasePatternAlphabet)){
						count += combinedLongerBasePatternAlphabetCountMap.get(combinedBasePatternAlphabet);
					}
					combinedLongerBasePatternAlphabetCountMap.put(combinedBasePatternAlphabet, count);
					
					count = nonOverlappingBasePatternIndexContributingLongerPatternMap.size();
					if(combinedLongerBasePatternAlphabetNonDuplicateCountMap.containsKey(combinedBasePatternAlphabet)){
						count += combinedLongerBasePatternAlphabetNonDuplicateCountMap.get(combinedBasePatternAlphabet);
					}
					combinedLongerBasePatternAlphabetNonDuplicateCountMap.put(combinedBasePatternAlphabet, count);
					
					instanceCount = noIdenticalTraces;
					if(combinedLongerBasePatternAlphabetInstanceCountMap.containsKey(combinedBasePatternAlphabet))
						instanceCount += combinedLongerBasePatternAlphabetInstanceCountMap.get(combinedBasePatternAlphabet);
					combinedLongerBasePatternAlphabetInstanceCountMap.put(combinedBasePatternAlphabet, instanceCount);
					
					instanceCount = 1;
					if(combinedLongerBasePatternAlphabetNonDuplicateInstanceCountMap.containsKey(combinedBasePatternAlphabet))
						instanceCount += combinedLongerBasePatternAlphabetNonDuplicateInstanceCountMap.get(combinedBasePatternAlphabet);
					combinedLongerBasePatternAlphabetNonDuplicateInstanceCountMap.put(combinedBasePatternAlphabet, instanceCount);

					//Check if this basePattern alphabet is also a tandem array and/or maximal repeat
					Set<Integer> tandemArrayLongerBasePatternIndicesSet = new HashSet<Integer>();
					Set<Integer> maximalRepeatLongerBasePatternIndicesSet = new HashSet<Integer>();
					if(entireLogTandemArrayBasePatternAlphabetPatternSetMap.containsKey(combinedBasePatternAlphabet)){
						tandemArrayBasePatternAlphabetPatternSet = entireLogTandemArrayBasePatternAlphabetPatternSetMap.get(combinedBasePatternAlphabet);
						for(Integer nonOverlappingLongerPatternIndex : nonOverlappingBasePatternIndexContributingLongerPatternMap.keySet()){
							if(tandemArrayBasePatternAlphabetPatternSet.contains(nonOverlappingBasePatternIndexContributingLongerPatternMap.get(nonOverlappingLongerPatternIndex))){
								contributingTandemArrayLongerBasePatternSet.add(nonOverlappingBasePatternIndexContributingLongerPatternMap.get(nonOverlappingLongerPatternIndex));
								tandemArrayLongerBasePatternIndicesSet.add(nonOverlappingLongerPatternIndex);
							}
						}
						
						if(tandemArrayLongerBasePatternIndicesSet.size() > 0){
							count = tandemArrayLongerBasePatternIndicesSet.size()*noIdenticalTraces;
							if(tandemArrayLongerBasePatternAlphabetCountMap.containsKey(combinedBasePatternAlphabet)){
								count += tandemArrayLongerBasePatternAlphabetCountMap.get(combinedBasePatternAlphabet);
							}
							tandemArrayLongerBasePatternAlphabetCountMap.put(combinedBasePatternAlphabet, count);
							
							count = tandemArrayLongerBasePatternIndicesSet.size();
							if(tandemArrayLongerBasePatternAlphabetNonDuplicateCountMap.containsKey(combinedBasePatternAlphabet)){
								count += tandemArrayLongerBasePatternAlphabetNonDuplicateCountMap.get(combinedBasePatternAlphabet);
							}
							tandemArrayLongerBasePatternAlphabetNonDuplicateCountMap.put(combinedBasePatternAlphabet, count);
							
							instanceCount = noIdenticalTraces;
							if(tandemArrayLongerBasePatternAlphabetInstanceCountMap.containsKey(combinedBasePatternAlphabet))
								instanceCount += tandemArrayLongerBasePatternAlphabetInstanceCountMap.get(combinedBasePatternAlphabet);
							tandemArrayLongerBasePatternAlphabetInstanceCountMap.put(combinedBasePatternAlphabet, instanceCount);
							
							instanceCount = 1;
							if(tandemArrayLongerBasePatternAlphabetNonDuplicateInstanceCountMap.containsKey(combinedBasePatternAlphabet))
								instanceCount += tandemArrayLongerBasePatternAlphabetNonDuplicateInstanceCountMap.get(combinedBasePatternAlphabet);
							tandemArrayLongerBasePatternAlphabetNonDuplicateInstanceCountMap.put(combinedBasePatternAlphabet, instanceCount);
						}
					}
					
					if(entireLogMaximalRepeatBasePatternAlphabetPatternSetMap.containsKey(combinedBasePatternAlphabet)){
						maximalRepeatBasePatternAlphabetPatternSet = entireLogMaximalRepeatBasePatternAlphabetPatternSetMap.get(combinedBasePatternAlphabet);
						
						for(Integer nonOverlappingLongerPatternIndex : nonOverlappingBasePatternIndexContributingLongerPatternMap.keySet()){
							if(maximalRepeatBasePatternAlphabetPatternSet.contains(nonOverlappingBasePatternIndexContributingLongerPatternMap.get(nonOverlappingLongerPatternIndex))){
								contributingMaximalRepeatLongerBasePatternSet.add(nonOverlappingBasePatternIndexContributingLongerPatternMap.get(nonOverlappingLongerPatternIndex));
								maximalRepeatLongerBasePatternIndicesSet.add(nonOverlappingLongerPatternIndex);
							}
						}
						if(maximalRepeatLongerBasePatternIndicesSet.size() > 0){
							count = maximalRepeatLongerBasePatternIndicesSet.size()*noIdenticalTraces;
							if(maximalRepeatLongerBasePatternAlphabetCountMap.containsKey(combinedBasePatternAlphabet)){
								count += maximalRepeatLongerBasePatternAlphabetCountMap.get(combinedBasePatternAlphabet);
							}
							maximalRepeatLongerBasePatternAlphabetCountMap.put(combinedBasePatternAlphabet, count);
							
							count = maximalRepeatLongerBasePatternIndicesSet.size();
							if(maximalRepeatLongerBasePatternAlphabetNonDuplicateCountMap.containsKey(combinedBasePatternAlphabet)){
								count += maximalRepeatLongerBasePatternAlphabetNonDuplicateCountMap.get(combinedBasePatternAlphabet);
							}
							maximalRepeatLongerBasePatternAlphabetNonDuplicateCountMap.put(combinedBasePatternAlphabet, count);
							
							instanceCount = noIdenticalTraces;
							if(maximalRepeatLongerBasePatternAlphabetInstanceCountMap.containsKey(combinedBasePatternAlphabet))
								instanceCount += maximalRepeatLongerBasePatternAlphabetInstanceCountMap.get(combinedBasePatternAlphabet);
							maximalRepeatLongerBasePatternAlphabetInstanceCountMap.put(combinedBasePatternAlphabet, instanceCount);
							
							instanceCount = 1;
							if(maximalRepeatLongerBasePatternAlphabetNonDuplicateInstanceCountMap.containsKey(combinedBasePatternAlphabet))
								instanceCount += maximalRepeatLongerBasePatternAlphabetNonDuplicateInstanceCountMap.get(combinedBasePatternAlphabet);
							maximalRepeatLongerBasePatternAlphabetNonDuplicateInstanceCountMap.put(combinedBasePatternAlphabet, instanceCount);
						}
					}

				}
			}
			
		}
		
		int noTraces = encodedLog.getNoTraces();
		int noUniqueTraces = encodedLog.getNoUniqueTraces();
		//The instance counts contain absolute count values, we need to change them to percentages
		int instanceCountPercentage;
		for(TreeSet<String> combinedBasePatternAlphabet : combinedShorterBasePatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(combinedShorterBasePatternAlphabetInstanceCountMap.get(combinedBasePatternAlphabet)*100.0f/noTraces);
			combinedShorterBasePatternAlphabetInstanceCountMap.put(combinedBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> combinedBasePatternAlphabet : combinedShorterBasePatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(combinedShorterBasePatternAlphabetNonDuplicateInstanceCountMap.get(combinedBasePatternAlphabet)*100.0f/noUniqueTraces);
			combinedShorterBasePatternAlphabetNonDuplicateInstanceCountMap.put(combinedBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> combinedPatternAlphabet : combinedShorterPatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(combinedShorterPatternAlphabetInstanceCountMap.get(combinedPatternAlphabet)*100.0f/noTraces);
			combinedShorterPatternAlphabetInstanceCountMap.put(combinedPatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> combinedPatternAlphabet : combinedShorterPatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(combinedShorterPatternAlphabetNonDuplicateInstanceCountMap.get(combinedPatternAlphabet)*100.0f/noUniqueTraces);
			combinedShorterPatternAlphabetNonDuplicateInstanceCountMap.put(combinedPatternAlphabet, instanceCountPercentage);
		}
		
		for(TreeSet<String> tandemArrayBasePatternAlphabet : tandemArrayShorterBasePatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(tandemArrayShorterBasePatternAlphabetInstanceCountMap.get(tandemArrayBasePatternAlphabet)*100.0f/noTraces);
			tandemArrayShorterBasePatternAlphabetInstanceCountMap.put(tandemArrayBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> tandemArrayBasePatternAlphabet : tandemArrayShorterBasePatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(tandemArrayShorterBasePatternAlphabetNonDuplicateInstanceCountMap.get(tandemArrayBasePatternAlphabet)*100.0f/noUniqueTraces);
			tandemArrayShorterBasePatternAlphabetNonDuplicateInstanceCountMap.put(tandemArrayBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> tandemArrayPatternAlphabet : tandemArrayShorterPatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(tandemArrayShorterPatternAlphabetInstanceCountMap.get(tandemArrayPatternAlphabet)*100.0f/noTraces);
			tandemArrayShorterPatternAlphabetInstanceCountMap.put(tandemArrayPatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> tandemArrayPatternAlphabet : tandemArrayShorterPatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(tandemArrayShorterPatternAlphabetNonDuplicateInstanceCountMap.get(tandemArrayPatternAlphabet)*100.0f/noUniqueTraces);
			tandemArrayShorterPatternAlphabetNonDuplicateInstanceCountMap.put(tandemArrayPatternAlphabet, instanceCountPercentage);
		}
		
		for(TreeSet<String> maximalRepeatBasePatternAlphabet : maximalRepeatShorterBasePatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(maximalRepeatShorterBasePatternAlphabetInstanceCountMap.get(maximalRepeatBasePatternAlphabet)*100.0f/noTraces);
			maximalRepeatShorterBasePatternAlphabetInstanceCountMap.put(maximalRepeatBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> maximalRepeatBasePatternAlphabet : maximalRepeatShorterBasePatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(maximalRepeatShorterBasePatternAlphabetNonDuplicateInstanceCountMap.get(maximalRepeatBasePatternAlphabet)*100.0f/noUniqueTraces);
			maximalRepeatShorterBasePatternAlphabetNonDuplicateInstanceCountMap.put(maximalRepeatBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> maximalRepeatPatternAlphabet : maximalRepeatShorterPatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(maximalRepeatShorterPatternAlphabetInstanceCountMap.get(maximalRepeatPatternAlphabet)*100.0f/noTraces);
			maximalRepeatShorterPatternAlphabetInstanceCountMap.put(maximalRepeatPatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> maximalRepeatPatternAlphabet : maximalRepeatShorterPatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(maximalRepeatShorterPatternAlphabetNonDuplicateInstanceCountMap.get(maximalRepeatPatternAlphabet)*100.0f/noUniqueTraces);
			maximalRepeatShorterPatternAlphabetNonDuplicateInstanceCountMap.put(maximalRepeatPatternAlphabet, instanceCountPercentage);
		}
		
		//DO it for Longer Pattern Metrics
		for(TreeSet<String> combinedBasePatternAlphabet : combinedLongerBasePatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(combinedLongerBasePatternAlphabetInstanceCountMap.get(combinedBasePatternAlphabet)*100.0f/noTraces);
			combinedLongerBasePatternAlphabetInstanceCountMap.put(combinedBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> combinedBasePatternAlphabet : combinedLongerBasePatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(combinedLongerBasePatternAlphabetNonDuplicateInstanceCountMap.get(combinedBasePatternAlphabet)*100.0f/noUniqueTraces);
			combinedLongerBasePatternAlphabetNonDuplicateInstanceCountMap.put(combinedBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> combinedPatternAlphabet : combinedLongerPatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(combinedLongerPatternAlphabetInstanceCountMap.get(combinedPatternAlphabet)*100.0f/noTraces);
			combinedLongerPatternAlphabetInstanceCountMap.put(combinedPatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> combinedPatternAlphabet : combinedLongerPatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(combinedLongerPatternAlphabetNonDuplicateInstanceCountMap.get(combinedPatternAlphabet)*100.0f/noUniqueTraces);
			combinedLongerPatternAlphabetNonDuplicateInstanceCountMap.put(combinedPatternAlphabet, instanceCountPercentage);
		}
		
		for(TreeSet<String> tandemArrayBasePatternAlphabet : tandemArrayLongerBasePatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(tandemArrayLongerBasePatternAlphabetInstanceCountMap.get(tandemArrayBasePatternAlphabet)*100.0f/noTraces);
			tandemArrayLongerBasePatternAlphabetInstanceCountMap.put(tandemArrayBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> tandemArrayBasePatternAlphabet : tandemArrayLongerBasePatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(tandemArrayLongerBasePatternAlphabetNonDuplicateInstanceCountMap.get(tandemArrayBasePatternAlphabet)*100.0f/noUniqueTraces);
			tandemArrayLongerBasePatternAlphabetNonDuplicateInstanceCountMap.put(tandemArrayBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> tandemArrayPatternAlphabet : tandemArrayLongerPatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(tandemArrayLongerPatternAlphabetInstanceCountMap.get(tandemArrayPatternAlphabet)*100.0f/noTraces);
			tandemArrayLongerPatternAlphabetInstanceCountMap.put(tandemArrayPatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> tandemArrayPatternAlphabet : tandemArrayLongerPatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(tandemArrayLongerPatternAlphabetNonDuplicateInstanceCountMap.get(tandemArrayPatternAlphabet)*100.0f/noUniqueTraces);
			tandemArrayLongerPatternAlphabetNonDuplicateInstanceCountMap.put(tandemArrayPatternAlphabet, instanceCountPercentage);
		}
		
		for(TreeSet<String> maximalRepeatBasePatternAlphabet : maximalRepeatLongerBasePatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(maximalRepeatLongerBasePatternAlphabetInstanceCountMap.get(maximalRepeatBasePatternAlphabet)*100.0f/noTraces);
			maximalRepeatLongerBasePatternAlphabetInstanceCountMap.put(maximalRepeatBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> maximalRepeatBasePatternAlphabet : maximalRepeatLongerBasePatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(maximalRepeatLongerBasePatternAlphabetNonDuplicateInstanceCountMap.get(maximalRepeatBasePatternAlphabet)*100.0f/noUniqueTraces);
			maximalRepeatLongerBasePatternAlphabetNonDuplicateInstanceCountMap.put(maximalRepeatBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> maximalRepeatPatternAlphabet : maximalRepeatLongerPatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(maximalRepeatLongerPatternAlphabetInstanceCountMap.get(maximalRepeatPatternAlphabet)*100.0f/noTraces);
			maximalRepeatLongerPatternAlphabetInstanceCountMap.put(maximalRepeatPatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> maximalRepeatPatternAlphabet : maximalRepeatLongerPatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(maximalRepeatLongerPatternAlphabetNonDuplicateInstanceCountMap.get(maximalRepeatPatternAlphabet)*100.0f/noUniqueTraces);
			maximalRepeatLongerPatternAlphabetNonDuplicateInstanceCountMap.put(maximalRepeatPatternAlphabet, instanceCountPercentage);
		}
		
		//Set the nonOverlapMetric
		EquivalenceClass equivalenceClass = new EquivalenceClass();
		nonOverlappingShorterPatternAlphabetCount.setCombinedBasePatternAlphabetCountMap(combinedShorterBasePatternAlphabetCountMap);
		nonOverlappingShorterPatternAlphabetCount.setCombinedBasePatternAlphabetInstanceCountMap(combinedShorterBasePatternAlphabetInstanceCountMap);
		nonOverlappingShorterPatternAlphabetCount.setCombinedBasePatternAlphabetNonDuplicateCountMap(combinedShorterBasePatternAlphabetNonDuplicateCountMap);
		nonOverlappingShorterPatternAlphabetCount.setCombinedBasePatternAlphabetNonDuplicateInstanceCountMap(combinedShorterBasePatternAlphabetNonDuplicateInstanceCountMap);
		
		nonOverlappingShorterPatternAlphabetCount.setCombinedPatternAlphabetCountMap(combinedShorterPatternAlphabetCountMap);
		nonOverlappingShorterPatternAlphabetCount.setCombinedPatternAlphabetInstanceCountMap(combinedShorterPatternAlphabetInstanceCountMap);
		nonOverlappingShorterPatternAlphabetCount.setCombinedPatternAlphabetNonDuplicateCountMap(combinedShorterPatternAlphabetNonDuplicateCountMap);
		nonOverlappingShorterPatternAlphabetCount.setCombinedPatternAlphabetNonDuplicateInstanceCountMap(combinedShorterPatternAlphabetNonDuplicateInstanceCountMap);
		
		nonOverlappingShorterPatternAlphabetCount.setContributingCombinedBasePatternAlphabetPatternSetMap(equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, contributingCombinedShorterBasePatternSet));
		nonOverlappingShorterPatternAlphabetCount.setContributingCombinedPatternAlphabetPatternSetMap(equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, contributingCombinedShorterPatternSet));

		nonOverlappingShorterPatternAlphabetCount.setTandemArrayBasePatternAlphabetCountMap(tandemArrayShorterBasePatternAlphabetCountMap);
		nonOverlappingShorterPatternAlphabetCount.setTandemArrayBasePatternAlphabetInstanceCountMap(tandemArrayShorterBasePatternAlphabetInstanceCountMap);
		nonOverlappingShorterPatternAlphabetCount.setTandemArrayBasePatternAlphabetNonDuplicateCountMap(tandemArrayShorterBasePatternAlphabetNonDuplicateCountMap);
		nonOverlappingShorterPatternAlphabetCount.setTandemArrayBasePatternAlphabetNonDuplicateInstanceCountMap(tandemArrayShorterBasePatternAlphabetNonDuplicateInstanceCountMap);
		
		nonOverlappingShorterPatternAlphabetCount.setTandemArrayPatternAlphabetCountMap(tandemArrayShorterPatternAlphabetCountMap);
		nonOverlappingShorterPatternAlphabetCount.setTandemArrayPatternAlphabetInstanceCountMap(tandemArrayShorterPatternAlphabetInstanceCountMap);
		nonOverlappingShorterPatternAlphabetCount.setTandemArrayPatternAlphabetNonDuplicateCountMap(tandemArrayShorterPatternAlphabetNonDuplicateCountMap);
		nonOverlappingShorterPatternAlphabetCount.setTandemArrayPatternAlphabetNonDuplicateInstanceCountMap(tandemArrayShorterPatternAlphabetNonDuplicateInstanceCountMap);
		
		nonOverlappingShorterPatternAlphabetCount.setContributingTandemArrayBasePatternAlphabetPatternSetMap(equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, contributingTandemArrayShorterBasePatternSet));
		nonOverlappingShorterPatternAlphabetCount.setContributingTandemArrayPatternAlphabetPatternSetMap(equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, contributingTandemArrayShorterPatternSet));
		
		nonOverlappingShorterPatternAlphabetCount.setMaximalRepeatBasePatternAlphabetCountMap(maximalRepeatShorterBasePatternAlphabetCountMap);
		nonOverlappingShorterPatternAlphabetCount.setMaximalRepeatBasePatternAlphabetInstanceCountMap(maximalRepeatShorterBasePatternAlphabetInstanceCountMap);
		nonOverlappingShorterPatternAlphabetCount.setMaximalRepeatBasePatternAlphabetNonDuplicateCountMap(maximalRepeatShorterBasePatternAlphabetNonDuplicateCountMap);
		nonOverlappingShorterPatternAlphabetCount.setMaximalRepeatBasePatternAlphabetNonDuplicateInstanceCountMap(maximalRepeatShorterBasePatternAlphabetNonDuplicateInstanceCountMap);
		
		nonOverlappingShorterPatternAlphabetCount.setMaximalRepeatPatternAlphabetCountMap(maximalRepeatShorterPatternAlphabetCountMap);
		nonOverlappingShorterPatternAlphabetCount.setMaximalRepeatPatternAlphabetInstanceCountMap(maximalRepeatShorterPatternAlphabetInstanceCountMap);
		nonOverlappingShorterPatternAlphabetCount.setMaximalRepeatPatternAlphabetNonDuplicateCountMap(maximalRepeatShorterPatternAlphabetNonDuplicateCountMap);
		nonOverlappingShorterPatternAlphabetCount.setMaximalRepeatPatternAlphabetNonDuplicateInstanceCountMap(maximalRepeatShorterPatternAlphabetNonDuplicateInstanceCountMap);
		
		nonOverlappingShorterPatternAlphabetCount.setContributingMaximalRepeatBasePatternAlphabetPatternSetMap(equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, contributingMaximalRepeatShorterBasePatternSet));
		nonOverlappingShorterPatternAlphabetCount.setContributingMaximalRepeatPatternAlphabetPatternSetMap(equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, contributingMaximalRepeatShorterPatternSet));

		//Setting for longer pattern alphabet count
		nonOverlappingLongerPatternAlphabetCount.setCombinedBasePatternAlphabetCountMap(combinedLongerBasePatternAlphabetCountMap);
		nonOverlappingLongerPatternAlphabetCount.setCombinedBasePatternAlphabetInstanceCountMap(combinedLongerBasePatternAlphabetInstanceCountMap);
		nonOverlappingLongerPatternAlphabetCount.setCombinedBasePatternAlphabetNonDuplicateCountMap(combinedLongerBasePatternAlphabetNonDuplicateCountMap);
		nonOverlappingLongerPatternAlphabetCount.setCombinedBasePatternAlphabetNonDuplicateInstanceCountMap(combinedLongerBasePatternAlphabetNonDuplicateInstanceCountMap);
		
		nonOverlappingLongerPatternAlphabetCount.setCombinedPatternAlphabetCountMap(combinedLongerPatternAlphabetCountMap);
		nonOverlappingLongerPatternAlphabetCount.setCombinedPatternAlphabetInstanceCountMap(combinedLongerPatternAlphabetInstanceCountMap);
		nonOverlappingLongerPatternAlphabetCount.setCombinedPatternAlphabetNonDuplicateCountMap(combinedLongerPatternAlphabetNonDuplicateCountMap);
		nonOverlappingLongerPatternAlphabetCount.setCombinedPatternAlphabetNonDuplicateInstanceCountMap(combinedLongerPatternAlphabetNonDuplicateInstanceCountMap);
		
		nonOverlappingLongerPatternAlphabetCount.setContributingCombinedBasePatternAlphabetPatternSetMap(equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, contributingCombinedLongerBasePatternSet));
		nonOverlappingLongerPatternAlphabetCount.setContributingCombinedPatternAlphabetPatternSetMap(equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, contributingCombinedLongerPatternSet));

		nonOverlappingLongerPatternAlphabetCount.setTandemArrayBasePatternAlphabetCountMap(tandemArrayLongerBasePatternAlphabetCountMap);
		nonOverlappingLongerPatternAlphabetCount.setTandemArrayBasePatternAlphabetInstanceCountMap(tandemArrayLongerBasePatternAlphabetInstanceCountMap);
		nonOverlappingLongerPatternAlphabetCount.setTandemArrayBasePatternAlphabetNonDuplicateCountMap(tandemArrayLongerBasePatternAlphabetNonDuplicateCountMap);
		nonOverlappingLongerPatternAlphabetCount.setTandemArrayBasePatternAlphabetNonDuplicateInstanceCountMap(tandemArrayLongerBasePatternAlphabetNonDuplicateInstanceCountMap);
		
		nonOverlappingLongerPatternAlphabetCount.setTandemArrayPatternAlphabetCountMap(tandemArrayLongerPatternAlphabetCountMap);
		nonOverlappingLongerPatternAlphabetCount.setTandemArrayPatternAlphabetInstanceCountMap(tandemArrayLongerPatternAlphabetInstanceCountMap);
		nonOverlappingLongerPatternAlphabetCount.setTandemArrayPatternAlphabetNonDuplicateCountMap(tandemArrayLongerPatternAlphabetNonDuplicateCountMap);
		nonOverlappingLongerPatternAlphabetCount.setTandemArrayPatternAlphabetNonDuplicateInstanceCountMap(tandemArrayLongerPatternAlphabetNonDuplicateInstanceCountMap);
		
		nonOverlappingLongerPatternAlphabetCount.setContributingTandemArrayBasePatternAlphabetPatternSetMap(equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, contributingTandemArrayLongerBasePatternSet));
		nonOverlappingLongerPatternAlphabetCount.setContributingTandemArrayPatternAlphabetPatternSetMap(equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, contributingTandemArrayLongerPatternSet));
		
		nonOverlappingLongerPatternAlphabetCount.setMaximalRepeatBasePatternAlphabetCountMap(maximalRepeatLongerBasePatternAlphabetCountMap);
		nonOverlappingLongerPatternAlphabetCount.setMaximalRepeatBasePatternAlphabetInstanceCountMap(maximalRepeatLongerBasePatternAlphabetInstanceCountMap);
		nonOverlappingLongerPatternAlphabetCount.setMaximalRepeatBasePatternAlphabetNonDuplicateCountMap(maximalRepeatLongerBasePatternAlphabetNonDuplicateCountMap);
		nonOverlappingLongerPatternAlphabetCount.setMaximalRepeatBasePatternAlphabetNonDuplicateInstanceCountMap(maximalRepeatLongerBasePatternAlphabetNonDuplicateInstanceCountMap);
		
		nonOverlappingLongerPatternAlphabetCount.setMaximalRepeatPatternAlphabetCountMap(maximalRepeatLongerPatternAlphabetCountMap);
		nonOverlappingLongerPatternAlphabetCount.setMaximalRepeatPatternAlphabetInstanceCountMap(maximalRepeatLongerPatternAlphabetInstanceCountMap);
		nonOverlappingLongerPatternAlphabetCount.setMaximalRepeatPatternAlphabetNonDuplicateCountMap(maximalRepeatLongerPatternAlphabetNonDuplicateCountMap);
		nonOverlappingLongerPatternAlphabetCount.setMaximalRepeatPatternAlphabetNonDuplicateInstanceCountMap(maximalRepeatLongerPatternAlphabetNonDuplicateInstanceCountMap);
		
		nonOverlappingLongerPatternAlphabetCount.setContributingMaximalRepeatBasePatternAlphabetPatternSetMap(equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, contributingMaximalRepeatLongerBasePatternSet));
		nonOverlappingLongerPatternAlphabetCount.setContributingMaximalRepeatPatternAlphabetPatternSetMap(equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, contributingMaximalRepeatLongerPatternSet));
	}

	private void computeNOGACMetrics(){
		System.out.println("In computeNOGACMetrics");
		nonOverlappingGlobalShorterPatternAlphabetCount = new MetricPatternAlphabetCount();
		nonOverlappingGlobalLongerPatternAlphabetCount = new MetricPatternAlphabetCount();
		
		Map<TreeSet<String>, Integer> tandemArrayShorterPatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayShorterBasePatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayShorterPatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayShorterBasePatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		Map<TreeSet<String>, Integer> maximalRepeatShorterPatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatShorterBasePatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatShorterPatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatShorterBasePatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		Map<TreeSet<String>, Integer> combinedShorterPatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedShorterBasePatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedShorterPatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedShorterBasePatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		Map<TreeSet<String>, Integer> tandemArrayShorterPatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayShorterBasePatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayShorterPatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayShorterBasePatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		Map<TreeSet<String>, Integer> maximalRepeatShorterPatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatShorterBasePatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatShorterPatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatShorterBasePatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		Map<TreeSet<String>, Integer> combinedShorterPatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedShorterBasePatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedShorterPatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedShorterBasePatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		
		Map<TreeSet<String>, Integer> tandemArrayLongerPatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayLongerBasePatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayLongerPatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayLongerBasePatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		Map<TreeSet<String>, Integer> maximalRepeatLongerPatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatLongerBasePatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatLongerPatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatLongerBasePatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		Map<TreeSet<String>, Integer> combinedLongerPatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedLongerBasePatternAlphabetCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedLongerPatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedLongerBasePatternAlphabetInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		Map<TreeSet<String>, Integer> tandemArrayLongerPatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayLongerBasePatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayLongerPatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> tandemArrayLongerBasePatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		Map<TreeSet<String>, Integer> maximalRepeatLongerPatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatLongerBasePatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatLongerPatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> maximalRepeatLongerBasePatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		Map<TreeSet<String>, Integer> combinedLongerPatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedLongerBasePatternAlphabetNonDuplicateCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedLongerPatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		Map<TreeSet<String>, Integer> combinedLongerBasePatternAlphabetNonDuplicateInstanceCountMap = new HashMap<TreeSet<String>, Integer>();
		
		Map<String, TreeSet<String>> patternPatternAlphabetMap = new HashMap<String, TreeSet<String>>();
		Map<String, TreeSet<String>> basePatternPatternAlphabetMap = new HashMap<String, TreeSet<String>>();
		Set<String> tandemArrayPatternAlphabetPatternSet, maximalRepeatPatternAlphabetPatternSet;
		Set<String> tandemArrayBasePatternAlphabetPatternSet, maximalRepeatBasePatternAlphabetPatternSet;
		
		Set<String> contributingCombinedShorterPatternSet = new HashSet<String>();
		Set<String> contributingTandemArrayShorterPatternSet = new HashSet<String>();
		Set<String> contributingMaximalRepeatShorterPatternSet = new HashSet<String>();

		Set<String> contributingCombinedLongerPatternSet = new HashSet<String>();
		Set<String> contributingTandemArrayLongerPatternSet = new HashSet<String>();
		Set<String> contributingMaximalRepeatLongerPatternSet = new HashSet<String>();

		Set<String> contributingCombinedShorterBasePatternSet = new HashSet<String>();
		Set<String> contributingTandemArrayShorterBasePatternSet = new HashSet<String>();
		Set<String> contributingMaximalRepeatShorterBasePatternSet = new HashSet<String>();

		Set<String> contributingCombinedLongerBasePatternSet = new HashSet<String>();
		Set<String> contributingTandemArrayLongerBasePatternSet = new HashSet<String>();
		Set<String> contributingMaximalRepeatLongerBasePatternSet = new HashSet<String>();

		
		Set<TreeSet<String>> encodedTraceContributingShorterCombinedPatternAlphabetSet = new HashSet<TreeSet<String>>();
		Set<TreeSet<String>> encodedTraceContributingShorterTandemArrayPatternAlphabetSet = new HashSet<TreeSet<String>>();
		Set<TreeSet<String>> encodedTraceContributingShorterMaximalRepeatPatternAlphabetSet = new HashSet<TreeSet<String>>();
		
		Set<TreeSet<String>> encodedTraceContributingLongerCombinedPatternAlphabetSet = new HashSet<TreeSet<String>>();
		Set<TreeSet<String>> encodedTraceContributingLongerTandemArrayPatternAlphabetSet = new HashSet<TreeSet<String>>();
		Set<TreeSet<String>> encodedTraceContributingLongerMaximalRepeatPatternAlphabetSet = new HashSet<TreeSet<String>>();
		
		Set<TreeSet<String>> encodedTraceContributingShorterCombinedBasePatternAlphabetSet = new HashSet<TreeSet<String>>();
		Set<TreeSet<String>> encodedTraceContributingShorterTandemArrayBasePatternAlphabetSet = new HashSet<TreeSet<String>>();
		Set<TreeSet<String>> encodedTraceContributingShorterMaximalRepeatBasePatternAlphabetSet = new HashSet<TreeSet<String>>();
		
		Set<TreeSet<String>> encodedTraceContributingLongerCombinedBasePatternAlphabetSet = new HashSet<TreeSet<String>>();
		Set<TreeSet<String>> encodedTraceContributingLongerTandemArrayBasePatternAlphabetSet = new HashSet<TreeSet<String>>();
		Set<TreeSet<String>> encodedTraceContributingLongerMaximalRepeatBasePatternAlphabetSet = new HashSet<TreeSet<String>>();
		
		/*
		 * Get the global pattern sets
		 */
		Set<String> combinedPatternAlphabetPatternSet, combinedBasePatternAlphabetPatternSet;
		for(TreeSet<String> combinedPatternAlphabet : entireLogCombinedPatternAlphabetPatternSetMap.keySet()){
			if(ignorablePatternAlphabetSet != null && ignorablePatternAlphabetSet.contains(combinedPatternAlphabet))
				continue;
			combinedPatternAlphabetPatternSet = entireLogCombinedPatternAlphabetPatternSetMap.get(combinedPatternAlphabet);
			for(String pattern : combinedPatternAlphabetPatternSet)
				patternPatternAlphabetMap.put(pattern, combinedPatternAlphabet);
		}
		
		for(TreeSet<String> combinedBasePatternAlphabet : entireLogCombinedBasePatternAlphabetPatternSetMap.keySet()){
			if(ignorablePatternAlphabetSet != null && ignorablePatternAlphabetSet.contains(combinedBasePatternAlphabet))
				continue;
			
			combinedBasePatternAlphabetPatternSet = entireLogCombinedBasePatternAlphabetPatternSetMap.get(combinedBasePatternAlphabet);
			for(String pattern : combinedBasePatternAlphabetPatternSet)
				basePatternPatternAlphabetMap.put(pattern, combinedBasePatternAlphabet);
		}
		
		int encodingLength = encodedLog.getEncodingLength();
		List<String> uniqueEncodedTraceList = encodedLog.getUniqueEncodedTraceList();
		Map<String, TreeSet<Integer>> encodedTraceIdenticalTraceIndicesSetMap = encodedLog.getEncodedTraceIdenticalTraceIndicesSetMap();
		
		int noIdenticalTraces;
		UkkonenSuffixTree suffixTree;
		
		Map<Integer, SortedSet<String>> indexPatternSetMap = new HashMap<Integer, SortedSet<String>>();
		Map<Integer, SortedSet<String>> indexBasePatternSetMap = new HashMap<Integer, SortedSet<String>>();
		
		SortedSet<String> indexPatternSet;
		SortedSet<String> indexBasePatternSet;
		
		int noMatches;
		int[] indexPos;
		
		for(String encodedTrace : uniqueEncodedTraceList){
			encodedTraceContributingShorterCombinedPatternAlphabetSet.clear();
			encodedTraceContributingShorterTandemArrayPatternAlphabetSet.clear();
			encodedTraceContributingShorterMaximalRepeatPatternAlphabetSet.clear();
			
			encodedTraceContributingLongerCombinedPatternAlphabetSet.clear();
			encodedTraceContributingLongerTandemArrayPatternAlphabetSet.clear();
			encodedTraceContributingLongerMaximalRepeatPatternAlphabetSet.clear();
			
			encodedTraceContributingShorterCombinedBasePatternAlphabetSet.clear();
			encodedTraceContributingShorterTandemArrayBasePatternAlphabetSet.clear();
			encodedTraceContributingShorterMaximalRepeatBasePatternAlphabetSet.clear();
			
			encodedTraceContributingLongerCombinedBasePatternAlphabetSet.clear();
			encodedTraceContributingLongerTandemArrayBasePatternAlphabetSet.clear();
			encodedTraceContributingLongerMaximalRepeatBasePatternAlphabetSet.clear();
			
			noIdenticalTraces = encodedTraceIdenticalTraceIndicesSetMap.get(encodedTrace).size();
			suffixTree = new UkkonenSuffixTree(encodingLength, encodedTrace);

			indexPatternSetMap.clear();
			
			for(String pattern : patternPatternAlphabetMap.keySet()){
				noMatches = suffixTree.noMatches(pattern);
				if(noMatches <= 0)
					continue;
				
				indexPos = suffixTree.getMatches(pattern);
				for(int index : indexPos){
					if(indexPatternSetMap.containsKey(index))
						indexPatternSet = indexPatternSetMap.get(index);
					else
						indexPatternSet = new TreeSet<String>(new AscendingStringLengthComparator());
					indexPatternSet.add(pattern);
					indexPatternSetMap.put(index, indexPatternSet);
				}
			}
			
			/*
			 * Update the counts, first do it for shorter patterns in combined pattern alphabet, 
			 */
			int nextIndex = -1, count, instanceCount;
			String contributingShorterPattern, contributingLongerPattern;
			TreeSet<String> contributingShorterPatternAlphabet, contributingLongerPatternAlphabet;
			
			TreeSet<Integer> sortedIndexSet = new TreeSet<Integer>();
			sortedIndexSet.addAll(indexPatternSetMap.keySet());
			
			for(Integer index : sortedIndexSet){
				if(index < nextIndex)
					continue;
				
				indexPatternSet = indexPatternSetMap.get(index);
				
				//Check if it is of length 1 and if so, consider it only if it is the only pattern at this index
				if(indexPatternSet.size() == 1 && indexPatternSet.first().length()==encodingLength){
					//consider it as it is the only one
					contributingShorterPattern = indexPatternSet.first();
				}else if(indexPatternSet.first().length()==encodingLength){
					//Get the second one
					contributingShorterPattern = (String)indexPatternSet.toArray()[1];
				}else{
					//consider the first one as the shorter pattern
					contributingShorterPattern = indexPatternSet.first();
				}
				contributingShorterPatternAlphabet = patternPatternAlphabetMap.get(contributingShorterPattern);
				
				contributingCombinedShorterPatternSet.add(contributingShorterPattern);
				encodedTraceContributingShorterCombinedPatternAlphabetSet.add(contributingShorterPatternAlphabet);
				
				count = noIdenticalTraces;
				if(combinedShorterPatternAlphabetCountMap.containsKey(contributingShorterPatternAlphabet)){
					count += combinedShorterPatternAlphabetCountMap.get(contributingShorterPatternAlphabet);
				}
				combinedShorterPatternAlphabetCountMap.put(contributingShorterPatternAlphabet, count);

				count = 1;
				if(combinedShorterPatternAlphabetNonDuplicateCountMap.containsKey(contributingShorterPatternAlphabet))
					count += combinedShorterPatternAlphabetNonDuplicateCountMap.get(contributingShorterPatternAlphabet);
				combinedShorterPatternAlphabetNonDuplicateCountMap.put(contributingShorterPatternAlphabet, count);
				
				nextIndex = index+contributingShorterPattern.length()/encodingLength;
			}
			//Update the instance counts
			for(TreeSet<String> contributingPatternAlphabet : encodedTraceContributingShorterCombinedPatternAlphabetSet){
				instanceCount = noIdenticalTraces;
				if(combinedShorterPatternAlphabetInstanceCountMap.containsKey(contributingPatternAlphabet)){
					instanceCount += combinedShorterPatternAlphabetInstanceCountMap.get(contributingPatternAlphabet);
				}
				combinedShorterPatternAlphabetInstanceCountMap.put(contributingPatternAlphabet, instanceCount);
			}
			for(TreeSet<String> contributingPatternAlphabet : encodedTraceContributingShorterCombinedPatternAlphabetSet){
				instanceCount = 1;
				if(combinedShorterPatternAlphabetNonDuplicateInstanceCountMap.containsKey(contributingPatternAlphabet)){
					instanceCount += combinedShorterPatternAlphabetNonDuplicateInstanceCountMap.get(contributingPatternAlphabet);
				}
				combinedShorterPatternAlphabetNonDuplicateInstanceCountMap.put(contributingPatternAlphabet, instanceCount);
			}
			// Do it for tandem arrays
			nextIndex = -1;
			String patternAtIndex;
			boolean patternExistsAtIndex;
			for(Integer index : sortedIndexSet){
				if(index < nextIndex)
					continue;

				patternExistsAtIndex = false;
				contributingShorterPattern = "";
				indexPatternSet = indexPatternSetMap.get(index);
				
				Iterator<String> it = indexPatternSet.iterator();
				while(it.hasNext()){
					patternAtIndex = it.next();
					if(entireLogTandemArrayPatternAlphabetPatternSetMap.containsKey(patternPatternAlphabetMap.get(patternAtIndex))){
						tandemArrayPatternAlphabetPatternSet = entireLogTandemArrayPatternAlphabetPatternSetMap.get(patternPatternAlphabetMap.get(patternAtIndex));
						if(tandemArrayPatternAlphabetPatternSet.contains(patternAtIndex)){
							patternExistsAtIndex = true;
							contributingShorterPattern = patternAtIndex;
							if(contributingShorterPattern.length() != encodingLength)
								break;
						}
					}
				}
				if(patternExistsAtIndex){
					contributingShorterPatternAlphabet = patternPatternAlphabetMap.get(contributingShorterPattern);
					
					contributingTandemArrayShorterPatternSet.add(contributingShorterPattern);
					encodedTraceContributingShorterTandemArrayPatternAlphabetSet.add(contributingShorterPatternAlphabet);
					
					count = noIdenticalTraces;
					if(tandemArrayShorterPatternAlphabetCountMap.containsKey(contributingShorterPatternAlphabet)){
						count += tandemArrayShorterPatternAlphabetCountMap.get(contributingShorterPatternAlphabet);
					}
					tandemArrayShorterPatternAlphabetCountMap.put(contributingShorterPatternAlphabet, count);
	
					count = 1;
					if(tandemArrayShorterPatternAlphabetNonDuplicateCountMap.containsKey(contributingShorterPatternAlphabet))
						count += tandemArrayShorterPatternAlphabetNonDuplicateCountMap.get(contributingShorterPatternAlphabet);
					tandemArrayShorterPatternAlphabetNonDuplicateCountMap.put(contributingShorterPatternAlphabet, count);
					
					nextIndex = index+contributingShorterPattern.length()/encodingLength;
				}
			}
			//Update the instance counts
			for(TreeSet<String> contributingPatternAlphabet : encodedTraceContributingShorterTandemArrayPatternAlphabetSet){
				instanceCount = noIdenticalTraces;
				if(tandemArrayShorterPatternAlphabetInstanceCountMap.containsKey(contributingPatternAlphabet)){
					instanceCount += tandemArrayShorterPatternAlphabetInstanceCountMap.get(contributingPatternAlphabet);
				}
				tandemArrayShorterPatternAlphabetInstanceCountMap.put(contributingPatternAlphabet, instanceCount);
			}
			for(TreeSet<String> contributingPatternAlphabet : encodedTraceContributingShorterTandemArrayPatternAlphabetSet){
				instanceCount = 1;
				if(tandemArrayShorterPatternAlphabetNonDuplicateInstanceCountMap.containsKey(contributingPatternAlphabet)){
					instanceCount += tandemArrayShorterPatternAlphabetNonDuplicateInstanceCountMap.get(contributingPatternAlphabet);
				}
				tandemArrayShorterPatternAlphabetNonDuplicateInstanceCountMap.put(contributingPatternAlphabet, instanceCount);
			}

			// Do it for maximal repeats
			nextIndex = -1;
			for(Integer index : sortedIndexSet){
				if(index < nextIndex)
					continue;

				patternExistsAtIndex = false;
				contributingShorterPattern = "";
				indexPatternSet = indexPatternSetMap.get(index);
				
				Iterator<String> it = indexPatternSet.iterator();
				while(it.hasNext()){
					patternAtIndex = it.next();
					if(entireLogMaximalRepeatPatternAlphabetPatternSetMap.containsKey(patternPatternAlphabetMap.get(patternAtIndex))){
						maximalRepeatPatternAlphabetPatternSet = entireLogMaximalRepeatPatternAlphabetPatternSetMap.get(patternPatternAlphabetMap.get(patternAtIndex));
						if(maximalRepeatPatternAlphabetPatternSet.contains(patternAtIndex)){
							patternExistsAtIndex = true;
							contributingShorterPattern = patternAtIndex;
							if(contributingShorterPattern.length() != encodingLength)
								break;
						}
					}
				}
				if(patternExistsAtIndex){
					contributingShorterPatternAlphabet = patternPatternAlphabetMap.get(contributingShorterPattern);
					
					contributingMaximalRepeatShorterPatternSet.add(contributingShorterPattern);
					encodedTraceContributingShorterMaximalRepeatPatternAlphabetSet.add(contributingShorterPatternAlphabet);
					
					count = noIdenticalTraces;
					if(maximalRepeatShorterPatternAlphabetCountMap.containsKey(contributingShorterPatternAlphabet)){
						count += maximalRepeatShorterPatternAlphabetCountMap.get(contributingShorterPatternAlphabet);
					}
					maximalRepeatShorterPatternAlphabetCountMap.put(contributingShorterPatternAlphabet, count);
	
					count = 1;
					if(maximalRepeatShorterPatternAlphabetNonDuplicateCountMap.containsKey(contributingShorterPatternAlphabet))
						count += maximalRepeatShorterPatternAlphabetNonDuplicateCountMap.get(contributingShorterPatternAlphabet);
					maximalRepeatShorterPatternAlphabetNonDuplicateCountMap.put(contributingShorterPatternAlphabet, count);
					
					nextIndex = index+contributingShorterPattern.length()/encodingLength;
				}
			}
			//Update the instance counts
			for(TreeSet<String> contributingPatternAlphabet : encodedTraceContributingShorterMaximalRepeatPatternAlphabetSet){
				instanceCount = noIdenticalTraces;
				if(maximalRepeatShorterPatternAlphabetInstanceCountMap.containsKey(contributingPatternAlphabet)){
					instanceCount += maximalRepeatShorterPatternAlphabetInstanceCountMap.get(contributingPatternAlphabet);
				}
				maximalRepeatShorterPatternAlphabetInstanceCountMap.put(contributingPatternAlphabet, instanceCount);
			}
			for(TreeSet<String> contributingPatternAlphabet : encodedTraceContributingShorterMaximalRepeatPatternAlphabetSet){
				instanceCount = 1;
				if(maximalRepeatShorterPatternAlphabetNonDuplicateInstanceCountMap.containsKey(contributingPatternAlphabet)){
					instanceCount += maximalRepeatShorterPatternAlphabetNonDuplicateInstanceCountMap.get(contributingPatternAlphabet);
				}
				maximalRepeatShorterPatternAlphabetNonDuplicateInstanceCountMap.put(contributingPatternAlphabet, instanceCount);
			}
			
			/*
			 * Update the counts, do it for longer patterns in combined pattern alphabet, 
			 */
			nextIndex = -1;
			for(Integer index : sortedIndexSet){
				if(index < nextIndex)
					continue;
				
				indexPatternSet = indexPatternSetMap.get(index);
				
				contributingLongerPattern = indexPatternSet.last();
				contributingLongerPatternAlphabet = patternPatternAlphabetMap.get(contributingLongerPattern);
				
				contributingCombinedLongerPatternSet.add(contributingLongerPattern);
				encodedTraceContributingLongerCombinedPatternAlphabetSet.add(contributingLongerPatternAlphabet);
				
				count = noIdenticalTraces;
				if(combinedLongerPatternAlphabetCountMap.containsKey(contributingLongerPatternAlphabet)){
					count += combinedLongerPatternAlphabetCountMap.get(contributingLongerPatternAlphabet);
				}
				combinedLongerPatternAlphabetCountMap.put(contributingLongerPatternAlphabet, count);

				count = 1;
				if(combinedLongerPatternAlphabetNonDuplicateCountMap.containsKey(contributingLongerPatternAlphabet))
					count += combinedLongerPatternAlphabetNonDuplicateCountMap.get(contributingLongerPatternAlphabet);
				combinedLongerPatternAlphabetNonDuplicateCountMap.put(contributingLongerPatternAlphabet, count);
				
				nextIndex = index+contributingLongerPattern.length()/encodingLength;
			}
			//Update the instance counts
			for(TreeSet<String> contributingPatternAlphabet : encodedTraceContributingLongerCombinedPatternAlphabetSet){
				instanceCount = noIdenticalTraces;
				if(combinedLongerPatternAlphabetInstanceCountMap.containsKey(contributingPatternAlphabet)){
					instanceCount += combinedLongerPatternAlphabetInstanceCountMap.get(contributingPatternAlphabet);
				}
				combinedLongerPatternAlphabetInstanceCountMap.put(contributingPatternAlphabet, instanceCount);
			}
			for(TreeSet<String> contributingPatternAlphabet : encodedTraceContributingLongerCombinedPatternAlphabetSet){
				instanceCount = 1;
				if(combinedLongerPatternAlphabetNonDuplicateInstanceCountMap.containsKey(contributingPatternAlphabet)){
					instanceCount += combinedLongerPatternAlphabetNonDuplicateInstanceCountMap.get(contributingPatternAlphabet);
				}
				combinedLongerPatternAlphabetNonDuplicateInstanceCountMap.put(contributingPatternAlphabet, instanceCount);
			}
			
			// Do it for tandem arrays
			nextIndex = -1;
			for(Integer index : sortedIndexSet){
				if(index < nextIndex)
					continue;

				patternExistsAtIndex = false;
				contributingLongerPattern = "";
				indexPatternSet = indexPatternSetMap.get(index);
				
				int noPatterns = indexPatternSet.size();
				Object[] indexPatternSetAsArray = indexPatternSet.toArray();
				for(int i = noPatterns-1; i >= 0; i--){
					patternAtIndex = (String)indexPatternSetAsArray[i];
					if(entireLogTandemArrayPatternAlphabetPatternSetMap.containsKey(patternPatternAlphabetMap.get(patternAtIndex))){
						tandemArrayPatternAlphabetPatternSet = entireLogTandemArrayPatternAlphabetPatternSetMap.get(patternPatternAlphabetMap.get(patternAtIndex));
						if(tandemArrayPatternAlphabetPatternSet.contains(patternAtIndex)){
							patternExistsAtIndex = true;
							contributingLongerPattern = patternAtIndex;
							break;
						}
					}
				}
				if(patternExistsAtIndex){
//					System.out.println("Tandem Array Exists at index: "+index+"@ "+contributingLongerPattern);
					contributingLongerPatternAlphabet = patternPatternAlphabetMap.get(contributingLongerPattern);
					contributingTandemArrayLongerPatternSet.add(contributingLongerPattern);
					encodedTraceContributingLongerTandemArrayPatternAlphabetSet.add(contributingLongerPatternAlphabet);
					
					count = noIdenticalTraces;
					if(tandemArrayLongerPatternAlphabetCountMap.containsKey(contributingLongerPatternAlphabet)){
						count += tandemArrayLongerPatternAlphabetCountMap.get(contributingLongerPatternAlphabet);
					}
					tandemArrayLongerPatternAlphabetCountMap.put(contributingLongerPatternAlphabet, count);
	
					count = 1;
					if(tandemArrayLongerPatternAlphabetNonDuplicateCountMap.containsKey(contributingLongerPatternAlphabet))
						count += tandemArrayLongerPatternAlphabetNonDuplicateCountMap.get(contributingLongerPatternAlphabet);
					tandemArrayLongerPatternAlphabetNonDuplicateCountMap.put(contributingLongerPatternAlphabet, count);
					
					nextIndex = index+contributingLongerPattern.length()/encodingLength;
				}
			}
			//Update the instance counts
			for(TreeSet<String> contributingPatternAlphabet : encodedTraceContributingLongerTandemArrayPatternAlphabetSet){
				instanceCount = noIdenticalTraces;
				if(tandemArrayLongerPatternAlphabetInstanceCountMap.containsKey(contributingPatternAlphabet)){
					instanceCount += tandemArrayLongerPatternAlphabetInstanceCountMap.get(contributingPatternAlphabet);
				}
				tandemArrayLongerPatternAlphabetInstanceCountMap.put(contributingPatternAlphabet, instanceCount);
			}
			for(TreeSet<String> contributingPatternAlphabet : encodedTraceContributingLongerTandemArrayPatternAlphabetSet){
				instanceCount = 1;
				if(tandemArrayLongerPatternAlphabetNonDuplicateInstanceCountMap.containsKey(contributingPatternAlphabet)){
					instanceCount += tandemArrayLongerPatternAlphabetNonDuplicateInstanceCountMap.get(contributingPatternAlphabet);
				}
				tandemArrayLongerPatternAlphabetNonDuplicateInstanceCountMap.put(contributingPatternAlphabet, instanceCount);
			}

			// Do it for maximal repeats
			nextIndex = -1;
			for(Integer index : sortedIndexSet){
				if(index < nextIndex)
					continue;

				patternExistsAtIndex = false;
				contributingLongerPattern = "";
				indexPatternSet = indexPatternSetMap.get(index);
				
				int noPatterns = indexPatternSet.size();
				Object[] indexPatternSetAsArray = indexPatternSet.toArray();
				for(int i = noPatterns-1; i >= 0; i--){
					patternAtIndex = (String)indexPatternSetAsArray[i];
					if(entireLogMaximalRepeatPatternAlphabetPatternSetMap.containsKey(patternPatternAlphabetMap.get(patternAtIndex))){
						maximalRepeatPatternAlphabetPatternSet = entireLogMaximalRepeatPatternAlphabetPatternSetMap.get(patternPatternAlphabetMap.get(patternAtIndex));
						if(maximalRepeatPatternAlphabetPatternSet.contains(patternAtIndex)){
							patternExistsAtIndex = true;
							contributingLongerPattern = patternAtIndex;
							break;
						}
					}
				}
				if(patternExistsAtIndex){
					contributingLongerPatternAlphabet = patternPatternAlphabetMap.get(contributingLongerPattern);
					contributingMaximalRepeatLongerPatternSet.add(contributingLongerPattern);
					encodedTraceContributingLongerMaximalRepeatPatternAlphabetSet.add(contributingLongerPatternAlphabet);
					
					count = noIdenticalTraces;
					if(maximalRepeatLongerPatternAlphabetCountMap.containsKey(contributingLongerPatternAlphabet)){
						count += maximalRepeatLongerPatternAlphabetCountMap.get(contributingLongerPatternAlphabet);
					}
					maximalRepeatLongerPatternAlphabetCountMap.put(contributingLongerPatternAlphabet, count);
	
					count = 1;
					if(maximalRepeatLongerPatternAlphabetNonDuplicateCountMap.containsKey(contributingLongerPatternAlphabet))
						count += maximalRepeatLongerPatternAlphabetNonDuplicateCountMap.get(contributingLongerPatternAlphabet);
					maximalRepeatLongerPatternAlphabetNonDuplicateCountMap.put(contributingLongerPatternAlphabet, count);
					
					nextIndex = index+contributingLongerPattern.length()/encodingLength;
				}
			}
			
			//Update the instance counts
			for(TreeSet<String> contributingPatternAlphabet : encodedTraceContributingLongerMaximalRepeatPatternAlphabetSet){
				instanceCount = noIdenticalTraces;
				if(maximalRepeatLongerPatternAlphabetInstanceCountMap.containsKey(contributingPatternAlphabet)){
					instanceCount += maximalRepeatLongerPatternAlphabetInstanceCountMap.get(contributingPatternAlphabet);
				}
				maximalRepeatLongerPatternAlphabetInstanceCountMap.put(contributingPatternAlphabet, instanceCount);
			}
			for(TreeSet<String> contributingPatternAlphabet : encodedTraceContributingLongerMaximalRepeatPatternAlphabetSet){
				instanceCount = 1;
				if(maximalRepeatLongerPatternAlphabetNonDuplicateInstanceCountMap.containsKey(contributingPatternAlphabet)){
					instanceCount += maximalRepeatLongerPatternAlphabetNonDuplicateInstanceCountMap.get(contributingPatternAlphabet);
				}
				maximalRepeatLongerPatternAlphabetNonDuplicateInstanceCountMap.put(contributingPatternAlphabet, instanceCount);
			}
			
			/*
			 * Do it for base patterns
			 */
			indexBasePatternSetMap.clear();
			for(String pattern : basePatternPatternAlphabetMap.keySet()){
				noMatches = suffixTree.noMatches(pattern);
				if(noMatches <= 0)
					continue;
				
				indexPos = suffixTree.getMatches(pattern);
				for(int index : indexPos){
					if(indexBasePatternSetMap.containsKey(index))
						indexBasePatternSet = indexBasePatternSetMap.get(index);
					else
						indexBasePatternSet = new TreeSet<String>(new AscendingStringLengthComparator());
					indexBasePatternSet.add(pattern);
					indexBasePatternSetMap.put(index, indexBasePatternSet);
				}
			}
			
			/*
			 * Update the counts, first do it for shorter patterns in combined pattern alphabet, 
			 */
			nextIndex = -1;
			String contributingShorterBasePattern, contributingLongerBasePattern;
			TreeSet<String> contributingShorterBasePatternAlphabet, contributingLongerBasePatternAlphabet;
			
			sortedIndexSet.clear();
			sortedIndexSet.addAll(indexBasePatternSetMap.keySet());
			
			for(Integer index : sortedIndexSet){
				if(index < nextIndex)
					continue;
				
				indexBasePatternSet = indexBasePatternSetMap.get(index);
				
				//Check if it is of length 1 and if so, consider it only if it is the only pattern at this index
				if(indexBasePatternSet.size() == 1 && indexBasePatternSet.first().length()==encodingLength){
					//consider it as it is the only one
					contributingShorterBasePattern = indexBasePatternSet.first();
				}else if(indexBasePatternSet.first().length()==encodingLength){
					//Get the second one
					contributingShorterBasePattern = (String)indexBasePatternSet.toArray()[1];
				}else{
					//consider the first one as the shorter pattern
					contributingShorterBasePattern = indexBasePatternSet.first();
				}
				contributingShorterBasePatternAlphabet = basePatternPatternAlphabetMap.get(contributingShorterBasePattern);
				
				contributingCombinedShorterBasePatternSet.add(contributingShorterBasePattern);
				encodedTraceContributingShorterCombinedBasePatternAlphabetSet.add(contributingShorterBasePatternAlphabet);
				
				count = noIdenticalTraces;
				if(combinedShorterBasePatternAlphabetCountMap.containsKey(contributingShorterBasePatternAlphabet)){
					count += combinedShorterBasePatternAlphabetCountMap.get(contributingShorterBasePatternAlphabet);
				}
				combinedShorterBasePatternAlphabetCountMap.put(contributingShorterBasePatternAlphabet, count);

				count = 1;
				if(combinedShorterBasePatternAlphabetNonDuplicateCountMap.containsKey(contributingShorterBasePatternAlphabet))
					count += combinedShorterBasePatternAlphabetNonDuplicateCountMap.get(contributingShorterBasePatternAlphabet);
				combinedShorterBasePatternAlphabetNonDuplicateCountMap.put(contributingShorterBasePatternAlphabet, count);
				
				nextIndex = index+contributingShorterBasePattern.length()/encodingLength;
			}
			//Update the instance counts
			for(TreeSet<String> contributingBasePatternAlphabet : encodedTraceContributingShorterCombinedBasePatternAlphabetSet){
				instanceCount = noIdenticalTraces;
				if(combinedShorterBasePatternAlphabetInstanceCountMap.containsKey(contributingBasePatternAlphabet)){
					instanceCount += combinedShorterBasePatternAlphabetInstanceCountMap.get(contributingBasePatternAlphabet);
				}
				combinedShorterBasePatternAlphabetInstanceCountMap.put(contributingBasePatternAlphabet, instanceCount);
			}
			for(TreeSet<String> contributingBasePatternAlphabet : encodedTraceContributingShorterCombinedBasePatternAlphabetSet){
				instanceCount = 1;
				if(combinedShorterBasePatternAlphabetNonDuplicateInstanceCountMap.containsKey(contributingBasePatternAlphabet)){
					instanceCount += combinedShorterBasePatternAlphabetNonDuplicateInstanceCountMap.get(contributingBasePatternAlphabet);
				}
				combinedShorterBasePatternAlphabetNonDuplicateInstanceCountMap.put(contributingBasePatternAlphabet, instanceCount);
			}
			// Do it for tandem arrays
			nextIndex = -1;
			for(Integer index : sortedIndexSet){
				if(index < nextIndex)
					continue;

				patternExistsAtIndex = false;
				contributingShorterBasePattern = "";
				indexBasePatternSet = indexBasePatternSetMap.get(index);
				
				Iterator<String> it = indexBasePatternSet.iterator();
				while(it.hasNext()){
					patternAtIndex = it.next();
					if(entireLogTandemArrayBasePatternAlphabetPatternSetMap.containsKey(basePatternPatternAlphabetMap.get(patternAtIndex))){
						tandemArrayBasePatternAlphabetPatternSet = entireLogTandemArrayBasePatternAlphabetPatternSetMap.get(basePatternPatternAlphabetMap.get(patternAtIndex));
						if(tandemArrayBasePatternAlphabetPatternSet.contains(patternAtIndex)){
							patternExistsAtIndex = true;
							contributingShorterBasePattern = patternAtIndex;
							if(contributingShorterBasePattern.length() != encodingLength)
								break;
						}
					}
				}
				if(patternExistsAtIndex){
					contributingShorterBasePatternAlphabet = basePatternPatternAlphabetMap.get(contributingShorterBasePattern);
					
					contributingTandemArrayShorterBasePatternSet.add(contributingShorterBasePattern);
					encodedTraceContributingShorterTandemArrayBasePatternAlphabetSet.add(contributingShorterBasePatternAlphabet);
					
					count = noIdenticalTraces;
					if(tandemArrayShorterBasePatternAlphabetCountMap.containsKey(contributingShorterBasePatternAlphabet)){
						count += tandemArrayShorterBasePatternAlphabetCountMap.get(contributingShorterBasePatternAlphabet);
					}
					tandemArrayShorterBasePatternAlphabetCountMap.put(contributingShorterBasePatternAlphabet, count);
	
					count = 1;
					if(tandemArrayShorterBasePatternAlphabetNonDuplicateCountMap.containsKey(contributingShorterBasePatternAlphabet))
						count += tandemArrayShorterBasePatternAlphabetNonDuplicateCountMap.get(contributingShorterBasePatternAlphabet);
					tandemArrayShorterBasePatternAlphabetNonDuplicateCountMap.put(contributingShorterBasePatternAlphabet, count);
					
					nextIndex = index+contributingShorterBasePattern.length()/encodingLength;
				}
			}
			//Update the instance counts
			for(TreeSet<String> contributingBasePatternAlphabet : encodedTraceContributingShorterTandemArrayBasePatternAlphabetSet){
				instanceCount = noIdenticalTraces;
				if(tandemArrayShorterBasePatternAlphabetInstanceCountMap.containsKey(contributingBasePatternAlphabet)){
					instanceCount += tandemArrayShorterBasePatternAlphabetInstanceCountMap.get(contributingBasePatternAlphabet);
				}
				tandemArrayShorterBasePatternAlphabetInstanceCountMap.put(contributingBasePatternAlphabet, instanceCount);
			}
			for(TreeSet<String> contributingBasePatternAlphabet : encodedTraceContributingShorterTandemArrayBasePatternAlphabetSet){
				instanceCount = 1;
				if(tandemArrayShorterBasePatternAlphabetNonDuplicateInstanceCountMap.containsKey(contributingBasePatternAlphabet)){
					instanceCount += tandemArrayShorterBasePatternAlphabetNonDuplicateInstanceCountMap.get(contributingBasePatternAlphabet);
				}
				tandemArrayShorterBasePatternAlphabetNonDuplicateInstanceCountMap.put(contributingBasePatternAlphabet, instanceCount);
			}

			// Do it for maximal repeats
			nextIndex = -1;
			for(Integer index : sortedIndexSet){
				if(index < nextIndex)
					continue;

				patternExistsAtIndex = false;
				contributingShorterBasePattern = "";
				indexBasePatternSet = indexBasePatternSetMap.get(index);
				
				Iterator<String> it = indexBasePatternSet.iterator();
				while(it.hasNext()){
					patternAtIndex = it.next();
					if(entireLogMaximalRepeatBasePatternAlphabetPatternSetMap.containsKey(basePatternPatternAlphabetMap.get(patternAtIndex))){
						maximalRepeatBasePatternAlphabetPatternSet = entireLogMaximalRepeatBasePatternAlphabetPatternSetMap.get(basePatternPatternAlphabetMap.get(patternAtIndex));
						if(maximalRepeatBasePatternAlphabetPatternSet.contains(patternAtIndex)){
							patternExistsAtIndex = true;
							contributingShorterBasePattern = patternAtIndex;
							if(contributingShorterBasePattern.length() != encodingLength)
								break;
						}
					}
				}
				if(patternExistsAtIndex){
					contributingShorterBasePatternAlphabet = basePatternPatternAlphabetMap.get(contributingShorterBasePattern);
					
					contributingMaximalRepeatShorterBasePatternSet.add(contributingShorterBasePattern);
					encodedTraceContributingShorterMaximalRepeatBasePatternAlphabetSet.add(contributingShorterBasePatternAlphabet);
					
					count = noIdenticalTraces;
					if(maximalRepeatShorterBasePatternAlphabetCountMap.containsKey(contributingShorterBasePatternAlphabet)){
						count += maximalRepeatShorterBasePatternAlphabetCountMap.get(contributingShorterBasePatternAlphabet);
					}
					maximalRepeatShorterBasePatternAlphabetCountMap.put(contributingShorterBasePatternAlphabet, count);
	
					count = 1;
					if(maximalRepeatShorterBasePatternAlphabetNonDuplicateCountMap.containsKey(contributingShorterBasePatternAlphabet))
						count += maximalRepeatShorterBasePatternAlphabetNonDuplicateCountMap.get(contributingShorterBasePatternAlphabet);
					maximalRepeatShorterBasePatternAlphabetNonDuplicateCountMap.put(contributingShorterBasePatternAlphabet, count);
					
					nextIndex = index+contributingShorterBasePattern.length()/encodingLength;
				}
			}
			//Update the instance counts
			for(TreeSet<String> contributingBasePatternAlphabet : encodedTraceContributingShorterMaximalRepeatBasePatternAlphabetSet){
				instanceCount = noIdenticalTraces;
				if(maximalRepeatShorterBasePatternAlphabetInstanceCountMap.containsKey(contributingBasePatternAlphabet)){
					instanceCount += maximalRepeatShorterBasePatternAlphabetInstanceCountMap.get(contributingBasePatternAlphabet);
				}
				maximalRepeatShorterBasePatternAlphabetInstanceCountMap.put(contributingBasePatternAlphabet, instanceCount);
			}
			for(TreeSet<String> contributingBasePatternAlphabet : encodedTraceContributingShorterMaximalRepeatBasePatternAlphabetSet){
				instanceCount = 1;
				if(maximalRepeatShorterBasePatternAlphabetNonDuplicateInstanceCountMap.containsKey(contributingBasePatternAlphabet)){
					instanceCount += maximalRepeatShorterBasePatternAlphabetNonDuplicateInstanceCountMap.get(contributingBasePatternAlphabet);
				}
				maximalRepeatShorterBasePatternAlphabetNonDuplicateInstanceCountMap.put(contributingBasePatternAlphabet, instanceCount);
			}
			
			/*
			 * Update the counts, do it for longer patterns in combined pattern alphabet, 
			 */
			nextIndex = -1;
			for(Integer index : sortedIndexSet){
				if(index < nextIndex)
					continue;
				
				indexBasePatternSet = indexBasePatternSetMap.get(index);
				
				contributingLongerBasePattern = indexBasePatternSet.last();
				contributingLongerBasePatternAlphabet = basePatternPatternAlphabetMap.get(contributingLongerBasePattern);
				
				contributingCombinedLongerBasePatternSet.add(contributingLongerBasePattern);
				encodedTraceContributingLongerCombinedBasePatternAlphabetSet.add(contributingLongerBasePatternAlphabet);
				
				count = noIdenticalTraces;
				if(combinedLongerBasePatternAlphabetCountMap.containsKey(contributingLongerBasePatternAlphabet)){
					count += combinedLongerBasePatternAlphabetCountMap.get(contributingLongerBasePatternAlphabet);
				}
				combinedLongerBasePatternAlphabetCountMap.put(contributingLongerBasePatternAlphabet, count);

				count = 1;
				if(combinedLongerBasePatternAlphabetNonDuplicateCountMap.containsKey(contributingLongerBasePatternAlphabet))
					count += combinedLongerBasePatternAlphabetNonDuplicateCountMap.get(contributingLongerBasePatternAlphabet);
				combinedLongerBasePatternAlphabetNonDuplicateCountMap.put(contributingLongerBasePatternAlphabet, count);
				
				nextIndex = index+contributingLongerBasePattern.length()/encodingLength;
			}
			//Update the instance counts
			for(TreeSet<String> contributingBasePatternAlphabet : encodedTraceContributingLongerCombinedBasePatternAlphabetSet){
				instanceCount = noIdenticalTraces;
				if(combinedLongerBasePatternAlphabetInstanceCountMap.containsKey(contributingBasePatternAlphabet)){
					instanceCount += combinedLongerBasePatternAlphabetInstanceCountMap.get(contributingBasePatternAlphabet);
				}
				combinedLongerBasePatternAlphabetInstanceCountMap.put(contributingBasePatternAlphabet, instanceCount);
			}
			for(TreeSet<String> contributingBasePatternAlphabet : encodedTraceContributingLongerCombinedBasePatternAlphabetSet){
				instanceCount = 1;
				if(combinedLongerBasePatternAlphabetNonDuplicateInstanceCountMap.containsKey(contributingBasePatternAlphabet)){
					instanceCount += combinedLongerBasePatternAlphabetNonDuplicateInstanceCountMap.get(contributingBasePatternAlphabet);
				}
				combinedLongerBasePatternAlphabetNonDuplicateInstanceCountMap.put(contributingBasePatternAlphabet, instanceCount);
			}
			
			// Do it for tandem arrays
			nextIndex = -1;
			for(Integer index : sortedIndexSet){
				if(index < nextIndex)
					continue;

				patternExistsAtIndex = false;
				contributingLongerBasePattern = "";
				indexBasePatternSet = indexBasePatternSetMap.get(index);
				
				int noBasePatterns = indexBasePatternSet.size();
				Object[] indexBasePatternSetAsArray = indexBasePatternSet.toArray();
				for(int i = noBasePatterns-1; i >= 0; i--){
					patternAtIndex = (String)indexBasePatternSetAsArray[i];
					if(entireLogTandemArrayBasePatternAlphabetPatternSetMap.containsKey(basePatternPatternAlphabetMap.get(patternAtIndex))){
						tandemArrayBasePatternAlphabetPatternSet = entireLogTandemArrayBasePatternAlphabetPatternSetMap.get(basePatternPatternAlphabetMap.get(patternAtIndex));
						if(tandemArrayBasePatternAlphabetPatternSet.contains(patternAtIndex)){
							patternExistsAtIndex = true;
							contributingLongerBasePattern = patternAtIndex;
							break;
						}
					}
				}
				if(patternExistsAtIndex){
					contributingLongerBasePatternAlphabet = basePatternPatternAlphabetMap.get(contributingLongerBasePattern);
					contributingTandemArrayLongerBasePatternSet.add(contributingLongerBasePattern);
					encodedTraceContributingLongerTandemArrayBasePatternAlphabetSet.add(contributingLongerBasePatternAlphabet);
					
					count = noIdenticalTraces;
					if(tandemArrayLongerBasePatternAlphabetCountMap.containsKey(contributingLongerBasePatternAlphabet)){
						count += tandemArrayLongerBasePatternAlphabetCountMap.get(contributingLongerBasePatternAlphabet);
					}
					tandemArrayLongerBasePatternAlphabetCountMap.put(contributingLongerBasePatternAlphabet, count);
	
					count = 1;
					if(tandemArrayLongerBasePatternAlphabetNonDuplicateCountMap.containsKey(contributingLongerBasePatternAlphabet))
						count += tandemArrayLongerBasePatternAlphabetNonDuplicateCountMap.get(contributingLongerBasePatternAlphabet);
					tandemArrayLongerBasePatternAlphabetNonDuplicateCountMap.put(contributingLongerBasePatternAlphabet, count);
					
					nextIndex = index+contributingLongerBasePattern.length()/encodingLength;
				}
			}
			//Update the instance counts
			for(TreeSet<String> contributingBasePatternAlphabet : encodedTraceContributingLongerTandemArrayBasePatternAlphabetSet){
				instanceCount = noIdenticalTraces;
				if(tandemArrayLongerBasePatternAlphabetInstanceCountMap.containsKey(contributingBasePatternAlphabet)){
					instanceCount += tandemArrayLongerBasePatternAlphabetInstanceCountMap.get(contributingBasePatternAlphabet);
				}
				tandemArrayLongerBasePatternAlphabetInstanceCountMap.put(contributingBasePatternAlphabet, instanceCount);
			}
			for(TreeSet<String> contributingBasePatternAlphabet : encodedTraceContributingLongerTandemArrayBasePatternAlphabetSet){
				instanceCount = 1;
				if(tandemArrayLongerBasePatternAlphabetNonDuplicateInstanceCountMap.containsKey(contributingBasePatternAlphabet)){
					instanceCount += tandemArrayLongerBasePatternAlphabetNonDuplicateInstanceCountMap.get(contributingBasePatternAlphabet);
				}
				tandemArrayLongerBasePatternAlphabetNonDuplicateInstanceCountMap.put(contributingBasePatternAlphabet, instanceCount);
			}

			// Do it for maximal repeats
			nextIndex = -1;
			for(Integer index : sortedIndexSet){
				if(index < nextIndex)
					continue;

				patternExistsAtIndex = false;
				contributingLongerBasePattern = "";
				indexBasePatternSet = indexBasePatternSetMap.get(index);
				
				int noBasePatterns = indexBasePatternSet.size();
				Object[] indexBasePatternSetAsArray = indexBasePatternSet.toArray();
				for(int i = noBasePatterns-1; i >= 0; i--){
					patternAtIndex = (String)indexBasePatternSetAsArray[i];
					if(entireLogMaximalRepeatBasePatternAlphabetPatternSetMap.containsKey(basePatternPatternAlphabetMap.get(patternAtIndex))){
						maximalRepeatBasePatternAlphabetPatternSet = entireLogMaximalRepeatBasePatternAlphabetPatternSetMap.get(basePatternPatternAlphabetMap.get(patternAtIndex));
						if(maximalRepeatBasePatternAlphabetPatternSet.contains(patternAtIndex)){
							patternExistsAtIndex = true;
							contributingLongerBasePattern = patternAtIndex;
							break;
						}
					}
				}
				if(patternExistsAtIndex){
					contributingLongerBasePatternAlphabet = basePatternPatternAlphabetMap.get(contributingLongerBasePattern);
					contributingMaximalRepeatLongerBasePatternSet.add(contributingLongerBasePattern);
					encodedTraceContributingLongerMaximalRepeatBasePatternAlphabetSet.add(contributingLongerBasePatternAlphabet);
					
					count = noIdenticalTraces;
					if(maximalRepeatLongerBasePatternAlphabetCountMap.containsKey(contributingLongerBasePatternAlphabet)){
						count += maximalRepeatLongerBasePatternAlphabetCountMap.get(contributingLongerBasePatternAlphabet);
					}
					maximalRepeatLongerBasePatternAlphabetCountMap.put(contributingLongerBasePatternAlphabet, count);
	
					count = 1;
					if(maximalRepeatLongerBasePatternAlphabetNonDuplicateCountMap.containsKey(contributingLongerBasePatternAlphabet))
						count += maximalRepeatLongerBasePatternAlphabetNonDuplicateCountMap.get(contributingLongerBasePatternAlphabet);
					maximalRepeatLongerBasePatternAlphabetNonDuplicateCountMap.put(contributingLongerBasePatternAlphabet, count);
					
					nextIndex = index+contributingLongerBasePattern.length()/encodingLength;
				}
			}
			
			//Update the instance counts
			for(TreeSet<String> contributingBasePatternAlphabet : encodedTraceContributingLongerMaximalRepeatBasePatternAlphabetSet){
				instanceCount = noIdenticalTraces;
				if(maximalRepeatLongerBasePatternAlphabetInstanceCountMap.containsKey(contributingBasePatternAlphabet)){
					instanceCount += maximalRepeatLongerBasePatternAlphabetInstanceCountMap.get(contributingBasePatternAlphabet);
				}
				maximalRepeatLongerBasePatternAlphabetInstanceCountMap.put(contributingBasePatternAlphabet, instanceCount);
			}
			for(TreeSet<String> contributingBasePatternAlphabet : encodedTraceContributingLongerMaximalRepeatBasePatternAlphabetSet){
				instanceCount = 1;
				if(maximalRepeatLongerBasePatternAlphabetNonDuplicateInstanceCountMap.containsKey(contributingBasePatternAlphabet)){
					instanceCount += maximalRepeatLongerBasePatternAlphabetNonDuplicateInstanceCountMap.get(contributingBasePatternAlphabet);
				}
				maximalRepeatLongerBasePatternAlphabetNonDuplicateInstanceCountMap.put(contributingBasePatternAlphabet, instanceCount);
			}
		}
		
		//The instance counts contain absolute count values, we need to change them to percentages
		int noTraces = encodedLog.getNoTraces();
		int noUniqueTraces = encodedLog.getNoUniqueTraces();
		
		int instanceCountPercentage;
		for(TreeSet<String> combinedBasePatternAlphabet : combinedShorterBasePatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(combinedShorterBasePatternAlphabetInstanceCountMap.get(combinedBasePatternAlphabet)*100.0f/noTraces);
			combinedShorterBasePatternAlphabetInstanceCountMap.put(combinedBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> combinedBasePatternAlphabet : combinedShorterBasePatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(combinedShorterBasePatternAlphabetNonDuplicateInstanceCountMap.get(combinedBasePatternAlphabet)*100.0f/noUniqueTraces);
			combinedShorterBasePatternAlphabetNonDuplicateInstanceCountMap.put(combinedBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> combinedPatternAlphabet : combinedShorterPatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(combinedShorterPatternAlphabetInstanceCountMap.get(combinedPatternAlphabet)*100.0f/noTraces);
			combinedShorterPatternAlphabetInstanceCountMap.put(combinedPatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> combinedPatternAlphabet : combinedShorterPatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(combinedShorterPatternAlphabetNonDuplicateInstanceCountMap.get(combinedPatternAlphabet)*100.0f/noUniqueTraces);
			combinedShorterPatternAlphabetNonDuplicateInstanceCountMap.put(combinedPatternAlphabet, instanceCountPercentage);
		}
		
		for(TreeSet<String> tandemArrayBasePatternAlphabet : tandemArrayShorterBasePatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(tandemArrayShorterBasePatternAlphabetInstanceCountMap.get(tandemArrayBasePatternAlphabet)*100.0f/noTraces);
			tandemArrayShorterBasePatternAlphabetInstanceCountMap.put(tandemArrayBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> tandemArrayBasePatternAlphabet : tandemArrayShorterBasePatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(tandemArrayShorterBasePatternAlphabetNonDuplicateInstanceCountMap.get(tandemArrayBasePatternAlphabet)*100.0f/noUniqueTraces);
			tandemArrayShorterBasePatternAlphabetNonDuplicateInstanceCountMap.put(tandemArrayBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> tandemArrayPatternAlphabet : tandemArrayShorterPatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(tandemArrayShorterPatternAlphabetInstanceCountMap.get(tandemArrayPatternAlphabet)*100.0f/noTraces);
			tandemArrayShorterPatternAlphabetInstanceCountMap.put(tandemArrayPatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> tandemArrayPatternAlphabet : tandemArrayShorterPatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(tandemArrayShorterPatternAlphabetNonDuplicateInstanceCountMap.get(tandemArrayPatternAlphabet)*100.0f/noUniqueTraces);
			tandemArrayShorterPatternAlphabetNonDuplicateInstanceCountMap.put(tandemArrayPatternAlphabet, instanceCountPercentage);
		}
		
		for(TreeSet<String> maximalRepeatBasePatternAlphabet : maximalRepeatShorterBasePatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(maximalRepeatShorterBasePatternAlphabetInstanceCountMap.get(maximalRepeatBasePatternAlphabet)*100.0f/noTraces);
			maximalRepeatShorterBasePatternAlphabetInstanceCountMap.put(maximalRepeatBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> maximalRepeatBasePatternAlphabet : maximalRepeatShorterBasePatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(maximalRepeatShorterBasePatternAlphabetNonDuplicateInstanceCountMap.get(maximalRepeatBasePatternAlphabet)*100.0f/noUniqueTraces);
			maximalRepeatShorterBasePatternAlphabetNonDuplicateInstanceCountMap.put(maximalRepeatBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> maximalRepeatPatternAlphabet : maximalRepeatShorterPatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(maximalRepeatShorterPatternAlphabetInstanceCountMap.get(maximalRepeatPatternAlphabet)*100.0f/noTraces);
			maximalRepeatShorterPatternAlphabetInstanceCountMap.put(maximalRepeatPatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> maximalRepeatPatternAlphabet : maximalRepeatShorterPatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(maximalRepeatShorterPatternAlphabetNonDuplicateInstanceCountMap.get(maximalRepeatPatternAlphabet)*100.0f/noUniqueTraces);
			maximalRepeatShorterPatternAlphabetNonDuplicateInstanceCountMap.put(maximalRepeatPatternAlphabet, instanceCountPercentage);
		}
		
		//DO it for Longer Pattern Metrics
		for(TreeSet<String> combinedBasePatternAlphabet : combinedLongerBasePatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(combinedLongerBasePatternAlphabetInstanceCountMap.get(combinedBasePatternAlphabet)*100.0f/noTraces);
			combinedLongerBasePatternAlphabetInstanceCountMap.put(combinedBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> combinedBasePatternAlphabet : combinedLongerBasePatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(combinedLongerBasePatternAlphabetNonDuplicateInstanceCountMap.get(combinedBasePatternAlphabet)*100.0f/noUniqueTraces);
			combinedLongerBasePatternAlphabetNonDuplicateInstanceCountMap.put(combinedBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> combinedPatternAlphabet : combinedLongerPatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(combinedLongerPatternAlphabetInstanceCountMap.get(combinedPatternAlphabet)*100.0f/noTraces);
			combinedLongerPatternAlphabetInstanceCountMap.put(combinedPatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> combinedPatternAlphabet : combinedLongerPatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(combinedLongerPatternAlphabetNonDuplicateInstanceCountMap.get(combinedPatternAlphabet)*100.0f/noUniqueTraces);
			combinedLongerPatternAlphabetNonDuplicateInstanceCountMap.put(combinedPatternAlphabet, instanceCountPercentage);
		}
		
		for(TreeSet<String> tandemArrayBasePatternAlphabet : tandemArrayLongerBasePatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(tandemArrayLongerBasePatternAlphabetInstanceCountMap.get(tandemArrayBasePatternAlphabet)*100.0f/noTraces);
			tandemArrayLongerBasePatternAlphabetInstanceCountMap.put(tandemArrayBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> tandemArrayBasePatternAlphabet : tandemArrayLongerBasePatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(tandemArrayLongerBasePatternAlphabetNonDuplicateInstanceCountMap.get(tandemArrayBasePatternAlphabet)*100.0f/noUniqueTraces);
			tandemArrayLongerBasePatternAlphabetNonDuplicateInstanceCountMap.put(tandemArrayBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> tandemArrayPatternAlphabet : tandemArrayLongerPatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(tandemArrayLongerPatternAlphabetInstanceCountMap.get(tandemArrayPatternAlphabet)*100.0f/noTraces);
			tandemArrayLongerPatternAlphabetInstanceCountMap.put(tandemArrayPatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> tandemArrayPatternAlphabet : tandemArrayLongerPatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(tandemArrayLongerPatternAlphabetNonDuplicateInstanceCountMap.get(tandemArrayPatternAlphabet)*100.0f/noUniqueTraces);
			tandemArrayLongerPatternAlphabetNonDuplicateInstanceCountMap.put(tandemArrayPatternAlphabet, instanceCountPercentage);
		}
		
		for(TreeSet<String> maximalRepeatBasePatternAlphabet : maximalRepeatLongerBasePatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(maximalRepeatLongerBasePatternAlphabetInstanceCountMap.get(maximalRepeatBasePatternAlphabet)*100.0f/noTraces);
			maximalRepeatLongerBasePatternAlphabetInstanceCountMap.put(maximalRepeatBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> maximalRepeatBasePatternAlphabet : maximalRepeatLongerBasePatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(maximalRepeatLongerBasePatternAlphabetNonDuplicateInstanceCountMap.get(maximalRepeatBasePatternAlphabet)*100.0f/noUniqueTraces);
			maximalRepeatLongerBasePatternAlphabetNonDuplicateInstanceCountMap.put(maximalRepeatBasePatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> maximalRepeatPatternAlphabet : maximalRepeatLongerPatternAlphabetInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(maximalRepeatLongerPatternAlphabetInstanceCountMap.get(maximalRepeatPatternAlphabet)*100.0f/noTraces);
			maximalRepeatLongerPatternAlphabetInstanceCountMap.put(maximalRepeatPatternAlphabet, instanceCountPercentage);
		}
		for(TreeSet<String> maximalRepeatPatternAlphabet : maximalRepeatLongerPatternAlphabetNonDuplicateInstanceCountMap.keySet()){
			instanceCountPercentage = Math.round(maximalRepeatLongerPatternAlphabetNonDuplicateInstanceCountMap.get(maximalRepeatPatternAlphabet)*100.0f/noUniqueTraces);
			maximalRepeatLongerPatternAlphabetNonDuplicateInstanceCountMap.put(maximalRepeatPatternAlphabet, instanceCountPercentage);
		}
		
		EquivalenceClass equivalenceClass = new EquivalenceClass();
		nonOverlappingGlobalShorterPatternAlphabetCount.setCombinedPatternAlphabetCountMap(combinedShorterPatternAlphabetCountMap);
		nonOverlappingGlobalShorterPatternAlphabetCount.setCombinedPatternAlphabetInstanceCountMap(combinedShorterPatternAlphabetInstanceCountMap);
		nonOverlappingGlobalShorterPatternAlphabetCount.setCombinedPatternAlphabetNonDuplicateCountMap(combinedShorterPatternAlphabetNonDuplicateCountMap);
		nonOverlappingGlobalShorterPatternAlphabetCount.setCombinedPatternAlphabetNonDuplicateInstanceCountMap(combinedShorterPatternAlphabetNonDuplicateInstanceCountMap);
		
		nonOverlappingGlobalShorterPatternAlphabetCount.setCombinedBasePatternAlphabetCountMap(combinedShorterBasePatternAlphabetCountMap);
		nonOverlappingGlobalShorterPatternAlphabetCount.setCombinedBasePatternAlphabetInstanceCountMap(combinedShorterBasePatternAlphabetInstanceCountMap);
		nonOverlappingGlobalShorterPatternAlphabetCount.setCombinedBasePatternAlphabetNonDuplicateCountMap(combinedShorterBasePatternAlphabetNonDuplicateCountMap);
		nonOverlappingGlobalShorterPatternAlphabetCount.setCombinedBasePatternAlphabetNonDuplicateInstanceCountMap(combinedShorterBasePatternAlphabetNonDuplicateInstanceCountMap);
		
		nonOverlappingGlobalShorterPatternAlphabetCount.setContributingCombinedPatternAlphabetPatternSetMap(equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, contributingCombinedShorterPatternSet));
		nonOverlappingGlobalShorterPatternAlphabetCount.setContributingCombinedBasePatternAlphabetPatternSetMap(equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, contributingCombinedShorterBasePatternSet));
		
		nonOverlappingGlobalShorterPatternAlphabetCount.setTandemArrayPatternAlphabetCountMap(tandemArrayShorterPatternAlphabetCountMap);
		nonOverlappingGlobalShorterPatternAlphabetCount.setTandemArrayPatternAlphabetInstanceCountMap(tandemArrayShorterPatternAlphabetInstanceCountMap);
		nonOverlappingGlobalShorterPatternAlphabetCount.setTandemArrayPatternAlphabetNonDuplicateCountMap(tandemArrayShorterPatternAlphabetNonDuplicateCountMap);
		nonOverlappingGlobalShorterPatternAlphabetCount.setTandemArrayPatternAlphabetNonDuplicateInstanceCountMap(tandemArrayShorterPatternAlphabetNonDuplicateInstanceCountMap);
		
		nonOverlappingGlobalShorterPatternAlphabetCount.setTandemArrayBasePatternAlphabetCountMap(tandemArrayShorterBasePatternAlphabetCountMap);
		nonOverlappingGlobalShorterPatternAlphabetCount.setTandemArrayBasePatternAlphabetInstanceCountMap(tandemArrayShorterBasePatternAlphabetInstanceCountMap);
		nonOverlappingGlobalShorterPatternAlphabetCount.setTandemArrayBasePatternAlphabetNonDuplicateCountMap(tandemArrayShorterBasePatternAlphabetNonDuplicateCountMap);
		nonOverlappingGlobalShorterPatternAlphabetCount.setTandemArrayBasePatternAlphabetNonDuplicateInstanceCountMap(tandemArrayShorterBasePatternAlphabetNonDuplicateInstanceCountMap);
		
		nonOverlappingGlobalShorterPatternAlphabetCount.setContributingTandemArrayPatternAlphabetPatternSetMap(equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, contributingTandemArrayShorterPatternSet));
		nonOverlappingGlobalShorterPatternAlphabetCount.setContributingTandemArrayBasePatternAlphabetPatternSetMap(equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, contributingTandemArrayShorterBasePatternSet));

		nonOverlappingGlobalShorterPatternAlphabetCount.setMaximalRepeatPatternAlphabetCountMap(maximalRepeatShorterPatternAlphabetCountMap);
		nonOverlappingGlobalShorterPatternAlphabetCount.setMaximalRepeatPatternAlphabetInstanceCountMap(maximalRepeatShorterPatternAlphabetInstanceCountMap);
		nonOverlappingGlobalShorterPatternAlphabetCount.setMaximalRepeatPatternAlphabetNonDuplicateCountMap(maximalRepeatShorterPatternAlphabetNonDuplicateCountMap);
		nonOverlappingGlobalShorterPatternAlphabetCount.setMaximalRepeatPatternAlphabetNonDuplicateInstanceCountMap(maximalRepeatShorterPatternAlphabetNonDuplicateInstanceCountMap);
		
		nonOverlappingGlobalShorterPatternAlphabetCount.setMaximalRepeatBasePatternAlphabetCountMap(maximalRepeatShorterBasePatternAlphabetCountMap);
		nonOverlappingGlobalShorterPatternAlphabetCount.setMaximalRepeatBasePatternAlphabetInstanceCountMap(maximalRepeatShorterBasePatternAlphabetInstanceCountMap);
		nonOverlappingGlobalShorterPatternAlphabetCount.setMaximalRepeatBasePatternAlphabetNonDuplicateCountMap(maximalRepeatShorterBasePatternAlphabetNonDuplicateCountMap);
		nonOverlappingGlobalShorterPatternAlphabetCount.setMaximalRepeatBasePatternAlphabetNonDuplicateInstanceCountMap(maximalRepeatShorterBasePatternAlphabetNonDuplicateInstanceCountMap);
		
		nonOverlappingGlobalShorterPatternAlphabetCount.setContributingMaximalRepeatPatternAlphabetPatternSetMap(equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, contributingMaximalRepeatShorterPatternSet));
		nonOverlappingGlobalShorterPatternAlphabetCount.setContributingMaximalRepeatBasePatternAlphabetPatternSetMap(equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, contributingMaximalRepeatShorterBasePatternSet));

		
		nonOverlappingGlobalLongerPatternAlphabetCount.setCombinedPatternAlphabetCountMap(combinedLongerPatternAlphabetCountMap);
		nonOverlappingGlobalLongerPatternAlphabetCount.setCombinedPatternAlphabetInstanceCountMap(combinedLongerPatternAlphabetInstanceCountMap);
		nonOverlappingGlobalLongerPatternAlphabetCount.setCombinedPatternAlphabetNonDuplicateCountMap(combinedLongerPatternAlphabetNonDuplicateCountMap);
		nonOverlappingGlobalLongerPatternAlphabetCount.setCombinedPatternAlphabetNonDuplicateInstanceCountMap(combinedLongerPatternAlphabetNonDuplicateInstanceCountMap);
		
		nonOverlappingGlobalLongerPatternAlphabetCount.setCombinedBasePatternAlphabetCountMap(combinedLongerBasePatternAlphabetCountMap);
		nonOverlappingGlobalLongerPatternAlphabetCount.setCombinedBasePatternAlphabetInstanceCountMap(combinedLongerBasePatternAlphabetInstanceCountMap);
		nonOverlappingGlobalLongerPatternAlphabetCount.setCombinedBasePatternAlphabetNonDuplicateCountMap(combinedLongerBasePatternAlphabetNonDuplicateCountMap);
		nonOverlappingGlobalLongerPatternAlphabetCount.setCombinedBasePatternAlphabetNonDuplicateInstanceCountMap(combinedLongerBasePatternAlphabetNonDuplicateInstanceCountMap);
		
		nonOverlappingGlobalLongerPatternAlphabetCount.setContributingCombinedPatternAlphabetPatternSetMap(equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, contributingCombinedLongerPatternSet));
		nonOverlappingGlobalLongerPatternAlphabetCount.setContributingCombinedBasePatternAlphabetPatternSetMap(equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, contributingCombinedLongerBasePatternSet));
		
		nonOverlappingGlobalLongerPatternAlphabetCount.setTandemArrayPatternAlphabetCountMap(tandemArrayLongerPatternAlphabetCountMap);
		nonOverlappingGlobalLongerPatternAlphabetCount.setTandemArrayPatternAlphabetInstanceCountMap(tandemArrayLongerPatternAlphabetInstanceCountMap);
		nonOverlappingGlobalLongerPatternAlphabetCount.setTandemArrayPatternAlphabetNonDuplicateCountMap(tandemArrayLongerPatternAlphabetNonDuplicateCountMap);
		nonOverlappingGlobalLongerPatternAlphabetCount.setTandemArrayPatternAlphabetNonDuplicateInstanceCountMap(tandemArrayLongerPatternAlphabetNonDuplicateInstanceCountMap);
		
		nonOverlappingGlobalLongerPatternAlphabetCount.setTandemArrayBasePatternAlphabetCountMap(tandemArrayLongerBasePatternAlphabetCountMap);
		nonOverlappingGlobalLongerPatternAlphabetCount.setTandemArrayBasePatternAlphabetInstanceCountMap(tandemArrayLongerBasePatternAlphabetInstanceCountMap);
		nonOverlappingGlobalLongerPatternAlphabetCount.setTandemArrayBasePatternAlphabetNonDuplicateCountMap(tandemArrayLongerBasePatternAlphabetNonDuplicateCountMap);
		nonOverlappingGlobalLongerPatternAlphabetCount.setTandemArrayBasePatternAlphabetNonDuplicateInstanceCountMap(tandemArrayLongerBasePatternAlphabetNonDuplicateInstanceCountMap);
		
		nonOverlappingGlobalLongerPatternAlphabetCount.setContributingTandemArrayPatternAlphabetPatternSetMap(equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, contributingTandemArrayLongerPatternSet));
		nonOverlappingGlobalLongerPatternAlphabetCount.setContributingTandemArrayBasePatternAlphabetPatternSetMap(equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, contributingTandemArrayLongerBasePatternSet));

		nonOverlappingGlobalLongerPatternAlphabetCount.setMaximalRepeatPatternAlphabetCountMap(maximalRepeatLongerPatternAlphabetCountMap);
		nonOverlappingGlobalLongerPatternAlphabetCount.setMaximalRepeatPatternAlphabetInstanceCountMap(maximalRepeatLongerPatternAlphabetInstanceCountMap);
		nonOverlappingGlobalLongerPatternAlphabetCount.setMaximalRepeatPatternAlphabetNonDuplicateCountMap(maximalRepeatLongerPatternAlphabetNonDuplicateCountMap);
		nonOverlappingGlobalLongerPatternAlphabetCount.setMaximalRepeatPatternAlphabetNonDuplicateInstanceCountMap(maximalRepeatLongerPatternAlphabetNonDuplicateInstanceCountMap);
		
		nonOverlappingGlobalLongerPatternAlphabetCount.setMaximalRepeatBasePatternAlphabetCountMap(maximalRepeatLongerBasePatternAlphabetCountMap);
		nonOverlappingGlobalLongerPatternAlphabetCount.setMaximalRepeatBasePatternAlphabetInstanceCountMap(maximalRepeatLongerBasePatternAlphabetInstanceCountMap);
		nonOverlappingGlobalLongerPatternAlphabetCount.setMaximalRepeatBasePatternAlphabetNonDuplicateCountMap(maximalRepeatLongerBasePatternAlphabetNonDuplicateCountMap);
		nonOverlappingGlobalLongerPatternAlphabetCount.setMaximalRepeatBasePatternAlphabetNonDuplicateInstanceCountMap(maximalRepeatLongerBasePatternAlphabetNonDuplicateInstanceCountMap);
		
		nonOverlappingGlobalLongerPatternAlphabetCount.setContributingMaximalRepeatPatternAlphabetPatternSetMap(equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, contributingMaximalRepeatLongerPatternSet));
		nonOverlappingGlobalLongerPatternAlphabetCount.setContributingMaximalRepeatBasePatternAlphabetPatternSetMap(equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, contributingMaximalRepeatLongerBasePatternSet));
	}
	
	private void computeConservednessMetric(){
		MetricPatternAlphabetCount metricPatternAlphabetCount;
		if(patternLengthPreference == PatternLengthPreference.Shorter)
			metricPatternAlphabetCount = nonOverlappingShorterPatternAlphabetCount;
		else
			metricPatternAlphabetCount = nonOverlappingLongerPatternAlphabetCount;
		
		conservednessMetric = new ConservednessMetric();
		
		conservednessMetric.setCombinedPatternAlphabetConservednessMap(getConservednessMetric(metricPatternAlphabetCount.getCombinedPatternAlphabetCountMap(), false));
		conservednessMetric.setCombinedPatternAlphabetNonDuplicateConservednessMap(getConservednessMetric(metricPatternAlphabetCount.getCombinedPatternAlphabetNonDuplicateCountMap(), true));
		conservednessMetric.setCombinedBasePatternAlphabetConservednessMap(getConservednessMetric(metricPatternAlphabetCount.getCombinedBasePatternAlphabetCountMap(), false));
		conservednessMetric.setCombinedBasePatternAlphabetNonDuplicateConservednessMap(getConservednessMetric(metricPatternAlphabetCount.getCombinedBasePatternAlphabetNonDuplicateCountMap(), true));
		
		conservednessMetric.setTandemArrayPatternAlphabetConservednessMap(getConservednessMetric(metricPatternAlphabetCount.getTandemArrayPatternAlphabetCountMap(), false));
		conservednessMetric.setTandemArrayPatternAlphabetNonDuplicateConservednessMap(getConservednessMetric(metricPatternAlphabetCount.getTandemArrayPatternAlphabetNonDuplicateCountMap(), true));
		conservednessMetric.setTandemArrayBasePatternAlphabetConservednessMap(getConservednessMetric(metricPatternAlphabetCount.getTandemArrayBasePatternAlphabetCountMap(), false));
		conservednessMetric.setTandemArrayBasePatternAlphabetNonDuplicateConservednessMap(getConservednessMetric(metricPatternAlphabetCount.getTandemArrayBasePatternAlphabetNonDuplicateCountMap(), true));
	
		conservednessMetric.setMaximalRepeatPatternAlphabetConservednessMap(getConservednessMetric(metricPatternAlphabetCount.getMaximalRepeatPatternAlphabetCountMap(), false));
		conservednessMetric.setMaximalRepeatPatternAlphabetNonDuplicateConservednessMap(getConservednessMetric(metricPatternAlphabetCount.getMaximalRepeatPatternAlphabetNonDuplicateCountMap(), true));
		conservednessMetric.setMaximalRepeatBasePatternAlphabetConservednessMap(getConservednessMetric(metricPatternAlphabetCount.getMaximalRepeatBasePatternAlphabetCountMap(), false));
		conservednessMetric.setMaximalRepeatBasePatternAlphabetNonDuplicateConservednessMap(getConservednessMetric(metricPatternAlphabetCount.getMaximalRepeatBasePatternAlphabetNonDuplicateCountMap(), true));
		
//		conservednessMetric.printMetrics();
	}
	
	private Map<TreeSet<String>, Integer> getConservednessMetric(Map<TreeSet<String>, Integer> nonOverlappingPatternAlphabetCountMap, boolean isNonDuplicates){
		Map<TreeSet<String>, Integer> patternAlphabetConservednessMap = new HashMap<TreeSet<String>, Integer>();
		
		float sum, mean, stddev, conservedness;
		int activityCount;
		Map<String, Integer> activityCountMap;
		
		if(isNonDuplicates)
			activityCountMap = this.nonDuplicateActivityCountMap;
		else
			activityCountMap = this.activityCountMap;
		
		for(TreeSet<String> patternAlphabet : nonOverlappingPatternAlphabetCountMap.keySet()){
			sum = 0.0f;
			for(String encodedActivity : patternAlphabet)
				sum += activityCountMap.get(encodedActivity);
			mean = sum/patternAlphabet.size();
			
			stddev = 0;
			for(String encodedActivity : patternAlphabet){
				activityCount = activityCountMap.get(encodedActivity);
				stddev += (activityCount-mean)*(activityCount-mean);
			}
			stddev = (float)Math.sqrt(stddev/patternAlphabet.size());
			
			conservedness = (nonOverlappingPatternAlphabetCountMap.get(patternAlphabet)/mean) * (1- stddev/mean)*100;
			patternAlphabetConservednessMap.put(patternAlphabet, new Double(Math.floor(conservedness)).intValue());
		}
		
		return patternAlphabetConservednessMap;
	}
	
	public void setMaxStringLength(int maxStrLength){
		MAX_STRLENGTH = maxStrLength;
	}
	
	public EncodedLog getEncodedLog(){
		return encodedLog;
	}
	
	protected void printPatterns(Map<TreeSet<String>, TreeSet<String>> patternAlphabetPatternSetMap){
		System.out.println("No. Pattern Alphabets: "+patternAlphabetPatternSetMap.size());
		Map<Integer, Set<TreeSet<String>>> alphabetSizeAlphabetSetMap = new HashMap<Integer, Set<TreeSet<String>>>();
		
		int index = 0;
		Set<TreeSet<String>> alphabetSizeAlphabetSet = new HashSet<TreeSet<String>>();
		int alphabetSize;
		for(TreeSet<String> patternAlphabet : patternAlphabetPatternSetMap.keySet()){
			alphabetSize = patternAlphabet.size();
			if(alphabetSizeAlphabetSetMap.containsKey(alphabetSize)){
				alphabetSizeAlphabetSet = alphabetSizeAlphabetSetMap.get(alphabetSize);
			}else{
				alphabetSizeAlphabetSet = new HashSet<TreeSet<String>>();
			}
			alphabetSizeAlphabetSet.add(patternAlphabet);
			alphabetSizeAlphabetSetMap.put(alphabetSize, alphabetSizeAlphabetSet);
		}
		
		int alphabetSizes[] = new int[alphabetSizeAlphabetSetMap.size()];
		for(Integer alpSize : alphabetSizeAlphabetSetMap.keySet())
			alphabetSizes[index++] = alpSize;
		
		
		Arrays.sort(alphabetSizes);
		System.out.println(alphabetSizeAlphabetSetMap.keySet()+" @ "+alphabetSizes.length);
		
		for(int i = 0; i < index; i++){
			alphabetSizeAlphabetSet = alphabetSizeAlphabetSetMap.get(alphabetSizes[i]);
			for(Set<String> patternAlphabet : alphabetSizeAlphabetSet){
				System.out.println(patternAlphabet+" @ "+patternAlphabetPatternSetMap.get(patternAlphabet));
			}
		}
	}

	/*
	 * A string comparator for sorting strings based on their lengths in a decreasing order. 
	 */
	private class DescendingStringLengthComparator implements Comparator<String> {
		public int compare(String str1, String str2) {
			if(str1.equals(str2))
				return 0;
			return (str1.length() < str2.length()) ? 1 : -1;
		}
	}

	/*
	 * A string comparator for sorting strings based on their lengths in a ascending order. 
	 */
	private class AscendingStringLengthComparator implements Comparator<String> {
		public int compare(String str1, String str2) {
			if(str1.equals(str2))
				return 0;
			return (str1.length() < str2.length()) ? -1 : 1;
		}
	}
	
	public XLog getLogSelectedPatterns(Set<String> patternSet){
		XLog selectedPatternLog = new XLogImpl(log.getAttributes());
		Set<Integer> identicalTraceIndicesSet;
		
		boolean hasRepeat;
		for (String encodedTrace : encodedLog.getUniqueEncodedTraceList()) {
			hasRepeat = false;
			for (String pattern : patternSet) {
				if (encodedTrace.contains(pattern)) {
					hasRepeat = true;
					break;
				}
			}
			if (hasRepeat) {
				identicalTraceIndicesSet = encodedLog.getEncodedTraceIdenticalTraceIndicesSetMap().get(encodedTrace);
				for (Integer traceIndex : identicalTraceIndicesSet)
					selectedPatternLog.add(log.get(traceIndex));
			}
		}
		
		return selectedPatternLog;
	}
	
	public MetricPatternAlphabetCount getOverlappingPatternAlphabetCount() {
		return overlappingPatternAlphabetCount;
	}

	public MetricPatternAlphabetCount getNonOverlappingShorterPatternAlphabetCount() {
		return nonOverlappingShorterPatternAlphabetCount;
	}

	public MetricPatternAlphabetCount getNonOverlappingLongerPatternAlphabetCount() {
		return nonOverlappingLongerPatternAlphabetCount;
	}
	
	public MetricPatternAlphabetCount getNonOverlappingGlobalShorterPatternAlphabetCount() {
		return nonOverlappingGlobalShorterPatternAlphabetCount;
	}

	public MetricPatternAlphabetCount getNonOverlappingGlobalLongerPatternAlphabetCount() {
		return nonOverlappingGlobalLongerPatternAlphabetCount;
	}
	
	public ConservednessMetric getConservednessMetric(){
		return conservednessMetric;
	}
	
	public Set<String> getEncodedAlphabet(Set<String> decodedAlphabet) {
		Set<String> encodedAlphabet = new HashSet<String>();
		for (String decodedActivity : decodedAlphabet)
			encodedAlphabet.add(encodedLog.getActivityCharMap().get(decodedActivity));
		return encodedAlphabet;
	}

	public Set<String> getDecodedAlphabet(Set<String> encodedAlphabet) {
		Set<String> decodedAlphabet = new HashSet<String>();
		for (String encodedActivity : encodedAlphabet)
			decodedAlphabet.add(encodedLog.getCharActivityMap().get(encodedActivity));
		return decodedAlphabet;
	}

	public Map<TreeSet<String>, TreeSet<String>> getEntireLogTandemArrayPatternAlphabetPatternSetMap() {
		return entireLogTandemArrayPatternAlphabetPatternSetMap;
	}

	public Map<TreeSet<String>, TreeSet<String>> getEntireLogMaximalRepeatPatternAlphabetPatternSetMap() {
		return entireLogMaximalRepeatPatternAlphabetPatternSetMap;
	}

	public Map<TreeSet<String>, TreeSet<String>> getEntireLogCombinedPatternAlphabetPatternSetMap() {
		return entireLogCombinedPatternAlphabetPatternSetMap;
	}

	public Map<TreeSet<String>, TreeSet<String>> getEntireLogTandemArrayBasePatternAlphabetPatternSetMap() {
		return entireLogTandemArrayBasePatternAlphabetPatternSetMap;
	}

	public Map<TreeSet<String>, TreeSet<String>> getEntireLogMaximalRepeatBasePatternAlphabetPatternSetMap() {
		return entireLogMaximalRepeatBasePatternAlphabetPatternSetMap;
	}

	public Map<TreeSet<String>, TreeSet<String>> getEntireLogCombinedBasePatternAlphabetPatternSetMap() {
		return entireLogCombinedBasePatternAlphabetPatternSetMap;
	}

	public PatternLengthPreference getPatternLengthPreference() {
		return patternLengthPreference;
	}

	public void setPatternLengthPreference(
			PatternLengthPreference patternLengthPreference) {
		this.patternLengthPreference = patternLengthPreference;
	}
}
